/*
 *	control.h
 *
 *	Base Control class definition.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef _OPaC_CONTROL_H_
#define	_OPaC_CONTROL_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

#ifndef _OPaC_VIEW_H_
#include <OPaC/view.h>
#endif

#ifndef	_OPaC_KEY_H_
#include <OPaC/key.h>
#endif

#ifndef	_OPaC_COLORPIXEL_H_
#include <OPaC/colorpixel.h>
#endif

/****************************************************************/

/*
 *	Describe what the control's shape will look like.
 */

OP_TYPEDEF (Card32, OPCtrlShape);				//	it is a bitset, therefore not an enum

#define	OP_CTRL_RELIEF			OP_BIT (0)		//	TRUE => control is 3D, flat otherwise
#define	OP_CTRL_ABOVE			OP_BIT (1)		//	TRUE => applicable only if relief
#define	OP_CTRL_EXTRUDE			OP_BIT (2)		//	TRUE => contents extruded from background

#define	OP_CTRL_NOFRAME			OP_BIT (7)		//	TRUE => don't draw the frame
#define	OP_CTRL_AUTOHILITE		OP_BIT (8)		//	TRUE => automatically hilite
#define	OP_CTRL_AUTOACTHILITE	OP_BIT (9)		//	TRUE => automatically actively hilite
#define	OP_CTRL_AUTOHOT			OP_BIT (10)		//	TRUE => automatically show when hot
#define	OP_CTRL_AUTOACTIVE		OP_BIT (11)		//	TRUE => automatically show when active
#define	OP_CTRL_AUTODRAG		OP_BIT (12)		//	TRUE => automatically show when drag source
#define	OP_CTRL_AUTODROP		OP_BIT (13)		//	TRUE => automatically show when drop dest.
#define	OP_CTRL_NOTMAINRESP		OP_BIT (14)		//	TRUE => does not become main responder on click
#define	OP_CTRL_NOTACTENTER		OP_BIT (15)		//	TRUE => does not become active on enter

#define	OP_CTRL_BROADHILITE		OP_BIT (16)		//	TRUE => broadcast the hilite

/****************************************************************/

/*
 *	The Control class is a somewhat specialized view
 *	which is used as a base class by many widgets.
 */

class OP_PUBLIC OPControl : public OPView
{
	
protected:
	OPAppRef		app;						//	app used to "tick"
	Bool			is_app_ticking;				//	TRUE => app is ticking
	
protected:
	OPCtrlShape		shape;						//	shape attributes
	Card32			sensitive_buttons;			//	buttons to which this control is sensitive
	OPColorPixel	relief_color;				//	base relief color
	
	Bool			do_hilite_on_action;		//	TRUE => hilite when user clicks / activates
	Bool			do_press_on_action;			//	TRUE => press when user clicks / activates
	Bool			do_cont_reports_on_action;	//	TRUE => produce continuous reports on action
	
	Bool			is_being_hilited;			//	TRUE => control is being highlighted
	Bool			is_actively_hilited;		//	TRUE => control is actively highlighted
	Bool			is_being_pressed;			//	TRUE => control is being pressed
	Bool			is_being_clicked;			//	TRUE => control is being clicked
	Bool			is_being_hovered;			//	TRUE => control is being hovered
	
	Bool			key_alt_known;				//	TRUE => window knows about keyboard alternative
	
	OPKey			key_alt;					//	keyboard alternative (if any)
	OPStringRef		tool_tip_caption;			//	text to be displayed in tool tip (if any)
	OPToolTipRef	tool_tip;					//	associated tool tip (if any)
	
	OPCoord			min_dx;						//	minimal width
	OPCoord			min_dy;						//	minimal height
	
protected:
	OPControl ();
	virtual ~ OPControl ();
	
	virtual void StoreInstance (OPStorageIn storage);		/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);		/* OVERRIDDEN */

	virtual void DisableTick ();
	virtual void EnableTick ();
	
	//	Subclasses will have to override the following methods in order to get
	//	some specific behaviour :
	
	virtual void ActionOn ();
	virtual void ActionOff ();
	virtual void Action (Bool clicked);
	
	virtual void HoverOn ();
	virtual void HoverOff ();

	//	The following methods may be overridden to change the basic display
	//	behaviour

	virtual void DrawActiveFrame (OPGrailPortIn, OPAdornerIn, OPCoord margin);
	
	//	The following methods get called by the "Action" method or some other
	//	internal method, that's why they have not been defined as public.
	
	virtual void SetHiliteState (Bool ok, Bool broadcast = TRUE);
	virtual void SetPressState (Bool ok, Bool broadcast = TRUE);
	
public:
	virtual void Delete ();									/* OVERRIDDEN */
	virtual Bool SetWindow (OPWindowIn window);				/* OVERRIDDEN */
	
	virtual Bool SetShapeStyle (OPCtrlShape shape);
	virtual Bool SetHiliteMode (Bool yes);
	virtual Bool SetPressMode (Bool yes);
	virtual Bool SetReportMode (Bool is_continuous); 
	virtual Bool SetSensitiveButtons (Card32 buttons);
	virtual Bool SetActiveHilite (Bool yes, Bool broadcast = TRUE);
	virtual Bool SetReliefColor (OPColorPixelIn color);
	virtual Bool SetToolTip (OPStringIn tip);
	
	virtual OPCtrlShape ReturnShapeStyle () const;
	virtual Bool   ReturnHiliteMode () const;
	virtual Bool   ReturnPressMode () const;
	virtual Bool   ReturnReportMode () const;
	virtual Card32 ReturnSensitiveButtons () const;
	virtual Bool   ReturnActiveHilite () const;
	virtual OPColorPixel ReturnReliefColor () const;
	virtual OPStringRet ReturnToolTip () const;
	
	virtual Bool SetDesiredMinSize (OPCoord dx, OPCoord dy);
	virtual Bool GetDesiredMinSize (OPCoordOut dx, OPCoordOut dy) const;
	
	//	The control may handle keyboard alternatives :
	
	virtual Bool SetKeyAlt (OPKeyIn key_alt);
	virtual Bool ClearKeyAlt ();
	
	virtual Bool GetKeyAlt (OPKeyOut key_alt) const;
	virtual Bool HasKeyAlt () const;
	virtual OPKey ReturnKeyAlt () const;

	virtual Bool HandleKeyAlt (OPEventIn event);
	virtual Bool ActionOnKeyAlt ();
	
	//	Inherited methods :
	
	virtual void GetMinSize (OPCoordOut, OPCoordOut) const;	/* OVERRIDDEN */
	virtual void GetMaxSize (OPCoordOut, OPCoordOut) const;	/* OVERRIDDEN */
	
	virtual void Broadcast (OPAtomIn m, OPObjectIn value);				/* OVERRIDDEN */
	virtual void Draw (OPGrailPortIn, OPRectIn, OPDrawOp = OP_DRAW_ALL);/* OVERRIDDEN */
	
	virtual void ButtonPress (OPEventIn event);				/* OVERRIDDEN */
	virtual void ButtonRelease (OPEventIn event);			/* OVERRIDDEN */
	virtual void EnterNotify (OPEventIn event);				/* OVERRIDDEN */
	virtual void LeaveNotify (OPEventIn event);				/* OVERRIDDEN */
	virtual void KeyDown (OPEventIn event);					/* OVERRIDDEN */
	virtual void KeyUp (OPEventIn event);					/* OVERRIDDEN */
	
	virtual void Tick ();									/* OVERRIDDEN */
	
	OP_DEFINE_SLOT (Bool, hilite_mode,   SetHiliteMode,   ReturnHiliteMode);
	OP_DEFINE_SLOT (Bool, press_mode,    SetPressMode,    ReturnPressMode);
	OP_DEFINE_SLOT (Bool, report_mode,   SetReportMode,   ReturnReportMode);
	OP_DEFINE_SLOT (Bool, active_hilite, SetActiveHilite, ReturnActiveHilite);
	OP_DEFINE_SLOT (Card32, shape_style, SetShapeStyle,   ReturnShapeStyle);
	OP_DEFINE_SLOT (Card32, sensitive_buttons, SetSensitiveButtons, ReturnSensitiveButtons);
	OP_DEFINE_SLOT (OPColorPixel, relief_color, SetReliefColor, ReturnReliefColor);
	OP_DEFINE_SLOT (OPString, tool_tip, SetToolTip, ReturnToolTip);
	OP_DEFINE_CLASS (OPControl, OPView);
};

/****************************************************************/

#endif /* _OPaC_CONTROL_H_ */
