/*
 *	except.h
 *
 *	This file defines the the classes used by OPaC to describe
 *	exception conditions. This is now based on the ANSI C++ model.
 *	
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_EXCEPT_H_
#define	_OPaC_EXCEPT_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

#ifndef _OPaC_ATOM_H_
#include <OPaC/atom.h>
#endif

class OPException;
class OPExcMemory;
class OPExcObject;
class OPExcTarget;
class OPExcSource;
class OPExcString;

/****************************************************************/

/*
 *	This macro allows to check the conformity of an object's
 *	class at run-time. It will throw an exception on mismatch.
 */

#define	OP_CLASS_CHECK(x,c)										\
	do {														\
		const OPObjectRef& __o = x;								\
		if (__o.IsInvalid ())									\
			throw OPExcObject (OPExcObject::NOT_INITIALISED);	\
		if (! __o->IsKindOf (OP_RUNTIME (c)))					\
			throw OPExcObject (OPExcObject::MISMATCH);			\
	} while (0)

#define OP_STRICT_CLASS_CHECK(x,c)								\
	do {														\
		const OPObjectRef& __o = x;								\
		if (__o.IsInvalid ())									\
			throw OPExcObject (OPExcObject::NOT_INITIALISED);	\
		if (! __o->IsA (OP_RUNTIME (c)))						\
			throw OPExcObject (OPExcObject::MISMATCH);			\
	} while (0)

/****************************************************************/

/*
 *	Base class for exception processing. Every exception
 *	class inherits its properties.
 */

class OP_PUBLIC OPException
{
	
protected:
	OPaC_CallFrame		throw_frame;			//	frame where throw occurred
	
public:
	OPException ();
	
	void* operator new (size_t);
	
	const OPaC_CallFrame* ReturnThrowFrame () const;
	Count ReturnThrowSrcLine () const;
	const LatinChar* ReturnThrowSrcFile () const;
};

/****************************************************************/

/*
 *	Memory exception class. A more detailed information
 *	about the memory exception is stored in each exception
 *	instance.
 */

class OP_PUBLIC OPExcMemory : public OPException
{
	
public:
	enum Error {
		FULL,									//	memory full (exhausted)
		BUFFER_TOO_SMALL,						//	specified buffer is to small
		NULL_PTR,								//	access to null pointer
		BAD_PTR,								//	access to bad pointer
		FREE									//	access to freed memory
	};
	
protected:
	Error				error;					//	error code
	void*				ptr;					//	faulty pointer (if BAD_PTR)
	
	Count				caller_line;			//	line # where call occurred
	const LatinChar*	caller_file;			//	file name where call occurred
	
public:
	OPExcMemory (Error e, void* p, Count l, const LatinChar* f)
		: error (e), ptr (p), caller_line (l), caller_file (f) { }
	
	Error ReturnError () const { return this->error; }
	void* ReturnFaultyPtr () const { return this->ptr; }
	Count ReturnFaultySrcLine () const { return this->caller_line; }
	const LatinChar* ReturnFaultySrcFile () const { return this->caller_file; }
};

/****************************************************************/

/*
 *	Object exception class. More information is stored in
 *	the exception instance.
 */

class OP_PUBLIC OPExcObject : public OPException
{
	
public:
	enum Error {
		MISMATCH,								//	object does not match requirements
		NOT_INITIALISED,						//	object has not been initialised
		NOT_FOUND								//	class has not been found
	};
	
protected:
	Error				error;					//	error code
	const OPObject*		object;					//	pointer to faulty object
	
public:
	OPExcObject (Error e, const OPObject* o = 0)
		: error (e), object (o) { }
	
	Error ReturnError () const { return this->error; }
	const OPObject* ReturnFaultyObject () const { return this->object; }
};

/****************************************************************/

/*
 *	Target exception class. An access to an invalid target
 *	has been attempted.
 */

class OP_PUBLIC OPExcTarget : public OPException
{
	
protected:
	OPAtomRef			atom;					//	target description
	
public:
	OPExcTarget () : atom ((OPAtom*)(0)) { }
	OPExcTarget (OPAtomIn a) : atom (a) { }
	const OPAtomRef& ReturnFaultyTarget () const { return this->atom; }
};

/****************************************************************/

/*
 *	Outlet exception class. An access to an invalid outlet
 *	has been attempted.
 */

class OP_PUBLIC OPExcOutlet : public OPException
{
	
protected:
	OPAtomRef			atom;					//	outlet description
	
public:
	OPExcOutlet () : atom ((OPAtom*)(0)) { }
	OPExcOutlet (OPAtomIn a) : atom (a) { }
	const OPAtomRef& ReturnFaultyOutlet () const { return this->atom; }
};

/****************************************************************/

/*
 *	General access error exception class. An access to an
 *	invalid element in an array has been attempted.
 */

class OP_PUBLIC OPExcAccess : public OPException
{
	
public:
	enum Error {
		INVALID_ACCESS,							//	access declared to be invalid
		OUT_OF_RANGE							//	access is out of range
	};
	
protected:
	Error				error;					//	error description
	
public:
	OPExcAccess (Error e) : error (e) { }
	Error ReturnError () const { return this->error; }
};

/****************************************************************/

/*
 *	The string exceptions are raised when some invalid
 *	string access takes place (eg. "out of range").
 */

class OP_PUBLIC OPExcString : public OPException
{
	
public:
	enum Error {
		OUT_OF_RANGE							//	access is out of range
	};
	
protected:
	Error				error;					//	error description
public:
	OPExcString (Error e) : error (e) { }
	Error ReturnError () const { return this->error; }
};

/****************************************************************/

#endif
