/*
 *	rbtree.h
 *
 *	Red-Black Tree Class interface definition. This class provides a fast way
 *	to insert and retrieve sorted information from a well balanced tree.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_RBTREE_H_
#define	_OPaC_RBTREE_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

/****************************************************************/

/*
 *	A Red-Black Tree consists of 2-3-4 nodes represented
 *	with simple 2-pointer nodes. The algorithms are being
 *	presented in Sedgewick's "Algorithms in C++" book.
 */

class OP_PUBLIC OPRBTree : public OPObject
{
	
protected:
	struct OP_PUBLIC Node
	{
		Node*			left;					//	left child node
		Node*			right;					//	right child node
		
		Bool			is_red;					//	TRUE => link with parent node is red
		Bool			is_valid;				//	TRUE => this node is valid
		
		Card32			key;					//	key used for the search
		OPObjectRef		value;					//	associated value (reference to object)
	};
	
private:
	Node*				head;					//	link to head node (empty node)
	Node*				zero;					//	link to zero node (conveniance NULL)
	
	Node				head_node;				//	memory for the dummy head node
	Node				zero_node;				//	memory for the dummy zero node
	
protected:
	OPRBTree ();
	virtual ~ OPRBTree ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	//	The allocation methods to manage the nodes may be overridden :
	
	virtual Node* AllocNode ();
	virtual void  FreeNode (Node* node);
	
	void FreeSubTree (Node* node);
	void StoreSubTree (Node* node, OPStorageIn storage);
	void RestoreSubTree (Node* node, OPStorageIn storage);
	
	//	The accessor methods have no reason to be overridden, so they have not
	//	been defined as being "virtual" :
	
	void  InsertNode (Node* node);
	Node* RemoveNode (Card32 index);
	
	Node* FindEQ (Card32 index);
	Node* FindGE (Card32 index);
	Node* FindLE (Card32 index);
	
	Node* FindMin () const;
	Node* FindMax () const;
	
public:
	virtual void Delete ();						/* OVERRIDDEN */
	
	virtual void Insert (Card32 index, OPObjectIn value);
	
	virtual Bool Remove (Card32 index);
	virtual Bool Remove (Card32 index, OPObjectOut value);
	
	virtual Bool Find (Card32 index);
	virtual Bool Find (Card32 index, OPObjectOut value);
	
	virtual Bool FindNearestPrev (Card32 index);
	virtual Bool FindNearestPrev (Card32 index, OPObjectOut value);
	
	virtual Bool FindNearestNext (Card32 index);
	virtual Bool FindNearestNext (Card32 index, OPObjectOut value);
	
	virtual Bool FindFirst (Card32& index) const;
	virtual Bool FindLast (Card32& index) const;

	OP_DEFINE_CLASS (OPRBTree, OPObject);
};

#endif /* _OPaC_RBTREE_H_ */
