/*
 *	stream.h
 *
 *	The stream class provides a unified interface enabling the access to files
 *	and interprocess communication using links.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_STREAM_H_
#define	_OPaC_STREAM_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

/****************************************************************/

#ifndef _OPaC_EXCEPT_H_
#include <OPaC/except.h>
#endif

#ifndef _OPaC_FILE_IO_H_
#include <OPaC/sd/file_io.h>
#endif

/****************************************************************/

/*
 *	The Stream class is just an abstract class which provides
 *	a canvas for the real streams.
 */

class OP_PUBLIC OPStream : public OPObject
{
protected:
	OPStringRef			name;					//	name of stream (if any)
	
protected:
	OPStream ();
	virtual ~ OPStream ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void SetStreamName (OPStringIn name);
	
public:
	virtual Count Read (void*, Count);
	virtual Int64 Skip (Int64);
	virtual Count Write (const void*, Count);
	
	virtual Bool Sync ();
	
	OP_DEFINE_CLASS (OPStream, OPObject);
};

/****************************************************************/

/*
 *	The File Stream class provides the base for a re-openable
 *	file read and file write class.
 */

class OP_PUBLIC OPFileStream : public OPStream
{
	
protected:
	Int64				index;					//	index into file
	Bool				is_open;				//	TRUE => stream is open
	
protected:
	OPFileStream ();
	virtual ~ OPFileStream ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void Open () = 0;
	virtual void Close () = 0;
	
	virtual void Awake ();								/* OVERRIDDEN */
	
	virtual Bool SetName (OPStringIn name);
	virtual OPStringRet ReturnName () const;
	virtual Bool GetName (OPStringOut name) const;
	
	OP_DEFINE_CLASS (OPFileStream, OPStream);
};

/****************************************************************/

/*
 *	The File Write Stream class is able to write data
 *	sequentially into a file.
 */

class OP_PUBLIC OPFileWrite : public OPFileStream
{
	
protected:
	OPaC_File			file;				//	file stream to write to (or 0)
	Bool				is_append_mode;		//	TRUE => append to end of file
	
protected:
	OPFileWrite ();
	virtual ~ OPFileWrite ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void Open ();								/* OVERRIDDEN */
	virtual void Close ();								/* OVERRIDDEN */
	
	virtual void Open (OPStringIn name, Bool append = FALSE);
	
	virtual Count Write (const void*, Count);			/* OVERRIDDEN */
	virtual Int64 Skip (Int64);							/* OVERRIDDEN */
	virtual Bool Sync ();								/* OVERRIDDEN */
	virtual void Trunc ();
	
	virtual Bool SetAppendMode (Bool on);
	virtual Bool ReturnAppendMode () const;
	
	OP_DEFINE_CLASS (OPFileWrite, OPFileStream);
};

/****************************************************************/

/*
 *	The File Read Stream class is able to read data
 *	sequentially from a file. It will generate an
 *	exception after the end of file.
 */

class OP_PUBLIC OPFileRead : public OPFileStream
{
	
protected:
	OPaC_File			file;				//	file stream to write to (or 0)
	
protected:
	OPFileRead ();
	virtual ~ OPFileRead ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void Open ();								/* OVERRIDDEN */
	virtual void Close ();								/* OVERRIDDEN */
	
	virtual void Open (OPStringIn name);
	
	virtual Count Read (void*, Count);					/* OVERRIDDEN */
	virtual Int64 Skip (Int64);							/* OVERRIDDEN */
	
	OP_DEFINE_CLASS (OPFileRead, OPFileStream);
};

/****************************************************************/

/*
 *	Exception class associated to the stream operations. The
 *	stream methods may throw such exceptions.
 */

class OP_PUBLIC OPExcStream : public OPException
{

public:
	enum Error {
		NO_WRITE,							//	stream does not accept any write
		NO_READ,							//	stream does not accept any read
		NO_SKIP,							//	stream does not accept skipping
		WRITE_ERROR,						//	an error occurred while writing data
		READ_ERROR,							//	an error occurred while reading data
		ALREADY_OPENED,						//	stream has already been open
		ALREADY_CLOSED,						//	stream has already been closed
		UNKNOWN_NAME,						//	unknown name has been specified
		TIME_OUT,							//	time out, no answer gotten
		CANNOT_OPEN							//	could not open the stream
	};
	
protected:
	Error				error;				// error code
	
public:
	OPExcStream (Error e) : error (e) { }
	virtual Error GetError () const { return error; }
};

#endif /* _OPaC_STREAM_H_ */
