/*
 *	textline.h
 *
 *	The Text Line is used to represent an editable line of text. It handles
 *	cursor blinking and highlighting.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_TEXTLINE_H_
#define	_OPaC_TEXTLINE_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

#ifndef _OPaC_CONTROL_H_
#include <OPaC/control.h>
#endif

#ifndef _OPaC_STRING_H_
#include <OPaC/string.h>
#endif

/****************************************************************/

/*
 *	Text Line class.
 */

class OP_PUBLIC OPTextLine : public OPControl
{
	
	static OPCoord		look_3D_effect;			//	size of 3D effect
	static OPCoord		look_frame_size;		//	size of frame around widgets
	static Count		look_cursor_off;		//	time while cursor off
	static Count		look_cursor_on;			//	time while cursor on
	
private:
	OPStringRef			text_string;			//	the text is stored in a string
	
	Card32				cursor_pos;				//	cursor position in text
	Card32				hilite_pos_1;			//	highlighting beginning position
	Card32				hilite_pos_2;			//	highlighting ending position
	Card32				temp_pos;				//	temporary position after highlighting
	
	OPCoord				text_dx;				//	text width
	OPCoord				cursor_x;				//	cursor position in view (horiz. offset)
	OPCoord				hilite_x_1;				//	highlighting beginning horizontal offset
	OPCoord				hilite_x_2;				//	highlighting ending horizontal offset
	OPCoord				start_x;				//	offset for the display
	OPRelCoord			cursor_ascender;		//	cursor height (over base line, positive)
	OPRelCoord			cursor_descender;		//	cursor height (below base line, negative)
	
	Bool				cursor_blink;			//	TRUE => cursor is being displayed
	Bool				cursor_enable;			//	TRUE => cursor is being enabled
	Bool				is_dragging;			//	TRUE => user currently dragging in text line
	Bool				is_frameless;			//	TRUE => text has no frame at all
	Bool				is_transparent;			//	TRUE => text has no specific background
	
protected:
	OPTextLine ();
	virtual ~ OPTextLine ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void Action (Bool clicked);					/* OVERRIDDEN */
	
	virtual void NotifyTextChange (Bool user);
	virtual void UpdateCursorGeometry ();
	virtual void RedrawCursor ();
	
	virtual OPGrailFontRet ReturnFont () const;
	virtual OPCoord ReturnTextIntervalWidth (Card32 pos_1, Card32 pos_2) const;
	virtual OPRelCoord ReturnTextAscender () const;
	virtual OPRelCoord ReturnTextDescender () const;
	virtual Card32 ComputePos (OPEventIn event) const;
	
	virtual OPCoord ReturnFrameSize () const;
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual void Awake ();								/* OVERRIDDEN */
	virtual void AwakeFromApp (OPAppIn app);			/* OVERRIDDEN */
	virtual void AwakeFromObject (OPObjectIn object);	/* OVERRIDDEN */
	
	virtual Bool SetText (OPStringIn text);
	virtual Bool GetText (OPStringOut text) const;
	virtual OPStringRet ReturnText () const;
	
	virtual void Clear ();
	virtual void SetFullHilite ();
	virtual void SetEmptyHilite ();
	
	virtual Bool SetCursorPos (Card32 pos);
	virtual Bool SetHiliteRange (Card32 pos_1, Card32 pos_2);
	virtual Bool SetFramelessMode (Bool yes);
	virtual Bool SetTransparentMode (Bool yes);
	
	virtual Card32 ReturnCursorPos () const;
	virtual Card32 ReturnHiliteBegin () const;
	virtual Card32 ReturnHiliteEnd () const;
	
	virtual Bool ReturnFramelessMode () const;
	virtual Bool ReturnTransparentMode () const;
	
	virtual Bool IsHilited () const;
	
	//	Inherited methods
	
	virtual Bool WouldAcceptMain () const;				/* OVERRIDDEN */
	virtual void BecomeMain (Bool clicked);				/* OVERRIDDEN */
	virtual void ResignMain ();							/* OVERRIDDEN */
	
	virtual void GetMinSize (OPCoordOut dx, OPCoordOut dy) const;/* OVERRIDDEN */
	virtual void Broadcast (OPAtomIn, OPObjectIn);		/* OVERRIDDEN */
	virtual void Draw (OPGrailPortIn, OPRectIn, OPDrawOp = OP_DRAW_ALL);/* OVERRIDDEN */
	
	virtual void ButtonPress (OPEventIn event);			/* OVERRIDDEN */
	virtual void ButtonRelease (OPEventIn event);		/* OVERRIDDEN */
	virtual void DragNotify (OPEventIn event);			/* OVERRIDDEN */
	virtual void KeyDown (OPEventIn event);				/* OVERRIDDEN */
	
	virtual void Tick ();								/* OVERRIDDEN */
	
	OP_DEFINE_SLOT (Card32, cursor_pos, SetCursorPos, ReturnCursorPos);
	OP_DEFINE_SLOT_READ (Bool, is_hilited, IsHilited);
	OP_DEFINE_SLOT (OPString, text, SetText, ReturnText);
	OP_DEFINE_SLOT (Bool, is_frameless, SetFramelessMode, ReturnFramelessMode);
	OP_DEFINE_SLOT (Bool, is_transparent, SetTransparentMode, ReturnTransparentMode);
	OP_DEFINE_CLASS (OPTextLine, OPControl);
};

#endif /* _OPaC_TEXTLINE_H_ */

