/*
 *	sd_grail.cxx
 *
 *	This file implements the interface to the graphical API. It
 *	relies on the platform independent "Grafix" library.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_grafix.h"

/****************************************************************/

#include <math.h>

#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_MORE_IMPLEMENTATION (OPGrailPort);

/****************************************************************/

#define	BUFFER_LENGTH		400
#define	MAX_LINE_LENGTH		2000

static LatinChar g_buffer_array[BUFFER_LENGTH];
static Card32    g_pos_array[MAX_LINE_LENGTH];
static Card8     g_err_array[MAX_LINE_LENGTH];

/****************************************************************/

/*
 *	Clean up a Grail Port. This will destroy everything
 *	that has been constructed by the constructor (logical,
 *	huh ?)
 */

OPGrailPort::~OPGrailPort ()
{
	OP_METHOD (~OPGrailPort, 0);
	
	if (this->surface) {
		GfxSurface::Delete (this->surface);
		this->surface = 0;
	}
}

/*
 *	Define a clipping rectangle, which will truncate every
 *	graphical operation. The new clipping region replaces
 *	the previous one.
 */

void
OPGrailPort::SetClipRect (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (SetClipRect, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		this->surface->SetClipRect (0, 0, 0, 0);
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	this->SetClipRect (x1, y1, x2, y2);
}

void
OPGrailPort::SetClipRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (SetClipRect, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->is_drawing);
	
	this->clip_x1 = (x1 > this->rect_x1) ? x1 : this->rect_x1;
	this->clip_y1 = (y1 > this->rect_y1) ? y1 : this->rect_y1;
	this->clip_x2 = (x2 < this->rect_x2) ? x2 : this->rect_x2;
	this->clip_y2 = (y2 < this->rect_y2) ? y2 : this->rect_y2;
	
	this->surface->SetClipRect ((int)(x1), (int)(y1), (int)(x2-x1+1), (int)(y2-y1+1));
}

/*
 *	Reset the clipping rectangle to the default one.
 */

void
OPGrailPort::ResetClipRect ()
{
	OP_METHOD (ResetClipRect, 0);
	
	this->clip_x1 = this->rect_x1;
	this->clip_y1 = this->rect_y1;
	this->clip_x2 = this->rect_x2;
	this->clip_y2 = this->rect_y2;
	
	int x1 = (int)this->clip_x1;
	int y1 = (int)this->clip_y1;
	int x2 = (int)this->clip_x2;
	int y2 = (int)this->clip_y2;
	
	this->surface->SetClipRect (x1, y1, x2-x1+1, y2-y1+1);
}




/****************************************************************
 *																*
 *																*
 *			GRAIL PORT CLASS : TARGET SPECIFIC METHODS			*
 *																*
 *																*
 ****************************************************************/


/*
 *	Draw just one pixel.
 */

void
OPGrailPort::DrawDot (OPCoord x, OPCoord y)
{
	OP_METHOD (DrawDot, 2);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ASSERT (this->surface);
	
	this->surface->SetPixelRGB ((int)x, (int)y, this->fore_color);
}

void
OPGrailPort::DrawDot (OPCoord x, OPCoord y, OPColorPixelIn color)
{
	OP_METHOD (DrawDot, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPColorPixelIn, color);
	OP_ASSERT (this->surface);
	
	this->surface->SetPixelRGB ((int)x, (int)y, color);
}


/*
 *	Draw a line (segment), using one of the predefined dash
 *	styles. The segment is always hair thin... The last dot
 *	will be drawn.
 *
 *	A line length of 0 won't draw anything !
 */

void
OPGrailPort::DrawLineDX (OPCoord x, OPCoord y, OPCoord dx)
{
	OP_METHOD (DrawLineDX, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	if (dx != 0) {
		this->surface->DrawLine ((int)(x), (int)(y), (int)(x+dx-1), (int)(y),
								 this->fore_color,
								 this->dash_array, this->dash_num);
	}
}

void
OPGrailPort::DrawLineDY (OPCoord x, OPCoord y, OPCoord dy)
{
	OP_METHOD (DrawLineDY, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dy);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	if (dy != 0) {
		this->surface->DrawLine ((int)(x), (int)(y), (int)(x), (int)(y+dy-1),
								 this->fore_color,
								 this->dash_array, this->dash_num);
	}
}



/*
 *	Draw a line (segment), using one of the predefined dash
 *	styles. The segment is always hair thin... The last dot
 *	will be drawn.
 *
 *	A line with both extremities at the same location will
 *	draw just a single dot.
 */

void
OPGrailPort::DrawLineToX (OPCoord x1, OPCoord y1, OPCoord x2)
{
	OP_METHOD (DrawLineToX, 3);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	if (x2 >= x1) {
		this->surface->DrawLine ((int)(x1), (int)(y1), (int)(x2), (int)(y1),
								 this->fore_color,
								 this->dash_array, this->dash_num);
	} else {
		this->surface->DrawLine ((int)(x1), (int)(y1), (int)(x2), (int)(y1),
								 this->fore_color,
								 this->dash_array, this->dash_num);
	}
}

void
OPGrailPort::DrawLineToY (OPCoord x1, OPCoord y1, OPCoord y2)
{
	OP_METHOD (DrawLineToY, 3);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, y2);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	if (y2 >= y1) {
		this->surface->DrawLine ((int)(x1), (int)(y1), (int)(x1), (int)(y2),
								 this->fore_color,
								 this->dash_array, this->dash_num);
	} else {
		this->surface->DrawLine ((int)(x1), (int)(y1), (int)(x1), (int)(y2),
								 this->fore_color,
								 this->dash_array, this->dash_num);
	}
}


/*
 *	Draw rectangular box. The bottom left corner will be
 *	drawn at [x2;y2]. This might not be the case on every
 *	system, so be careful to implent this the OPaC way.
 */

void
OPGrailPort::DrawRect (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (DrawRect, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	this->DrawRect (x1, y1, x2, y2);
}

void
OPGrailPort::DrawRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (DrawRect, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	this->surface->DrawLine ((int)x1, (int)y1, (int)x2-1, (int)y1, this->fore_color, this->dash_array, this->dash_num);
	this->surface->DrawLine ((int)x2, (int)y1, (int)x2, (int)y2-1, this->fore_color, this->dash_array, this->dash_num);
	this->surface->DrawLine ((int)x2, (int)y2, (int)x1+1, (int)y2, this->fore_color, this->dash_array, this->dash_num);
	this->surface->DrawLine ((int)x1, (int)y2, (int)x1, (int)y1+1, this->fore_color, this->dash_array, this->dash_num);
}



/*
 *	Fill a rectangular box. The surface drawn is exactly
 *	the same as with DrawRect. Beware, this is usually not
 *	the case on other systems !!!
 *
 *	BE CAREFUL TO IMPLEMENT THIS METHOD THE RIGHT WAY !
 */

void
OPGrailPort::FillRect (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (FillRect, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	Card32 x1 = (Card32) (box.X1 () + margin /* + 0.5 */);
	Card32 y1 = (Card32) (box.Y1 () + margin /* + 0.5 */);
	Card32 x2 = (Card32) (box.X2 () - margin /* + 0.5 */);
	Card32 y2 = (Card32) (box.Y2 () - margin /* + 0.5 */);
	
	if (this->fore_color.Transparency ()) {
		this->surface->RectFillBlend (x1, y1, x2-x1+1, y2-y1+1, this->fore_color);
	} else {
		this->surface->RectFill (x1, y1, x2-x1+1, y2-y1+1, this->fore_color);
	}
}

void
OPGrailPort::FillRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (FillRect, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	if (this->fore_color.Transparency ()) {
		this->surface->RectFillBlend ((int)(x1), (int)(y1), (int)(x2-x1+1), (int)(y2-y1+1), this->fore_color);
	} else {
		this->surface->RectFill ((int)(x1), (int)(y1), (int)(x2-x1+1), (int)(y2-y1+1), this->fore_color);
	}
}


/*
 *	Draw a line segment. The segment is always hair thin and
 *	is not dashed... The last dot will be drawn.
 */

void
OPGrailPort::DrawHairLine (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy)
{
	OP_METHOD (DrawHairLine, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, OPCoord, dy);
	
	this->DrawHairLineTo (x, y, x+dx, y+dy);
}


void
OPGrailPort::DrawHairLineTo (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (DrawHairLineTo, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->surface);
	
	this->surface->DrawLine ((int)x1, (int)y1, (int)x2, (int)y2, this->fore_color, this->dash_array, this->dash_num);
}


/*
 *	Draw a line segment with anti-aliasing. The segment is always
 *	hair thin and is not dashed... The last dot will be drawn.
 */

void
OPGrailPort::DrawAALine (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy)
{
	OP_METHOD (DrawAALine, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, OPCoord, dy);
	
	this->DrawAALineTo (x, y, x+dx, y+dy);
}

void
OPGrailPort::DrawAALineTo (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (DrawAALineTo, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->surface);
	
	this->surface->DrawAALine ((int)x1, (int)y1, (int)x2, (int)y2, this->fore_color);
}


/*
 *	Draw a piece of text, which will be extracted from a
 *	string starting at character "start" and up to "length"
 *	characters. The way to position the text is specified
 *	by "pos" and "dx" is used if aligned positioning has
 *	been requested.
 */

void
OPGrailPort::ShowText (OPCoord x, OPCoord y, OPStringIn text,
					   Count start, Count length,
					   Card8 pos, OPCoord dx)
{
	OP_METHOD (ShowText, 7);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPStringIn, text);
	OP_ARG (4, Count, start);
	OP_ARG (5, Count, length);
	OP_ARG (6, Card8, pos);
	OP_ARG (7, OPCoord, dx);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	if (this->font.IsInvalid ()) {
		return;
	}
	
	Count len = text->CountCharacters ();
	
	if (start >= len) {
		
		//	Do not draw anything, since the start of the text in the string is
		//	past the end of the string.
		
		return;
	}
	
	if ( (len < (start + length))
	  || (length == 0) ) {
		
		//	User is asking for a too long piece of text, longer than it really
		//	can be.
		
		length = len - start;
	}
	
	LatinChar* buffer_alloc = 0;
	LatinChar* buffer = 0;
	
	if (length < BUFFER_LENGTH) {
		
		//	Use the buffer from the stack.
		
		buffer = g_buffer_array;
		
	} else {
		
		//	Use the buffer from the temporary memory.
		
		buffer = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_temp_zone, LatinChar, length);
		buffer_alloc = buffer;
	}
	
	text->ExtractStringCopy (buffer, start, length);
	
	int hi = (pos & OP_TEXT_HPOS_MASK);
	int vi = (pos & OP_TEXT_VPOS_MASK);
	
	OPCoord text_w = this->font->ReturnWidth (buffer);
	OPCoord text_a = this->font->ReturnAscender ();
	OPCoord text_d = this->font->ReturnDescender ();
	
	if (hi == OP_TEXT_HPOS_CENTER) x -= text_w / 2;
	if (hi == OP_TEXT_HPOS_RIGHT)  x -= text_w;
	
	if (vi == OP_TEXT_VPOS_BASE)   y -= text_a;
	if (vi == OP_TEXT_VPOS_CENTER) y -= (text_a - text_d) / 2;
	if (vi == OP_TEXT_VPOS_BOTTOM) y -= text_a - text_d;
	
	y += text_a;
	
	for (Count index = 0; index < length; index++) {
		GfxFontCache::GlyphGeom geom;
		this->font->font->GetGlyphGeometry ((Card8)buffer[index], geom);
		this->font->font->DrawGlyph ((Card8)buffer[index], (int)x, (int)y, this->fore_color, this->surface);
		x += geom.x_advance;
		y += geom.y_advance;
	}
	
	if (buffer_alloc) {
		
		//	Free the memory if we need to.
		
		OP_FREE_FROM_ZONE (OPaC_temp_zone, buffer_alloc);
	}
}



/*
 *	Draw a piece of text in the specified rectangle. This
 *	uses the moreShowText
 */

void
OPGrailPort::ShowText (OPRectIn box, OPStringIn text,
					   Count start, Count length, Card8 pos, OPRelCoord margin)
{
	OP_METHOD (ShowText, 6);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPStringIn, text);
	OP_ARG (3, Count, start);
	OP_ARG (4, Count, length);
	OP_ARG (5, Card8, pos);
	OP_ARG (6, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	this->ShowText (x1, y1, x2, y2, text, start, length, pos);
}

void
OPGrailPort::ShowText (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2,
					   OPStringIn text,
					   Count start, Count length,
					   Card8 pos)
{
	OP_METHOD (ShowText, 8);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ARG (5, OPStringIn, text);
	OP_ARG (6, Count, start);
	OP_ARG (7, Count, length);
	OP_ARG (8, Card8, pos);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( (x2 <= x1) || (y2 <= y1) ) {
		return;
	}
	
	OPCoord x = x1;
	OPCoord y = y1;
	
	switch (pos & OP_TEXT_VPOS_MASK) {
		
		case OP_TEXT_VPOS_TOP:
			y = y1;
			break;
		
		case OP_TEXT_VPOS_CENTER:
			y = (y1 + y2 + 1) / 2;
			break;
			
		case OP_TEXT_VPOS_BOTTOM:
			y = y2;
			break;
	}
	
	switch (pos & OP_TEXT_HPOS_MASK) {
		
		case OP_TEXT_HPOS_LEFT:
			x = x1;
			break;
		
		case OP_TEXT_HPOS_CENTER:
			x = (x1 + x2 + 1) / 2;
			break;
		
		case OP_TEXT_HPOS_ALIGN:
			x = x1;
			break;
		
		case OP_TEXT_HPOS_RIGHT:
			x = x2;
			break;
	}
	
	this->ShowText (x, y, text, start, length, pos, x2-x1+1);
}



/*
 *	Scroll a rectangular area of the Grail Port. This will
 *	not erase the invalidated parts of the picture.
 */

void
OPGrailPort::ScrollRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2,
						 OPRelCoord move_x, OPRelCoord move_y)
{
	OP_METHOD (ScrollRect, 6);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ARG (5, OPRelCoord, move_x);
	OP_ARG (6, OPRelCoord, move_y);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	this->surface->RectCopy ((int)(x1+move_x), (int)(y1+move_y),
							 (int)(x2-x1+1), (int)(y2-y1+1), (int)x1, (int)y1);
}

void
OPGrailPort::ScrollRect (OPRectIn box, OPRelCoord move_x, OPRelCoord move_y,
						 OPRelCoord margin)
{
	OP_METHOD (ScrollRect, 4);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, move_x);
	OP_ARG (3, OPRelCoord, move_y);
	OP_ARG (4, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	this->ScrollRect (x1, y1, x2, y2, move_x, move_y);
}



/*
 *	Copy a rectangle from the source port to the current
 *	port, into the specified box, from the source position
 *	given by its top/left corner.
 *
 *	NB: This will copy the pixels, including the one of
 *		the bottom right corner. This is still the same
 *		policy as with the rectangle.
 */

void
OPGrailPort::CopyRect (OPGrailPortIn source,
					   OPCoord source_x, OPCoord source_y,
					   OPCoord dest_x1, OPCoord dest_y1,
					   OPCoord dest_x2, OPCoord dest_y2)
{
	OP_METHOD (CopyRect, 7);
	OP_ARG (1, OPGrailPortIn, source);
	OP_ARG (2, OPCoord, source_x);
	OP_ARG (3, OPCoord, source_y);
	OP_ARG (4, OPCoord, dest_x1);
	OP_ARG (5, OPCoord, dest_y1);
	OP_ARG (6, OPCoord, dest_x2);
	OP_ARG (7, OPCoord, dest_y2);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	OPGrailPort* src = source.ReturnRawObject ();
	
	OP_ASSERT (src->surface);
	
	this->surface->RectBlend ((int)dest_x1, (int)dest_y1,
							  (int)(dest_x2-dest_x1+1), (int)(dest_y2-dest_y1+1),
							  (int)source_x, (int)source_y, src->surface);
}

void
OPGrailPort::CopyRect (OPGrailPortIn source,
					   OPCoord source_x, OPCoord source_y,
					   OPRectIn dest_box, OPRelCoord dest_margin)
{
	OP_METHOD (CopyRect, 5);
	OP_ARG (1, OPGrailPortIn, source);
	OP_ARG (2, OPCoord, source_x);
	OP_ARG (3, OPCoord, source_y);
	OP_ARG (4, OPRectIn, dest_box);
	OP_ARG (5, OPRelCoord, dest_margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(dest_box.DX ()) < (2*dest_margin))
	  || ((OPRelCoord)(dest_box.DY ()) < (2*dest_margin)) ) {
		return;
	}
	
	OPCoord x1 = dest_box.X1 () + dest_margin;
	OPCoord y1 = dest_box.Y1 () + dest_margin;
	OPCoord x2 = dest_box.X2 () - dest_margin;
	OPCoord y2 = dest_box.Y2 () - dest_margin;
	
	this->CopyRect (source, source_x, source_y, x1, y1, x2, y2);
}


#if 0
/*
 *	Set a line of pixels with the specified pixel values. This
 *	is not very fast, but provides an easy way to build some
 *	pixel maps yourself.
 */

void
OPGrailPort::SetColorPixels (OPCoord x, OPCoord y, OPCoord dx,
							 const OPColorPixel* pixels)
{
	OP_METHOD (SetGrayPixels, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, const OPColorPixel*, pixels);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	Count delta = (Count)(dx);

	while (delta--) {
		this->surface->SetPixelRGB ((int)(x++), (int)y, *pixels++);
	}
}
#endif


/*
 *	Set a line of pixels with the specified gray values. This
 *	is not very fast, but provides an easy way to build some
 *	pixel maps yourself.
 */

void
OPGrailPort::SetGrayPixels (OPCoord x, OPCoord y, OPCoord dx,
							const Card8* intensity)
{
	OP_METHOD (SetGrayPixels, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, cont Card8*, intensity);
	OP_ASSERT (this->is_drawing == TRUE);
	
	this->SetRGBPixels (x, y, dx, intensity, intensity, intensity);
}



/*
 *	Set a line of pixels with the specified RGB values. This
 *	is not very fast, but provides an easy way to build some
 *	pixel maps yourself.
 */

void
OPGrailPort::SetRGBPixels (OPCoord x, OPCoord y, OPCoord dx,
						   const Card8* red,
						   const Card8* green,
						   const Card8* blue)
{
	OP_METHOD (SetRGBPixels, 6);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, const Card8*, red);
	OP_ARG (5, const Card8*, green);
	OP_ARG (6, const Card8*, blue);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	Count delta = (Count)(dx);

	while (delta--) {
		OPColorPixel rgb = (*red++ << 16) | (*green++ << 8) | (*blue++);
		this->surface->SetPixelRGB ((int)(x++), (int)y, rgb);
	}
}



/*
 *	Set a line of pixels with the specified RGB values. This
 *	is not very fast, but provides an easy way to build some
 *	pixel maps yourself.
 */

void
OPGrailPort::SetRGBPixels (OPCoord x, OPCoord y, OPCoord dx, const Card8* rgb)
{
	OP_METHOD (SetRGBPixels, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, const Card8*, rgb);
	
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	this->surface->SetPixelLineBGR24 ((int)(x), (int)y, (Card32)(dx), rgb);
}


/*
 *	Set a block of pixels with the specified monochrome values.
 *	The pixels set to "1" will be drawn with the current color,
 *	the others won't be modified.
 */

void
OPGrailPort::SetBitmap (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy, const Card8* bits)
{
	OP_METHOD (SetBitmap, 5);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (3, OPCoord, dy);
	OP_ARG (5, cont Card8*, bits);
	OP_ASSERT (this->is_drawing == TRUE);
	OP_ASSERT (this->surface);
	
	int width = (((int)(dx)) + 7) / 8;
	
	while (dy--) {
		this->surface->SetPixelLineMono ((int)(x), (int)(y), (Card32)(dx), this->fore_color, bits);
		bits += width;
		y++;
	}
}


