/*
 *	sd_file_io.cxx
 *	
 *	MS-Windows 32-bit file input/output functions.
 *	
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"
#include <io.h>

#include <OPaC/app.h>
#include <OPaC/string.h>
#include <OPaC/sd/file_io.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;

/****************************************************************/

#define	MAX_BUF_LENGTH	250
static const OPaC_FilePos mask32 = (Card32)(-1);
	
/****************************************************************/

static void
convert_to_win32 (char* name)
{
	name[MAX_BUF_LENGTH-1] = 0;
	while (*name) {
		switch (*name) {
			case '/':
				*name = '\\';
				break;
			default:
				break;
		}
		name++;
	}
}

static void
convert_from_win32 (char* name)
{
	name[MAX_BUF_LENGTH-1] = 0;
	while (*name) {
		switch (*name) {
			case '\\':
				*name = '/';
				break;
			default:
				break;
		}
		name++;
	}
}

/****************************************************************/

OPaC_File::OPaC_File ()
{
	this->desc       = 0;
	this->share_base = 0;
	this->share_size = 0;
	this->share_desc = 0;
	this->mode       = OPEN_NONE;
}

OPaC_File::~OPaC_File ()
{
	if (this->desc) {
		this->Close ();
	}
}

Bool
OPaC_File::ParsePathAndReplace (const char* path, char* buffer, int size)
{
	char temp[MAX_BUF_LENGTH];
	
	*buffer = 0;
	
	while (path[0] && size) {
		if (path[0] == '%') {
			
			path++;
			
			if (path[0] == '%') {
				*buffer++ = *path++;
				size--;
			}
			
			const char* replace = 0;
			const char* token   = path;
			
			while (path[0] && path[0] != '%') path++;
			int length  = path - token;
			int rep_len = 0;
			
			if (path[0] == 0) {
				return FALSE;
			}
			
			if ( (length == 4)
			  && (strncmp (token, "OPaC", 4) == 0) ) {
				OPApp::CurrentApp ()->ReturnOPaCDirectory ()->GetStringCopy (temp, MAX_BUF_LENGTH);
				replace = temp;
				rep_len = strlen (replace);
			}
			
			if (replace == 0) return FALSE;
			if (rep_len > size) return FALSE;
			
			strcat (buffer, replace);
			buffer += rep_len;
			size   -= rep_len;
			path++;
			
		} else {
			*buffer++ = *path++;
			size--;
		}
	}
	
	if (size) *buffer = 0;
	
	return (size) ? TRUE : FALSE;
}

/****************************************************************/

Bool
OPaC_File::Open (OPStringIn name, Mode mode)
{
	OP_METHOD (OpenRead, 1);
	OP_ARG (1, OPStringIn, name);
	
	char  buforg[MAX_BUF_LENGTH];
	char  buffer[MAX_BUF_LENGTH];
	char* file_name = buffer;
	name->GetStringCopy (buforg, MAX_BUF_LENGTH);
	ParsePathAndReplace (buforg, file_name, MAX_BUF_LENGTH);
	convert_to_win32 (file_name);
	
	DWORD rw   = 0;
	DWORD fmap = FILE_MAP_READ;
	DWORD open = OPEN_EXISTING;
	DWORD flag = FILE_ATTRIBUTE_NORMAL;
	
	if (mode & OPEN_READ) 	{ rw |= GENERIC_READ; }
	if (mode & OPEN_WRITE) 	{ rw |= GENERIC_WRITE; }
	if (mode & OPEN_APPEND) { rw |= GENERIC_WRITE; }
	if (mode & OPEN_CREATE) { open = OPEN_ALWAYS; }
	if (mode & OPEN_TRUNC)  { open = (mode & OPEN_CREATE) ? CREATE_ALWAYS : TRUNCATE_EXISTING; }
	if (mode & OPEN_SYNC)   { flag |= FILE_FLAG_WRITE_THROUGH; }
	
	if (mode & OPEN_WRITE) 		   { fmap = FILE_MAP_ALL_ACCESS; }
	if (mode & OPEN_COPY_ON_WRITE) { fmap = FILE_MAP_COPY; }
	
	if (strncmp (file_name, "mmap:", 5) == 0) {
		this->desc       = OpenFileMapping (fmap, TRUE, file_name+5);
		this->share_desc = this->desc;
	} else {
		this->desc       = CreateFile (file_name, rw, 0, NULL, open, flag, NULL);
		this->share_desc = 0;
	}
	
	if (this->desc == INVALID_HANDLE_VALUE) {
		this->desc = 0;
		return FALSE;
	}
	
	if (mode & OPEN_APPEND) { SetFilePointer (this->desc, 0, NULL, FILE_END); }
	
	this->mode = mode;
	
	return TRUE;
}

Bool
OPaC_File::Close ()
{
	OP_METHOD (Close, 0);
	
	if (this->share_base) {
		this->UnmapSharedMemory ();
	}
	if (this->share_desc) {
		if (this->share_desc != this->desc) {
			CloseHandle (this->share_desc);
		}
		this->share_desc = 0;
	}
	
	if (this->desc == 0) return TRUE;
	
	Bool ok = CloseHandle (this->desc) ? TRUE : FALSE;
	this->desc = 0;
	return ok;
}


/****************************************************************/


Size
OPaC_File::Read (void* void_buffer, Size size)
{
	OP_METHOD (Read, 2);
	OP_ARG (1, void*, void_buffer);
	OP_ARG (2, Size, size);
	
	OP_ASSERT (this->desc);
	
	DWORD total_read = 0;
	ReadFile (this->desc, void_buffer, (DWORD)size, &total_read, NULL);
	
	return total_read;
}


Size
OPaC_File::Write (const void* void_buffer, Size size)
{
	OP_METHOD (Write, 2);
	OP_ARG (1, const void*, void_buffer);
	OP_ARG (2, Size, size);
	
	OP_ASSERT (this->desc);
	
	DWORD total_write = 0;
	WriteFile (this->desc, void_buffer, (DWORD)size, &total_write, NULL);
	
	return total_write;
}


/* 
 *	Map the file into memory. This is only possible if the
 *	file descriptor refers to a memory mappable file (opened
 *	using a "mmap:file_name").
 */

void*
OPaC_File::MapSharedMemory (OPaC_FilePos offset, Size size, OPStringIn map_name)
{
	OP_METHOD (MapSharedMemory, 3);
	OP_ARG (1, OPaC_FilePos, offset);
	OP_ARG (2, Size, size);
	OP_ARG (3, OPStringIn, map_name);
	
	char  buffer[MAX_BUF_LENGTH];
	DWORD prot = PAGE_READONLY;
	char* name = 0;
	
	if (map_name.IsValid ()) {
		char* file_name = buffer;
		map_name->GetStringCopy (file_name, MAX_BUF_LENGTH);
		convert_to_win32 (file_name);
		name = (strncmp (file_name, "mmap:", 5) == 0) ? file_name+5 : file_name;
	}
	
	if (this->mode & OPEN_WRITE)		 { prot = PAGE_READWRITE; }
	if (this->mode & OPEN_COPY_ON_WRITE) { prot = PAGE_WRITECOPY; }
	
	if (this->desc) {
		if (this->share_desc == 0) {
			this->share_desc = CreateFileMapping (this->desc, NULL, prot, 0, size, name);
		}
		if (this->share_desc == 0) {
			return 0;
		}
	} else {
		return 0;	//	NIY : non-file-backed shared memory
	}
	
	DWORD map = FILE_MAP_READ;
	
	if (this->mode & OPEN_WRITE)         map = FILE_MAP_ALL_ACCESS;
	if (this->mode & OPEN_COPY_ON_WRITE) map = FILE_MAP_COPY;
	
	this->share_base = MapViewOfFile (this->share_desc, map, (DWORD)(offset >> 32), (DWORD)(offset), (DWORD)(size));
	this->share_size = size;
	
	return this->share_base;
}

/*
 *	Unmap the view of the file.
 */

Bool
OPaC_File::UnmapSharedMemory ()
{
	OP_METHOD (UnmapSharedMemory, 0);
	
	if (UnmapViewOfFile (this->share_base)) {
		this->share_base = 0;
		this->share_size = 0;
		return TRUE;
	}
	
	return FALSE;
}

/*
 *	Flush the contents of the mapped memory chunk to the
 *	underlying file. This is an asynchronous operation.
 */

Bool
OPaC_File::FlushSharedMemory (void* base, Size size)
{
	OP_METHOD (FlushSharedMemory, 2);
	OP_ARG (1, void*, base);
	OP_ARG (2, Size, size);
	
	return FlushViewOfFile (base, size) ? TRUE : FALSE;
}


Bool
OPaC_File::SeekAbs (OPaC_FilePos offset)
{
	OP_METHOD (SeekAbs, 1);
	OP_ARG (1, OPaC_FilePos, offset);
	
	OP_ASSERT ( (((offset >> 32) & mask32) == 0) || (((offset >> 32) & mask32) == mask32));
	
	DWORD mode = (offset < 0) ? FILE_END : FILE_BEGIN;
	DWORD pos  = (offset < 0) ? (DWORD)(-offset - 1) : (DWORD)(offset);
	
	return (SetFilePointer (this->desc, pos, NULL, mode) == 0xFFFFFFFF) ? FALSE : TRUE;
}


Bool
OPaC_File::SeekRel (OPaC_FilePos offset)
{
	OP_METHOD (SeekRel, 1);
	OP_ARG (1, OPaC_FilePos, offset);
	
	OP_ASSERT ( (((offset >> 32) & mask32) == 0) || (((offset >> 32) & mask32) == mask32));
	
	DWORD mode = FILE_CURRENT;
	DWORD pos  = (DWORD)(offset);
	
	return (SetFilePointer (this->desc, pos, NULL, mode) == 0xFFFFFFFF) ? FALSE : TRUE;
}


OPaC_FilePos
OPaC_File::TellPos () const
{
	OP_METHOD (TellPos, 0);
	
	DWORD mode   = FILE_CURRENT;
	DWORD pos    = 0;
	DWORD offset = SetFilePointer (this->desc, pos, NULL, mode);
	
	OP_ASSERT (offset != 0xFFFFFFFF);
	
	return (OPaC_FilePos)(offset);
}


OPaC_FilePos
OPaC_File::TellSize () const
{
	OP_METHOD (TellPos, 0);
	
	DWORD size = GetFileSize (this->desc, NULL);
	
	return (size != 0xFFFFFFFF) ? (OPaC_FilePos)(size) : 0;
}

Bool
OPaC_File::Flush ()
{
	OP_METHOD (Flush, 0);
	
	return FlushFileBuffers (this->desc) ? TRUE : FALSE;
}



Bool
OPaC_File::Truncate ()
{
	OP_METHOD (Truncate, 0);
	
	return SetEndOfFile (this->desc) ? TRUE : FALSE;
}



Bool
OPaC_File::Resize (OPaC_FilePos size)
{
	OP_METHOD (OPaC_Resize, 1);
	OP_ARG (1, OPaC_FilePos, size);
	
	DWORD mode = FILE_BEGIN;
	DWORD pos  = (DWORD)(size);
	
	SetFilePointer (this->desc, pos, NULL, mode);
	return SetEndOfFile (this->desc) ? TRUE : FALSE;
}





Bool
OPaC_File::DeleteFile (OPStringIn name)
{
	OP_MEMBER (DeleteFile, 1);
	OP_ARG (1, OPStringIn, name);
	
	char buffer[MAX_BUF_LENGTH];
	name->GetStringCopy (buffer, MAX_BUF_LENGTH);
	convert_to_win32 (buffer);
	
	return ::DeleteFile (buffer);
}

Bool
OPaC_File::CreateDirectory (OPStringIn name)
{
	OP_MEMBER (CreateDirectory, 1);
	OP_ARG (1, OPStringIn, name);
	
	char buffer[MAX_BUF_LENGTH];
	name->GetStringCopy (buffer, MAX_BUF_LENGTH);
	convert_to_win32 (buffer);
	
	return ::CreateDirectory (buffer, NULL);
}

Bool
OPaC_File::DeleteDirectory (OPStringIn name)
{
	OP_MEMBER (APP_PUBLICDeleteDirectory, 1);
	OP_ARG (1, OPStringIn, name);
	
	char buffer[MAX_BUF_LENGTH];
	name->GetStringCopy (buffer, MAX_BUF_LENGTH);
	convert_to_win32 (buffer);
	
	return ::RemoveDirectory (buffer);
}

Bool
OPaC_File::SetCurrentDirectory (OPStringIn name)
{
	OP_MEMBER (SetCurrentDirectory, 1);
	OP_ARG (1, OPStringIn, name);
	
	char buffer[MAX_BUF_LENGTH];
	name->GetStringCopy (buffer, MAX_BUF_LENGTH);
	convert_to_win32 (buffer);
	
	return ::SetCurrentDirectory (buffer);
}

Bool
OPaC_File::GetCurrentDirectory (OPStringOut name)
{
	OP_MEMBER (GetCurrentDirectory, 1);
	OP_ARG (1, OPStringOut, name);
	
	char buffer[MAX_BUF_LENGTH];
	::GetCurrentDirectory (MAX_BUF_LENGTH-1, buffer);
	convert_from_win32 (buffer);
	name = buffer;
	return TRUE;
}

Bool
OPaC_File::GetOPaCDirectory (OPStringOut name)
{
	OP_MEMBER (GetOPaCDirectory, 1);
	OP_ARG (1, OPStringOut, name);
	
	char buffer[MAX_BUF_LENGTH];
	::GetCurrentDirectory (MAX_BUF_LENGTH-1, buffer);
	convert_from_win32 (buffer);
	name = buffer;
	name->AppendString ("/../binaries");
	return TRUE;
}

Bool
OPaC_File::GetInitialDirectory (OPStringOut name)
{
	OP_MEMBER (GetInitialDirectory, 1);
	OP_ARG (1, OPStringOut, name);
	static Bool is_dir_ok = FALSE;
	static char buffer[MAX_BUF_LENGTH];
	
	if (is_dir_ok == FALSE) {
		::GetCurrentDirectory (MAX_BUF_LENGTH-1, buffer);
		convert_from_win32 (buffer);
		is_dir_ok = TRUE;
	}
	
	name = buffer;
	return TRUE;
}
