/*
 *	button.cxx
 *
 *	Button class implementation.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/button.h>
#include <OPaC/lookman.h>
#include <OPaC/grail.h>
#include <OPaC/new.h>
#include <OPaC/pubmacro.h>
#include <OPaC/texttag.h>
#include <OPaC/textstyle.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPButton);

/****************************************************************/

OP_PUBLISH_SLOT (OPButton, icon, ReturnIcon, SetIcon, "Button icon", OPView);
OP_PUBLISH_SLOT (OPButton, contents, ReturnContents, SetContents, "Button contents", OPView);
OP_PUBLISH_SLOT (OPButton, shortcut, ReturnShortcut, SetShortcut, "Button shortcut", OPView);
OP_PUBLISH_SLOT (OPButton, back_color, ReturnBackColor, SetBackColor, "Background color", OPColorPixel);

/****************************************************************/

/*
 *	The 3D effect used for buttons takes up some room. It
 *	is stored in global variables for speed :
 */

OPCoord    OPButton::flat_3D_effect  = 0;
OPRelCoord OPButton::press_3D_offset = 0;

/****************************************************************/

/*
 *	Initialise an empty button and define the button's
 *	behaviour according to the Look Manager's policy.
 */

OPButton::OPButton ()
{
	OP_METHOD (OPButton, 0);
	
	this->shape |= OP_CTRL_AUTOHILITE;
	this->shape |= OP_CTRL_AUTOACTHILITE;
	this->shape |= OP_CTRL_AUTOACTIVE;
	this->shape |= OP_CTRL_RELIEF;
	this->shape |= OP_CTRL_ABOVE;
	
	this->do_press_on_action  = OPLookMan::Current ()->IsPolicyHiliteButton () == FALSE;
	this->do_hilite_on_action = OPLookMan::Current ()->IsPolicyHiliteButton () == TRUE;
	
	this->icon       = 0;
	this->contents   = 0;
	this->shortcut   = 0;
	this->back_color = OP_COLOR_TRANSPARENT;
	
	//	This could be done only once, but it is more complicated to find out if
	//	it is useful than to do it every time...
	
	OPButton::flat_3D_effect  = OPLookMan::Current ()->Return3DEffectWidth ();
	OPButton::press_3D_offset = OPLookMan::Current ()->Return3DPressOffset ();
}



/*
 *	Free a button. This will automagically delete the
 *	contents and shortcut views if they are no longer
 *	of any use to anyone else.
 */

OPButton::~OPButton ()
{
	OP_METHOD (~OPButton, 0);
}



/*
 *	This casting function is needed by the OPaC extension
 *	class OPxTitleWidget.
 */

OPView*
OPButton::ReturnObjectAsView ()
{
	return this;
}


/*
 *	Save the button to the archive.
 */

void
OPButton::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPButton icon contents shortcut back_color>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->icon;
	storage << this->contents;
	storage << this->shortcut;
	storage << this->back_color;
	storage << OPStorageSync::end;
}



/*
 *	Restore the button from the archive.
 */

void
OPButton::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);

	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->icon;
	storage >> this->contents;
	storage >> this->shortcut;
	storage >> this->back_color;
	storage >> OPStorageSync::end;
}



/*
 *	Prepare the button for its destruction.
 */

void
OPButton::Delete ()
{
	OP_METHOD (Delete, 0);
	
	OPViewRef none = 0;
	
	if (this->icon.IsValid ())     this->icon->SetParent (none);
	if (this->contents.IsValid ()) this->contents->SetParent (none);
	if (this->shortcut.IsValid ()) this->shortcut->SetParent (none);
	
	this->icon     = 0;
	this->contents = 0;
	this->shortcut = 0;
	
	super::Delete ();
}



/*
 *	This gets called when the button's geometry changes
 *	(move or resize).
 */

void
OPButton::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
	this->UpdateButton ();
}



/*
 *	The button has been pressed / depressed.
 */

void
OPButton::SetPressState (Bool ok, Bool broadcast)
{
	OP_METHOD (SetPressState, 2);
	OP_ARG (1, Bool, ok);
	OP_ARG (2, Bool, broadcast);
	
	super::SetPressState (ok, broadcast);
	this->UpdateButton ();
}



/*
 *	Update the button's contents, icon(s) and shortcut geometry.
 *	This has to be called whenever the button is resized or its
 *	some or all of its contents have changed.
 */

void
OPButton::UpdateButton ()
{
	OP_METHOD (UpdateButton, 0);
	
	OPCoord widget_frame_size = OPLookMan::Current ()->ReturnWidgetFrameSize ();
	
	
	OPCoord fs = (this->shape & OP_CTRL_NOFRAME) ? 0 : widget_frame_size;
	OPCoord ox = (this->is_being_pressed) ? OPButton::press_3D_offset : 0;
	OPCoord oy = (this->is_being_pressed) ? OPButton::press_3D_offset : 0;
	OPCoord mx = OPButton::flat_3D_effect + fs;
	OPCoord my = OPButton::flat_3D_effect + fs;
	
	//	"mx" and "my" are the margins around the button's contents.
	//	"ox" and "oy" are the offsets to the button's contents
	
	OPCoord dx = this->box.DX () - 2 * mx - OPButton::press_3D_offset;
	OPCoord dy = this->box.DY () - 2 * my - OPButton::press_3D_offset;
	
	OPCoord idx = 0;		//	icon size
	OPCoord idy = 0;
	
	OPCoord sdx = 0;		//	shortcut size
	OPCoord sdy = 0;
	
	/*
	 *	The button is built of three distinct parts :
	 *
	 *	- The icon is always placed on the very left.
	 *	- The shortcut is always placed on the very right.
	 *	- The content occupies the remaining space at the center
	 *	  of the button.
	 *
	 *	 ******************************
	 *   * ----- -------------- ----- *
	 *   *| Icn |   Contents   | Sct |*
	 *   * ----- -------------- ----- *
	 *   ******************************
	 *    :     :              :     :
	 *	  :<idx>:<---  dx  --->:<sdx>:
	 */
	
	const OPCoord shortcut_right_margin = 4;
	const OPCoord icon_left_margin      = 4;
	
	if (this->shortcut.IsValid ()) {
		this->shortcut->GetSize (sdx, sdy);
		sdx += shortcut_right_margin;
		dx  -= sdx;
	}
	
	if (this->icon.IsValid ()) {
		this->icon->GetSize (idx, idy);
		idx += icon_left_margin;
		dx  -= idx;
	}
	
	//	Change the position of both icon and shortcut.
	
	if (this->shortcut.IsValid ()) {
		
		OPCoord x = this->box.X2 () - sdx - mx + ox;
		OPCoord y = this->box.Y1 () + ((dy - sdy) / 2) + my + oy;
		
		this->shortcut->SetOrigin (x, y);
	}
	
	if (this->icon.IsValid ()) {
		
		OPCoord x = this->box.X1 () + mx + ox;
		OPCoord y = this->box.Y1 () + ((dy - idy) / 2) + my + oy;
		
		this->icon->SetOrigin (x, y);
	}
	
	//	Change the position and size of the contents.
	
	if (this->contents.IsValid ()) {
		
		OPCoord x = this->box.X1 () + mx + ox + idx;
		OPCoord y = this->box.Y1 () + my + oy;
		
		this->contents->SetOrigin (x, y);
		this->contents->SetSize (dx, dy);
	}
}



/*
 *	Define the contents view. Subscribe the view to this
 *	button.
 */

Bool
OPButton::SetContents (OPViewIn contents)
{
	OP_METHOD (SetContents, 1);
	OP_ARG (1, OPViewIn, contents);
	
	if (this->contents.IsValid ()) {
		
		//	Detach the old contents view from the button.
		
		this->contents->SetParent (OPViewNull);
	}
	
	if (contents.IsValid ()) {
		
		//	Attach the new contents view to the button and set its internal
		//	state in such a way that it will not respond to "Detect" and to
		//	the user interface builder's moves.
		
		contents->SetParent (this);
		contents->SetFlagOkDetect (FALSE);
		contents->SetFlagOkUIMove (FALSE);
	}
	
	this->contents = contents;
	this->UpdateButton ();
	
	return TRUE;
}



/*
 *	Define the shortcut view. Subscribe the view to this
 *	button.
 */

Bool
OPButton::SetShortcut (OPViewIn shortcut)
{
	OP_METHOD (SetShortcut, 1);
	OP_ARG (1, OPViewIn, shortcut);
	
	if (this->shortcut.IsValid ()) {
		
		//	Detach the old shortcut view from the button.
		
		this->shortcut->SetParent (OPViewNull);
	}
	
	if (shortcut.IsValid ()) {
		
		//	Attach the new shortcut view to the button and set its internal
		//	state in such a way that it will not respond to "Detect" and to
		//	the user interface builder's moves.
		
		shortcut->SetParent (this);
		shortcut->SetFlagOkDetect (FALSE);
		shortcut->SetFlagOkUIMove (FALSE);
	}
	
	this->shortcut = shortcut;
	this->UpdateButton ();
	
	return TRUE;
}



/*
 *	Define the icon view. Subscribe the view to this button.
 */

Bool
OPButton::SetIcon (OPViewIn icon)
{
	OP_METHOD (SetIcon, 1);
	OP_ARG (1, OPViewIn, icon);
	
	if (this->icon.IsValid ()) {
		
		//	Detach the old icon view from the button.
		
		this->icon->SetParent (OPViewNull);
	}
	
	if (icon.IsValid ()) {
		
		//	Attach the new icon view to the button and set its internal
		//	state in such a way that it will not respond to "Detect" and to
		//	the user interface builder's moves.
		
		icon->SetParent (this);
		icon->SetFlagOkDetect (FALSE);
		icon->SetFlagOkUIMove (FALSE);
	}
	
	this->icon = icon;
	this->UpdateButton ();
	
	return TRUE;
}



/*
 *	Define the title of the button. This is a simplified
 *	alias for SetText.
 */

Bool
OPButton::SetTitle (OPStringIn text)
{
	OP_METHOD (SetTitle, 1);
	OP_ARG (1, OPStringIn, text);
	
	return this->SetText (text);
}


/*
 *	Define the title of the button by creating the
 *	text tag automatically from the name.
 */

Bool
OPButton::SetText (OPStringIn text, Card8 mode, Bool relief)
{
	OP_METHOD (SetText, 3);
	OP_ARG (1, OPStringIn, text);
	OP_ARG (2, Card8, mode);
	OP_ARG (3, Bool, relief);
	
	if (mode == 0xFF) {
		mode = OP_TEXT_HPOS_CENTER | OP_TEXT_VPOS_CENTER;
	}
	
	OPTextTagRef tag = OP_NEW (OPTextTag);
	
	tag->SetText (text);
	tag->SetMode (mode);

	if (relief) {
		this->shape |= OP_CTRL_RELIEF;
		this->shape |= OP_CTRL_ABOVE;
		this->SetPressMode (TRUE);
		this->SetHiliteMode (FALSE);
	} else {
		this->shape &= ~OP_CTRL_RELIEF;
		this->shape &= ~OP_CTRL_ABOVE;
		this->SetPressMode (FALSE);
		this->SetHiliteMode (TRUE);
	}
	
	return this->SetContents (tag);
}


/*
 *	Return the current contents view (there may be none).
 */

Bool
OPButton::GetContents (OPViewOut contents) const
{
	OP_METHOD (GetContents, 1);
	OP_ARG (1, OPViewOut, contents);
	
	contents = this->contents;
	
	return TRUE;
}

const OPViewRef&
OPButton::ReturnContents () const
{
	OP_METHOD (ReturnContents, 0);
	return this->contents;
}

/*
 *	Return the current shortcut view (there may be none).
 */

Bool
OPButton::GetShortcut (OPViewOut shortcut) const
{
	OP_METHOD (GetShortcut, 1);
	OP_ARG (1, OPViewOut, shortcut);
	
	shortcut = this->shortcut;
	
	return TRUE;
}

const OPViewRef&
OPButton::ReturnShortcut () const
{
	OP_METHOD (ReturnShortcut, 0);
	return this->shortcut;
}


/*
 *	Return the current icon view (there may be none).
 */

Bool
OPButton::GetIcon (OPViewOut icon) const
{
	OP_METHOD (GetIcon, 1);
	OP_ARG (1, OPViewOut, icon);
	
	icon = this->icon;
	
	return TRUE;
}

const OPViewRef&
OPButton::ReturnIcon () const
{
	OP_METHOD (ReturnIcon, 0);
	return this->icon;
}

/****************************************************************/

Bool
OPButton::SetBackColor (OPColorPixelIn color)
{
	OP_METHOD (SetBackColor, 1);
	OP_ARG (1, OPColorPixelIn, color);
	
	this->back_color = color;
	return TRUE;
}

Bool
OPButton::GetBackColor (OPColorPixelOut color) const
{
	OP_METHOD (GetBackColor, 1);
	OP_ARG (1, OPColorPixelOut, color);
	
	color = this->back_color;
	return TRUE;
}

const OPColorPixel&
OPButton::ReturnBackColor () const
{
	OP_METHOD (ReturnBackColor, 0);
	
	return this->back_color;
}

/****************************************************************/

/*
 *	Return the button's minimal size. This will be the
 *	content's size added to the shortcut's.
 */

void
OPButton::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	OPCoord fs  = (this->shape & OP_CTRL_NOFRAME) ? 0 : OPLookMan::Current ()->ReturnWidgetFrameSize ();
	OPCoord dx1 = 0;
	OPCoord dy1 = 0;
	OPCoord dx2 = 0;
	OPCoord dy2 = 0;
	OPCoord dx3 = 0;
	OPCoord dy3 = 0;
	
	if (this->contents.IsValid ()) {
		this->contents->GetMinSize (dx1, dy1);
	}
	
	if (this->shortcut.IsValid ()) {
		this->shortcut->GetMinSize (dx2, dy2);
		dx2 += 4;
	}
	
	if (this->icon.IsValid ()) {
		this->icon->GetMinSize (dx3, dy3);
		dx3 += 4;
	}
	
	dx = 2 * (OPButton::flat_3D_effect + fs);
	dy = 2 * (OPButton::flat_3D_effect + fs);
	
	dx += dx1 + dx2 + dx3;
	dy += (dy1 > dy2) ? ( (dy1 > dy3) ? dy1 : dy3)
					  : ( (dy2 > dy3) ? dy2 : dy3);
	
	dx += OPButton::press_3D_offset;
	dy += OPButton::press_3D_offset;
}



/*
 *	Return the button's maximum size. Return an arbitrary
 *	huge value.
 */

void
OPButton::GetMaxSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMaxSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	super::GetMaxSize (dx, dy);
}



/*
 *	A child is signalling that it has changed its geometry.
 *	The button might want to update its size in turn.
 */

void
OPButton::ChildChangedSize (OPViewIn child)
{
	OP_METHOD (ChildChangedSize, 1);
	OP_ARG (1, OPViewIn, child);
	
	super::ChildChangedSize (child);
	
	this->UpdateButton ();
}



/*
 *	The only message a button understands is that the look
 *	manager has changed.
 */

void
OPButton::Broadcast (OPAtomIn message, OPObjectIn object)
{
	OP_METHOD (Broadcast, 2);
	OP_ARG (1, OPAtomIn, message);
	OP_ARG (2, OPObjectIn, object);
	
	if (message == this->atom_look_man_changed) {
		
		this->do_press_on_action     = OPLookMan::Current ()->IsPolicyHiliteButton () == FALSE;
		this->do_hilite_on_action    = OPLookMan::Current ()->IsPolicyHiliteButton () == TRUE;
		
		OPButton::flat_3D_effect  = OPLookMan::Current ()->Return3DEffectWidth ();
		OPButton::press_3D_offset = OPLookMan::Current ()->Return3DPressOffset ();
		
		this->UpdateButton ();
	}
	
	super::Broadcast (message, object);
}



/*
 *	Draw the button. Let the control draw the special
 *	adornments and the child views.
 */

void
OPButton::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	const OPLookManRef& lookman = OPLookMan::Current ();
	
	OPCoord effect = (this->shape & OP_CTRL_RELIEF) ? lookman->Return3DEffectWidth () : 0;
	
	port->SetForeColor (this->back_color);
	port->FillRect (this->box, effect);
	
	super::Draw (port, rect, what);
}


