/*
 *	colslider.cxx
 *
 *	Colored Slider class definition.
 *	
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/colslider.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/window.h>

#include <OPaC/notmacro.h>
#include <OPaC/pubmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPColSlider);

/****************************************************************/

OP_PUBLISH_SLOT (OPColSlider, min_color, ReturnMinColor, SetMinColor, "Minimum color", OPColorPixel);
OP_PUBLISH_SLOT (OPColSlider, max_color, ReturnMaxColor, SetMaxColor, "Maximum color", OPColorPixel);
OP_PUBLISH_SLOT_READ (OPColSlider, current_color, ReturnCurrentColor, "Current color", OPColorPixel);

OP_NOTIFY_CHANGE (OPColSlider, current_color, Scroll);

/****************************************************************/

/*
 *	The colored slider specifies an upper and a lower bound
 *	for drawing the body of the slider.
 */

OPColSlider::OPColSlider ()
{
	OP_METHOD (OPColSlider, 0);
	
	this->min_color = OP_COLOR_WIDGET_BACK;
	this->max_color = OP_COLOR_WIDGET_BACK;
}



/*
 *	Clean up the colored slider.
 */

OPColSlider::~OPColSlider ()
{
	OP_METHOD (~OPColSlider, 0);
}



/*
 *	Store the slider...
 */

void
OPColSlider::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPColSlider min_color max_color>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->min_color << this->max_color;
	storage << OPStorageSync::end;
}



/*
 *	Restore the slider...
 */

void
OPColSlider::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;	
	storage >> OPStorageSync::begin;
	storage >> this->min_color >> this->max_color;
	storage >> OPStorageSync::end;
}


static OPColorPixel
mix_colors (OPColorPixelIn c1, OPColorPixelIn c2, Real ratio)
{
	Real x = 1.0 - ratio;
	Real y = ratio;
	
	Real r = c1.RealRed () * x + c2.RealRed () * y;
	Real g = c1.RealGreen () * x + c2.RealGreen () * y;
	Real b = c1.RealBlue () * x + c2.RealBlue () * y;
	Real a = c1.RealTransparency () * x + c2.RealTransparency () * y;
	
	return OPColorPixel (r, g, b, a);
}


/*
 *	Draw the colored slider's body.
 */

void
OPColSlider::DrawBody (OPGrailPortIn port)
{
	OP_METHOD (DrawBody, 1);
	OP_ARG (1, OPGrailPortIn, port);
	
	OP_ASSERT (this->is_knob_ok);
	
	port->SetLineStyle (OP_LINE_SOLID);
	
	if (this->is_horiz) {
		
		OPCoord x1 = this->knob_min + 6;
		OPCoord y1 = this->box.Y1 () + this->look_frame + 6;
		OPCoord x2 = this->knob_max - 6;
		OPCoord y2 = this->box.Y2 () - this->look_frame - 6;
		OPCoord dx = x2 - x1 + 1;
		OPCoord dy = y2 - y1 + 1;
		
		for (OPCoord x = x1; x <= x2; x++) {
			Real ratio = (x - x1) / dx;
			port->SetForeColor (mix_colors (this->min_color, this->max_color, ratio));
			port->DrawLineDY (x, y1, dy);
		}
		
	} else {
		
		OPCoord x1 = this->box.X1 () + this->look_frame + 6;
		OPCoord y1 = this->knob_min + 6;
		OPCoord x2 = this->box.X2 () - this->look_frame - 6;
		OPCoord y2 = this->knob_max - 6;
		OPCoord dx = x2 - x1 + 1;
		OPCoord dy = y2 - y1 + 1;
		
		for (OPCoord y = y1; y <= y2; y++) {
			Real ratio = (y - y1) / dy;
			port->SetForeColor (mix_colors (this->min_color, this->max_color, ratio));
			port->DrawLineDX (x1, y, dx);
		}
	}
}


Bool
OPColSlider::SetMinColor (OPColorPixelIn color)
{
	OP_METHOD (SetMinColor, 1);
	OP_ARG (1, OPColorPixelIn, color);
	
	this->min_color = color;
	
	return TRUE;
}

Bool
OPColSlider::SetMaxColor (OPColorPixelIn color)
{
	OP_METHOD (SetMaxColor, 1);
	OP_ARG (1, OPColorPixelIn, color);
	
	this->max_color = color;
	
	return TRUE;
}

OPColorPixel
OPColSlider::ReturnMinColor () const
{
	OP_METHOD (ReturnMinColor, 0);
	return this->min_color;
}

OPColorPixel
OPColSlider::ReturnMaxColor () const
{
	OP_METHOD (ReturnMaxColor, 0);
	return this->max_color;
}

OPColorPixel
OPColSlider::ReturnCurrentColor () const
{
	OP_METHOD (ReturnCurrentColor, 0);

	return mix_colors (this->min_color, this->max_color, this->value);
}

