/*
 *	dbscrlist.cxx
 *
 *	Database-based scroll list. This provides the mechanisms
 *	needed to easily represent a collection of items visually.
 *
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/database.h>
#include <OPaC/dbscrlist.h>
#include <OPaC/new.h>
#include <OPaC/objectpane.h>
#include <OPaC/scrgrid.h>
#include <OPaC/selgrid.h>
#include <OPaC/textcaption.h>
#include <OPaC/textstyle.h>
#include <OPaC/values.h>

#include <OPaC/pubmacro.h>
#include <OPaC/notmacro.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPDBScrollList);

OP_TARGET_ACTOR (OPDBScrollList, Update);
OP_TARGET_ACTOR (OPDBScrollList, RemoveAll);
OP_TARGET_ACTOR (OPDBScrollList, HandleCellActivated);

OP_NOTIFY_EVENT (OPDBScrollList, SelectionChanged);

/****************************************************************/

OPAtomRef OPDBScrollList::atom_index;
OPAtomRef OPDBScrollList::atom_name;
OPAtomRef OPDBScrollList::atom_data;
OPAtomRef OPDBScrollList::atom_text;
OPAtomRef OPDBScrollList::atom_large_icon;
OPAtomRef OPDBScrollList::atom_small_icon;
OPAtomRef OPDBScrollList::atom_mark_icon;

/****************************************************************/


OPDBScrollList::OPDBScrollList ()
{
	OP_METHOD (OPDBScrollList, 0);
	
	this->is_unique   = TRUE;
}

OPDBScrollList::~OPDBScrollList ()
{
	OP_METHOD (~OPDBScrollList, 0);
}


/*
 *	Save the data.
 */

void
OPDBScrollList::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPDBScrollList db db_model db_provider db_hash"
							   " pane pane_title caption grid is_unique>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->db << this->db_model << this->db_provider << this->db_hash;
	storage << this->pane << this->pane_title << this->caption << this->grid;
	storage << this->is_unique;
	storage << OPStorageSync::end;
}



/*
 *	Restore the data.
 */

void
OPDBScrollList::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->db >> this->db_model >> this->db_provider >> this->db_hash;
	storage >> this->pane >> this->pane_title >> this->caption >> this->grid;
	storage >> this->is_unique;
	storage >> OPStorageSync::end;
}



/*
 *	Setup the data base; also create the associated model
 *	and provider.
 */

Bool
OPDBScrollList::SetupDataBase (OPStringIn title, Bool unique)
{
	OP_METHOD (SetupDataBase, 2);
	OP_ARG (1, OPStringIn, title);
	OP_ARG (2, Bool, unique);
	
	OP_ASSERT (this->db.IsInvalid ());
	
	this->db          = OP_NEW (OPDataBase);
	this->db_hash     = OP_NEW (OPHashTable);
	this->db_model    = OP_NEW (OPEntityModel);
	this->db_provider = OP_NEW (OPDBProvider);
	
	this->db_model->Initialise (this->db, "ListModel", 7);
	this->db_provider->Initialise (this->db);
	
	OPDBScrollList::atom_index = "index";
	OPDBScrollList::atom_name  = "name";
	OPDBScrollList::atom_data  = "data";
	OPDBScrollList::atom_text  = "text";
	OPDBScrollList::atom_large_icon = "large icon";
	OPDBScrollList::atom_small_icon = "small icon";
	OPDBScrollList::atom_mark_icon  = "mark icon";
	
	this->db_model->AddAttribute (OPDBScrollList::atom_index);
	this->db_model->AddAttribute (OPDBScrollList::atom_name);
	this->db_model->AddAttribute (OPDBScrollList::atom_data);
	this->db_model->AddAttribute (OPDBScrollList::atom_text);
	this->db_model->AddAttribute (OPDBScrollList::atom_large_icon);
	this->db_model->AddAttribute (OPDBScrollList::atom_small_icon);
	this->db_model->AddAttribute (OPDBScrollList::atom_mark_icon);
	
	this->db_provider->SetDBNameForSorting (OPDBScrollList::atom_index);
	this->db_provider->SetDBNameForObject (OPDBScrollList::atom_data);
	this->db_provider->SetDBNameForText (OPDBScrollList::atom_text);
	this->db_provider->SetDBNameForLargeIcon (OPDBScrollList::atom_large_icon);
	this->db_provider->SetDBNameForSmallIcon (OPDBScrollList::atom_small_icon);
	this->db_provider->SetDBNameForMarkIcon (OPDBScrollList::atom_mark_icon);
	
	this->pane_title  = title;
	this->is_unique   = unique;
	
	//	Now, create the pane and the caption...
	
	this->pane    = OP_NEW (OPObjectPane);
	this->caption = OP_NEW (OPTextCaption);
	
	OPCoord dx = 0;
	OPCoord dy = 0;
	OPCoord x  = 0;
	OPCoord y  = 0;
	OPRect  box;
	
	this->GetSize (dx, dy);
	this->GetOrigin (x, y);
	
	box.SetOrigin (x, y+26);
	box.SetSize (dx, dy-26);
	
	this->pane->SetParent (this);
	this->pane->Initialise (box, this->db_provider, "OPSelGrid");
	this->pane->SetMode (OP_PANE_MODE_NAME_ONLY | OP_PANE_LAYOUT_LINES);
	
	this->caption->SetParent (this);
	this->caption->SetOrigin (x, y);
	this->caption->SetSize (dx, 24);
	this->caption->SetText (this->pane_title);
	
	this->pane->GetCellGrid (this->grid);
	
	this->grid->AttachObserver (this, "CellActivated", "HandleCellActivated");
	this->grid->SetSelectOnClick (TRUE);
	this->grid->SetSelectMultipleOk (FALSE);
	this->grid->SetSelectAtLeastOne (TRUE);
	this->grid->EnableDragging (FALSE);
	
	return TRUE;
}


/*
 *	Insert an element into the data base.
 */

Bool
OPDBScrollList::Insert (OPObjectIn index, OPTextIn name, OPObjectIn data,
						OPTextIn text, OPLookIcon large_icon, OPLookIcon small_icon,
						OPLookIcon mark_icon)
{
	OP_METHOD (Insert, 7);
	OP_ARG (1, OPObjectIn, index);
	OP_ARG (2, OPObjectIn, data);
	OP_ARG (3, OPTextIn, name);
	OP_ARG (4, OPTextIn, text);
	OP_ARG (5, OPLookIcon, large_icon);
	OP_ARG (6, OPLookIcon, small_icon);
	OP_ARG (7, OPLookIcon, mark_icon);
	
	if (this->db.IsInvalid ()) {
		return FALSE;
	}
	
	if ( (this->is_unique == FALSE)
	  || (this->db_hash->LookFor (name) == FALSE) ) {
		
		//	The element is not yet stored in the hash table or we don't require
		//	it to be unique. Therefore, add it and create a new entity in the
		//	data base.
		
		this->db_hash->Insert (name);
		
		OPEntityRef entity = OP_NEW (OPEntity);
		
		entity->Initialise (this->db_model);
		entity->SetAttribute (1, index);
		entity->SetAttribute (2, name);
		entity->SetAttribute (3, data);
		entity->SetAttribute (4, text);
		entity->SetAttribute (5, (OPValueRef)(Int32)large_icon);
		entity->SetAttribute (6, (OPValueRef)(Int32)small_icon);
		entity->SetAttribute (7, (OPValueRef)(Int32)mark_icon);
		
		this->db->Insert (entity);
		
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Update the pane according to the data base contents.
 */

Bool
OPDBScrollList::Update ()
{
	OP_METHOD (Update, 0);
	
	if (this->pane.IsValid ()) {
		this->pane->Update ();
	}
	
	return TRUE;
}


/*
 *	Remove every element which had been inserted previously.
 */

Bool
OPDBScrollList::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);
	
	this->grid->RemoveEveryHilite ();
	
	if (this->db.IsValid ()) {
		
		this->db->RemoveAll ();
		this->db_hash->RemoveAll ();
		
		if (this->pane.IsValid ()) {
			this->pane->Update ();
		}
		
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	Report that a change occurred.
 */

Bool
OPDBScrollList::HandleCellActivated ()
{
	OP_METHOD (HandleCellActivated, 0);
	
	this->QuickReport (notify_SelectionChanged);
	
	return TRUE;
}


/*
 *	Find the nth selected entity.
 */

Bool
OPDBScrollList::GetSelectedEntity (OPEntityOut entity, Count nth) const
{
	OP_METHOD (GetSelectedEntity, 2);
	OP_ARG (1, OPEntityOut, entity);
	OP_ARG (2, Count, nth);
	
	Count max = this->grid->ReturnSelectNum ();
	
	if (nth < max) {
		
		Count x = 0;
		Count y = 0;
		
		if (this->grid->FindActiveHilite (x, y, nth)) {
			return this->db->GetEntity (OPDBScrollList::atom_index, y, entity);
		}
	}
	
	return FALSE;
}

Count
OPDBScrollList::CountSelectedItems () const
{
	OP_METHOD (CountSelectedItems, 0);
	return this->grid->ReturnSelectNum ();
}

Bool
OPDBScrollList::GetSelectedIndex (OPObjectOut index, Count nth) const
{
	OP_METHOD (GetSelectedIndex, 2);
	OP_ARG (1, OPObjectOut, index);
	OP_ARG (2, Count, nth);
	
	OPEntityRef entity;
	
	if (this->GetSelectedEntity (entity, nth)) {
		return entity->GetAttribute (1, index);
	}
	
	return FALSE;
}

Bool
OPDBScrollList::GetSelectedName (OPTextOut name, Count nth) const
{
	OP_METHOD (GetSelectedName, 2);
	OP_ARG (1, OPTextOut, name);
	OP_ARG (2, Count, nth);
	
	OPEntityRef entity;
	
	if (this->GetSelectedEntity (entity, nth)) {
		return entity->GetAttribute (2, name);
	}
	
	return FALSE;
}

Bool
OPDBScrollList::GetSelectedData (OPObjectOut data, Count nth) const
{
	OP_METHOD (GetSelectedData, 2);
	OP_ARG (1, OPObjectOut, data);
	OP_ARG (2, Count, nth);
	
	OPEntityRef entity;
	
	if (this->GetSelectedEntity (entity, nth)) {
		return entity->GetAttribute (3, data);
	}
	
	return FALSE;
}

Bool
OPDBScrollList::GetSelectedText (OPTextOut text, Count nth) const
{
	OP_METHOD (GetSelectedText, 2);
	OP_ARG (1, OPObjectOut, text);
	OP_ARG (2, Count, nth);
	
	OPEntityRef entity;
	
	if (this->GetSelectedEntity (entity, nth)) {
		return entity->GetAttribute (4, text);
	}
	
	return FALSE;
}

Bool
OPDBScrollList::GetSelectedLargeIcon (OPLookIconOut icon, Count nth) const
{
	OP_METHOD (GetSelectedLargeIcon, 2);
	OP_ARG (1, OPLookIconOut, icon);
	OP_ARG (2, Count, nth);
	
	OPValueRef  value;
	OPEntityRef entity;
	
	if (this->GetSelectedEntity (entity, nth)) {
		if (entity->GetAttribute (5, value)) {
			icon = (OPLookIcon) value->ReturnInt32 ();
			return TRUE;
		}
	}
	
	return FALSE;
}

Bool
OPDBScrollList::GetSelectedSmallIcon (OPLookIconOut icon, Count nth) const
{
	OP_METHOD (GetSelectedSmallIcon, 2);
	OP_ARG (1, OPLookIconOut, icon);
	OP_ARG (2, Count, nth);
	
	OPValueRef  value;
	OPEntityRef entity;
	
	if (this->GetSelectedEntity (entity, nth)) {
		if (entity->GetAttribute (6, value)) {
			icon = (OPLookIcon) value->ReturnInt32 ();
			return TRUE;
		}
	}
	
	return FALSE;
}

Bool
OPDBScrollList::GetSelectedMarkIcon (OPLookIconOut icon, Count nth) const
{
	OP_METHOD (GetSelectedMarkIcon, 2);
	OP_ARG (1, OPLookIconOut, icon);
	OP_ARG (2, Count, nth);
	
	OPValueRef  value;
	OPEntityRef entity;
	
	if (this->GetSelectedEntity (entity, nth)) {
		if (entity->GetAttribute (7, value)) {
			icon = (OPLookIcon) value->ReturnInt32 ();
			return TRUE;
		}
	}
	
	return FALSE;
}


/*
 *	Select a cell in the grid.
 */

Bool
OPDBScrollList::SetSelectedCell (Count index, Bool report)
{
	OP_METHOD (SetSelectedCell, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, Bool, report);
	
	return this->grid->SetActiveHilite (0, index, TRUE, report);
}


/*
 *	Find which cell is selected. This can be used to find
 *	the "nth" selected cell, starting with 0.
 */

Bool
OPDBScrollList::GetSelectedCell (CountOut index, Count nth) const
{
	OP_METHOD (GetSelectedCell, 2);
	OP_ARG (1, CountOut, index);
	OP_ARG (2, Count, nth);
	
	Count max = this->grid->ReturnSelectNum ();
	
	if (nth < max) {
		
		Count x = 0;
		Count y = 0;
		
		if (this->grid->FindActiveHilite (x, y, nth)) {
			index = y;
			return TRUE;
		}
	}
	
	return FALSE;
}

