/*
 *	dragview.cxx
 *
 *	The drag view is provided in order to be able to move views around the
 *	screen and drop them on other OPaC views.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/dragview.h>
#include <OPaC/event.h>
#include <OPaC/grail.h>
#include <OPaC/gridview.h>
#include <OPaC/new.h>
#include <OPaC/window.h>

#include <OPaC/notmacro.h>
#include <OPaC/pubmacro.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPDragView);

/****************************************************************/

OP_TARGET_CLASS (OPDragView);
OP_TARGET_ARG (OPObjectRef, target, "target object");
OP_TARGET_CALL (CommitDrop (target));
OP_TARGET_END ();

OP_TARGET_CLASS (OPDragView);
OP_TARGET_ARG (OPObjectRef, target, "target object");
OP_TARGET_CALL (CommitConnect (target));
OP_TARGET_END ();

OP_TARGET_CLASS (OPDragView);
OP_TARGET_ARG (OPObjectRef, target, "target object");
OP_TARGET_CALL (SubstituteTarget (target));
OP_TARGET_END ();

OP_NOTIFY_EVENT (OPDragView, CommitDrop);
OP_NOTIFY_EVENT (OPDragView, CancelDrop);
OP_NOTIFY_EVENT (OPDragView, CommitSmartDrop);

OP_NOTIFY_CHANGE (OPDragView, drag_object, CommitDrop);
OP_NOTIFY_CHANGE (OPDragView, drag_object, CancelDrop);

OP_PUBLISH_SLOT (OPDragView, drag_object, ReturnDragObject, SetDragObject, "Drag object", OPObject);

/****************************************************************/

/*
 *	Initialise an empty drag view. The window won't be
 *	opened until Setup is called.
 */

OPDragView::OPDragView ()
{
	OP_METHOD (OPDragView, 0);
	
	this->style      = OP_WIN_NAKED | OP_WIN_TOP_LEVEL;
	this->dragging   = FALSE;
	this->auto_stop  = FALSE;
	this->auto_oops  = FALSE;
	this->auto_embed = FALSE;
	this->old_state  = 0;
	this->start_x    = 0;
	this->start_y    = 0;
	this->ox         = 0;
	this->oy         = 0;
	
	this->is_hot_spot_ok = FALSE;
	this->drop_only_in_view = FALSE;
	this->notifying_drop = FALSE;
	this->drop_cancelled = FALSE;
}



/*
 *	Destruction; nothing special to do here.
 */

OPDragView::~OPDragView ()
{
	OP_METHOD (~OPDragView, 0);
}



/*
 *	Store the drag view,
 */

void
OPDragView::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);

	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPDragView drag_view drag_object"
							   " auto_stop auto_oops auto_embed>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->drag_view << this->drag_object;
	storage << this->auto_stop << this->auto_oops << this->auto_embed;
	storage << OPStorageSync::end;
}



/*
 *	Restore the drag view.
 */

void
OPDragView::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);

	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->drag_view >> this->drag_object;
	storage >> this->auto_stop >> this->auto_oops >> this->auto_embed;
	storage >> OPStorageSync::end;
}



/*
 *	Prepare for suicide.
 */

void
OPDragView::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->drag_view.IsValid ()) {
		OPViewRef view = this->drag_view;
		this->drag_view = 0;
		view->Delete ();
	}

	super::Delete ();
}



/*
 *	Initialise the drag view : create the window and open
 *	it in order to be ready to do some work.
 */

void
OPDragView::Setup (OPRectIn box)
{
	OP_METHOD (Setup, 1);
	OP_ARG (1, OPRectIn, box);
	
	this->Close ();
	this->SetOrigin (box.X1 (), box.Y1 ());
	this->SetSize (box.DX (), box.DY ());
	this->ShowNormal ();
}



/*
 *	Initialise the drag view with the previous coordinates
 *	of the specified view. That one will be used for further
 *	dragging operations. Automatically call DragBegin.
 */

void
OPDragView::Setup (OPViewIn view, Bool auto_embed)
{
	OP_METHOD (Setup, 1);
	OP_ARG (1, OPViewIn, view);
	
	OPWindowRef window;
	
	if ( (view.IsValid ())
	  && (view->GetWindow (window)) ) {
		
		OPRect  box = view->ReturnBox ();
		OPCoord x   = 0;
		OPCoord y   = 0;
		
		window->GetOrigin (x, y);
		
		x += box.X1 ();
		y += box.Y1 ();
		
		box.SetOrigin (x, y);
		
		this->auto_embed = auto_embed;
		this->Setup (box);
		this->DragBegin (view);
		
	} else {
		
		this->DragEnd ();
		this->Close ();
	}
}


/*
 *	Constrain the drop to the specified view.
 */

void
OPDragView::DropOnlyInView (OPViewIn view)
{
	OP_METHOD (DropOnlyInView, 1);
	OP_ARG (1, OPViewIn, view);
	
	this->drop_only_in_view = view.IsValid ();
	
	if (this->drop_only_in_view) {
		view->GetWindow (this->drop_only_window);
		view->GetBox (this->drop_only_rect);
	} else {
		this->drop_only_window = OPWindowNull;
		this->drop_only_rect.MakeEmpty ();
	}
}


/*
 *	Set the object which is embedded into the drag view.
 */

Bool
OPDragView::SetDragObject (OPObjectIn object)
{
	OP_METHOD (SetDragView, 1);
	OP_ARG (1, OPObjectIn, object);
	
	this->drag_object = object;
	
	return TRUE;
}

/*
 *	Return the object which is embedded into the drag view.
 */

const OPObjectRef&
OPDragView::ReturnDragObject () const
{
	OP_METHOD (ReturnDragObject, 0);
	
	if (this->drag_object.IsValid ()) {
		return this->drag_object;
	}
	
	return this->drag_view;
}



/*
 *	Define the automatic working mode of the drag view.
 *
 *	- auto stop => handle release of button.
 *	- auto oops => always do the aborted dropping animation.
 *	- auto embed => embed view into window on drop.
 */

void
OPDragView::SetAutoStop (Bool yes)
{
	OP_METHOD (SetAutoStop, 1);
	OP_ARG (1, Bool, yes);
	
	this->auto_stop = yes;
}

void
OPDragView::SetAutoOops (Bool yes)
{
	OP_METHOD (SetAutoOops, 1);
	OP_ARG (1, Bool, yes);
	
	this->auto_oops = yes;
}

void
OPDragView::SetAutoEmbed (Bool yes)
{
	OP_METHOD (SetAutoEmbed, 1);
	OP_ARG (1, Bool, yes);
	
	this->auto_embed = yes;
}


/*
 *	Define the hot spot. This is corresponds to the mouse
 *	co-ordinates in the view where the original dragging
 *	operation started.
 */

void
OPDragView::SetHotSpot (OPCoord x, OPCoord y)
{
	OP_METHOD (SetHotSpot, 2);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	
	this->is_hot_spot_ok = TRUE;
	
	this->ox = (Int32) x;
	this->oy = (Int32) y;
}


/*
 *	Start the dragging operation. Setup everything so the
 *	user only has to click and move around the window.
 */

void
OPDragView::DragBegin (OPViewIn view)
{
	OP_METHOD (DragBegin, 1);
	OP_ARG (1, OPViewIn, view);
	
	if (this->drag_view.IsValid ()) {
		this->DragEnd ();
	}
	
	this->drag_view = view;
	
	if ( (this->drag_object.IsInvalid ())
	  && (this->drag_view->GetAttribute ("embedded_object_for_drag", this->drag_object)) ) {
		
		//	The view is just visual; it represents an object which will actually be
		//	dragged and dropped.
		
		this->auto_embed = FALSE;
	}
	
	if (this->drag_view.IsValid ()) {
		
		OPViewRef self   = this;
		OPViewRef parent = 0;
		OPCoord   dx     = 0;
		OPCoord   dy     = 0;
		
		this->drag_view->GetParent (parent);
		this->drag_view->GetSize (dx, dy);
		
		this->old_state = this->drag_view->ReturnState ();
		
		Card32 idx = (Card32)(dx+1) / 2;
		Card32 idy = (Card32)(dy+1) / 2;
		
		dx = 2*idx;
		dy = 2*idy;
		
		this->SetSize (dx, dy);
		
		//	Insert the view into the parent, that's self.
		
		this->drag_view->SetParent (self);
		this->drag_view->SetOrigin (0, 0);
		
		//	Disable some features of the view which is being dragged.
		
		this->drag_view->ChangeState (OP_VIEW_IS_ALIVE,   FALSE);
		this->drag_view->ChangeState (OP_VIEW_OK_DETECT,  FALSE);
		this->drag_view->ChangeState (OP_VIEW_OK_MAIN,    FALSE);
		this->drag_view->ChangeState (OP_VIEW_OK_UI_EDIT, FALSE);
		this->drag_view->ChangeState (OP_VIEW_OK_UI_MOVE, FALSE);
		
		this->Display ();
		
		if (parent.IsValid ()) {
			parent->Display ();
		}
	}
}



/*
 *	The dragging comes to an end. The view will no longer
 *	be our child.
 */

void
OPDragView::DragEnd ()
{
	OP_METHOD (DragEnd, 0);

	if (this->drag_view.IsValid ()) {
		OPViewRef none = 0;
		this->drag_view->SetParent (none);
		this->drag_view->ChangeState (this->old_state, TRUE);
	}
}



/*
 *	The drag operation has been cancelled.
 */

void
OPDragView::DragForget ()
{
	OP_METHOD (DragForget, 0);
	this->DragEnd ();
	this->drag_view = OPViewNull;
}



/*
 *	The drag operation resulted in a valid drop. The view will
 *	be pasted into the destination drop view if embedding is on.
 *	Otherwise, a "drag & drop" message will be sent.
 */

void
OPDragView::DragDrop ()
{
	OP_METHOD (DragDrop, 0);
	
	if (this->drag_view.IsValid ()) {
		
		OPWindowRef window;
		
		if ( (this->drop_view.IsValid ())
		  && (this->drop_view->GetWindow (window)) ) {
			
			OPCoord x = 0;
			OPCoord y = 0;
			
			window->GetOrigin (x, y);
			
			//	We have now [x;y] which correponds to the absolute position of
			//	the superview with respect to the screen. We may now compute the
			//	relative position...
			
			this->drop_x -= x;
			this->drop_y -= y;
			
		} else {
			
			this->drop_x = 0;
			this->drop_y = 0;
		}
		
		this->DragEnd ();
		
		//	Dropped on an "intelligent" view. Tell it about the drop so it
		//	can handle it in a meaningful way.
		
		this->drop_cancelled = FALSE;
		
		if ( (this->drop_view.IsValid ())
		  && (this->drop_view->DragAndDrop (this->drop_box, this, OP_DDP_DRAGnDROP_COMMIT))
		  && (this->drop_cancelled == FALSE) ) {
			
			//	The drop succeeded...
			
			this->QuickReport (notify_CommitDrop, OP_REPORT_NEW);
			
		} else {
			
			this->drag_view->SetParent (this);
			this->Display ();
			this->DragOops ();
			this->QuickReport (notify_CancelDrop, OP_REPORT_NEW);
			this->DragEnd ();
		}
		
		this->drag_view = 0;
		this->drop_view = 0;
	}
}



/*
 *	The user dropped the view anywhere, just bring it back
 *	to the starting point.
 */

void
OPDragView::DragOops ()
{
	OP_METHOD (DragOops, 0);
	
	Count steps = 10;
	Count index = 0;
	
	OPCoord x   = this->box.X1 ();
	OPCoord y   = this->box.Y1 ();
	OPCoord dx  = this->start_x - x;
	OPCoord dy  = this->start_y - y;
	
	while (index++ < steps) {
		
		OPCoord rx = (int)(x + dx*index/steps);
		OPCoord ry = (int)(y + dy*index/steps);

		this->box.SetOrigin (rx, ry);
		this->grail->SetOrigin (rx, ry);
		this->grail->Redraw ();
	}
}


/*
 *	Find the view over which we are hovering. Use the real
 *	drop co-ordinates, not the center of gravity of our box.
 */

Bool
OPDragView::FindDropView ()
{
	OP_METHOD (FindDropView, 0);
	
	OPGrailWindowRef grail;
	OPWindowRef window;
	OPViewRef view;
	OPViewRef old_view = this->drop_view;
	
	if (OPGrailWindow::FindOwnWindow (this->drop_x, this->drop_y, grail, 1)) {
		
		Bool    ok = TRUE;
		OPCoord x  = 0;
		OPCoord y  = 0;
		
		grail->GetWindow (window);
		window->GetOrigin (x, y);
		
		this->drop_box = this->box;
		this->drop_box.MoveOrigin (-x, -y);
		
		x = this->drop_x - x;
		y = this->drop_y - y;
		
		//	Either there is no constraining view or we have to match both the window
		//	and the bounding rectangle...
		
		if ( (this->drop_only_in_view == FALSE)
		  || ( (this->drop_only_window == window)
			&& (this->drop_only_rect.IsInside (x, y)) ) ) {
			
			//	Find out where the user would drop the view now. If not in any subview,
			//	just select the window.
			
			if (window->Detect (x, y, view) == FALSE) {
				view = window;
			}
			
			this->drop_view = view;
			
			//	Ask the view if it wants to be substituted by another one, then check
			//	to see if the new view would accept the drop.
			
			if (this->drop_view.IsValid ()) {
				this->drop_view->DragAndDrop (this->drop_box, this, OP_DDP_DRAGnDROP_SUBST);
			}
			
			if (old_view != this->drop_view) {
				ok = this->drop_view.IsValid ()
				  && this->drop_view->DragAndDrop (this->drop_box, this, OP_DDP_DRAGnDROP_CHECK);
			}
			
			//	The drop destination might have changed... In this case, notify the
			//	previous one (if any) that it has been left. Record the new target
			//	as the new active drop destination.
			
			if (old_view != this->drop_view) {
				
				if (old_view.IsValid ()) {
					old_view->DragAndDrop (this->drop_box, this, OP_DDP_DRAGnDROP_LEAVE);
				}
				
				if (ok) {
					this->drop_view->DragAndDrop (this->drop_box, this, OP_DDP_DRAGnDROP_ENTER);
				}
				
			} else if (this->drop_view.IsValid ()) {
				
				//	The user moved above the same view as before. Just tell the drop
				//	destination that there has been a motion.
				
				this->drop_view->DragAndDrop (this->drop_box, this, OP_DDP_DRAGnDROP_MOTION);
			}
			
			return this->drop_view.IsValid ();
		}
	}
	
	//	There is no valid drop destination. If there was a previous one, just
	//	tell it that it no longer is the active drop destination.
	
	if (this->drop_view.IsValid ()) {
		this->drop_view->DragAndDrop (this->drop_box, this, OP_DDP_DRAGnDROP_LEAVE);
		this->drop_view = OPViewNull;
	}
	
	return FALSE;
}


/*
 *	Handle the possible cases at the end of a drag and drop
 *	operation :
 *
 *	1. Drop is valid.
 *	2. Drop is invalid, return to starting position.
 *	3. Drop is invalid, just give up.
 */

void
OPDragView::ButtonRelease (OPEventIn event)
{
	OP_METHOD (ButtonRelease, 1);
	OP_ARG (1, OPEventIn, event);
	
	OPDragViewRef self = this;
	
	//	First, notify of latest drag operation, so that the window
	//	is positioned correctly.
	
	this->DragNotify (event);
	this->dragging = FALSE;
	this->is_hot_spot_ok = FALSE;
	
	OPEvent::GrabWindowEvents (this, FALSE);
	
	if (this->auto_stop) {
		
		if ( (this->drop_view.IsValid ())
		  && (this->drop_view->DragAndDrop (this->drop_box, this, OP_DDP_DRAGnDROP_CHECK)) ) {
			this->DragDrop ();
			this->Close ();
			self->Delete ();
			return;
		}
		
		if (this->auto_oops) {
			
			this->DragOops ();
			this->QuickReport (notify_CancelDrop, OP_REPORT_NEW);
			
		} else {
			
			this->DragOops ();
			this->DragForget ();
			this->QuickReport (notify_CancelDrop, OP_REPORT_NEW);
			this->Close ();
			self->Delete ();
		}
	}
}


/*
 *	When the cursor is being moved around, the view is being
 *	notified and so we can follow it by changing the window
 *	origin.
 */

void
OPDragView::DragNotify (OPEventIn event)
{
	OP_METHOD (DragNotify, 1);
	OP_ARG (1, OPEventIn, event);

	if (this->dragging == FALSE) {
		
		//	First time dragging occurs. Tell the event handler that we will be
		//	interested in any mouse move events, even those outside our window
		//	(which happen when the user moves too fast).
		
		OPEvent::GrabWindowEvents (this, TRUE);
		
		this->dragging = TRUE;
		
		if (! this->is_hot_spot_ok) {
			this->SetHotSpot (event->ReturnX (), event->ReturnY ());
		}
		
		this->start_x  = (Card32) this->box.X1 ();
		this->start_y  = (Card32) this->box.Y1 ();
		
		return;
	}
	
	OPCoord x = 0;
	OPCoord y = 0;
	
	this->grail->GetOrigin (x, y);
	
	this->drop_x = event->ReturnX () + x;
	this->drop_y = event->ReturnY () + y;
	
	x += event->ReturnX () - this->ox;
	y += event->ReturnY () - this->oy;
	
	this->box.SetOrigin (x, y);
	this->grail->SetOrigin (x, y);
	this->grail->Redraw ();
	
	this->FindDropView ();
}


/*
 *	Handle the drop operation... This method is used as a
 *	call-back for validating a drop.
 */

void
OPDragView::CommitDrop (OPObjectIn target)
{
	OP_METHOD (CommitDrop, 1);
	OP_ARG (1, OPObjectIn, target);
	
	if ( (this->auto_embed)
	  && (target->IsKindOf (OP_RUNTIME (OPView))) ) {
		
		OPViewRef drop = target;
		
		if ( (drop->ReturnState () & OP_VIEW_OK_UI_EDIT)
		  && (this->drag_object.IsInvalid ()) ) {
			
			//	Dropped on a user editable view. Just paste the view into the
			//	drop destination.
			
			OPGridViewRef grid;
			OPWindowRef window;
			
			OPCoord x  = this->drop_box.X1 ();
			OPCoord y  = this->drop_box.Y1 ();
			OPCoord dx = 0;
			OPCoord dy = 0;
			
			this->drag_view->GetSize (dx, dy);
			
			if ( (drop->GetWindow (window))
			  && (window->GetAttribute ("BuilderGrid", grid)) ) {
				
				OPCoord x1 = x;
				OPCoord y1 = y;
				OPCoord x2 = x + dx;
				OPCoord y2 = y + dy;
				OPCoord min_dx;
				OPCoord min_dy;
				
				grid->AlignOnGrid (x1, y1);
				grid->AlignOnGrid (x2, y2);
				this->drag_view->GetMinSize (min_dx, min_dy);
				
				x  = x1;
				y  = y1;
				dx = x2 - x1; if (dx < min_dx) dx = min_dx;
				dy = y2 - y1; if (dy < min_dy) dy = min_dy;
			}
			
			this->drag_view->SetParent (drop);
			this->drag_view->SetOrigin (x, y);
			this->drag_view->SetSize (dx, dy);
			this->drag_view->Display ();
			
		} else {
			
			this->drop_cancelled = TRUE;
		}
		
		return;
	}
	
	if (this->auto_embed == FALSE) {
		this->notifying_drop = TRUE;
		this->QuickReport (notify_CommitSmartDrop, OP_REPORT_NEW);
		this->notifying_drop = FALSE;
	}
}


/*
 *	Handle a connection operation. This is not implemented in
 *	this class.
 */

void
OPDragView::CommitConnect (OPObjectIn target)
{
	OP_METHOD (CommitConnect, 1);
	OP_ARG (1, OPObjectIn, target);
	
	//	NIY
}


/*
 *	Handle the substitution of the current drop target by another
 *	object...
 */

void
OPDragView::SubstituteTarget (OPObjectIn target)
{
	OP_METHOD (SubstituteTarget, 1);
	OP_ARG (1, OPObjectIn, target);
	
	if (target->IsKindOf (OP_RUNTIME (OPView))) {
		this->drop_view = target;
	}
}



/*
 *	Return the position of the drop point relative to
 *	the window where the drop would happen.
 */

Bool
OPDragView::GetDropOrigin (OPCoordOut x, OPCoordOut y) const
{
	OP_METHOD (GetDropOrigin, 2);
	OP_ARG (1, OPCoordOut, x);
	OP_ARG (2, OPCoordOut, y);
	
	if (this->notifying_drop) {
		
		x = this->drop_x;
		y = this->drop_y;
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPDragView::GetDropView (OPViewOut view) const
{
	OP_METHOD (GetDropView, 1);
	OP_ARG (1, OPViewOut, view);
	
	if (this->notifying_drop) {
		
		view = this->drop_view;
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPDragView::GetDropBox (OPRectOut box) const
{
	OP_METHOD (GetDropBox, 1);
	OP_ARG (1, OPRectOut, box);
	
	if (this->notifying_drop) {
		
		box = this->drop_box;
		
		return TRUE;
	}
	
	return FALSE;
}

