/*
 *	filerd.cxx
 *
 *	The FileRead class is a specialized stream used to read data from a file.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/stream.h>
#include <OPaC/new.h>
#include <OPaC/string.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPFileRead);

/****************************************************************/

/*
 *	The File Read Stream provides an interface for reading
 *	data sequentially from a file.
 *
 *	The file must exist when it is opened !
 */

OPFileRead::OPFileRead ()
{
	OP_METHOD (OPFileRead, 0);
}



/*
 *	Close the file if it hasn't been done yet.
 */

OPFileRead::~OPFileRead ()
{
	OP_METHOD (~OPFileRead, 0);
	
	this->file.Close ();
}

	

/*
 *	Don't store anything more.
 */

void
OPFileRead::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPFileRead>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << OPStorageSync::end;
}



/*
 *	Don't restore anything more.
 */

void
OPFileRead::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> OPStorageSync::end;
}


/*
 *	Open the file for reading from it. The name has to be a
 *	physical, OS dependent, file name.
 *
 *	Throws :
 *
 *	- OPExcStream (CANNOT_OPEN) if the open failed.
 */

void
OPFileRead::Open (OPStringIn name)
{
	OP_METHOD (Open, 1);
	OP_ARG (1, OPStringIn, name);
	
	this->SetName (name);
	this->Open ();
}

void
OPFileRead::Open ()
{
	OP_METHOD (Open, 0);
	
	this->file.Close ();
	
	this->is_open = this->file.Open (this->name, OPaC_File::OPEN_READ);
	this->index   = 0;
	
	if (this->is_open == FALSE) {
		throw OPExcStream (OPExcStream::CANNOT_OPEN);
	}
}



/*
 *	Close the file. This will never throw any kind of
 *	exception.
 */

void
OPFileRead::Close ()
{
	OP_METHOD (Close, 0);
	
	this->file.Close ();
	this->index   = 0;
	this->is_open = FALSE;
}



/*
 *	Read data from the file.
 *
 *	Throws :
 *
 *	- OPExcStream (NO_READ) if the file hasn't been opened.
 */

Count
OPFileRead::Read (void* buffer, Count size)
{
	OP_METHOD (Read, 2);
	OP_ARG (1, void*, buffer);
	OP_ARG (2, Count, size);
	
	if (this->is_open == FALSE) {
		throw OPExcStream (OPExcStream::NO_READ);
	}
	
	Count read = this->file.Read (buffer, size);
	this->index += read;
	
	return read;
}



/*
 *	Skip data in the file.
 *
 *	Throws :
 *
 *	- OPExcStream (READ_ERROR) if the end of file has been
 *	  reached while skipping data (this is not guaranteed
 *	  on every platform).
 *
 *	- OPExcStream (NO_READ) if the file hasn't been opened.
 */

Int64
OPFileRead::Skip (Int64 size)
{
	OP_METHOD (Skip, 1);
	OP_ARG (1, Int64, size);
	
	if (this->is_open == FALSE) {
		throw OPExcStream (OPExcStream::NO_SKIP);
	}
	
	if (this->file.SeekRel (size) == FALSE) {
		throw OPExcStream (OPExcStream::READ_ERROR);
	}
	
	this->index += size;
	
	return size;
}

