/*
 *	filestream.cxx
 *
 *	The File Stream class provides the base for a re-openable
 *	file read and file write class.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/new.h>
#include <OPaC/stream.h>
#include <OPaC/string.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_ABSTRACT_CLASS (OPFileStream);

/****************************************************************/

/*
 *	The file stream maintains an index into the file and a
 *	flag used to determine if the file is open or not.
 */

OPFileStream::OPFileStream ()
{
	OP_METHOD (OPFileStream, 0);
	
	this->index   = 0;
	this->is_open = FALSE;
}



/*
 *	Nothing to do on destruction.
 */

OPFileStream::~OPFileStream ()
{
	OP_METHOD (~OPFileStream, 0);
}

	

/*
 *	Store the state of the file stream.
 */

void
OPFileStream::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPFileStream index is_open>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->index << this->is_open;
	storage << OPStorageSync::end;
}



/*
 *	Restore the state of the file stream but don't re-open
 *	the stream. Let the super-class handle it as needed.
 */

void
OPFileStream::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->index >> this->is_open;
	storage >> OPStorageSync::end;
}

/****************************************************************/

/*
 *	Re-open a stream after it has been restored (if it was
 *	open when stored).
 */

void
OPFileStream::Awake ()
{
	OP_METHOD (Awake, 0);
	
	super::Awake ();
	
	Bool  is_open = this->is_open;
	Int64 index   = this->index;
	
	this->is_open = FALSE;
	this->index   = 0;
	
	if (is_open) {
		this->Open ();
		this->Skip (index);
	}
}

/****************************************************************/

Bool
OPFileStream::SetName (OPStringIn name)
{
	OP_METHOD (SetName, 1);
	OP_ARG (1, OPStringIn, name);
	
	this->SetStreamName (name);
	
	return TRUE;
}

OPStringRet
OPFileStream::ReturnName () const
{
	OP_METHOD (ReturnName, 0);
	return this->name;
}

Bool
OPFileStream::GetName (OPStringOut name) const
{
	OP_METHOD (GetName, 1);
	OP_ARG (1, OPStringOut, name);
	
	if (this->name.IsValid ()) {
		this->name->GetString (name);
		return TRUE;
	}
	
	name = OPStringNull;
	
	return FALSE;
}

/****************************************************************/
