/*
 *	grail.cxx
 *
 *	GRAphical Interface Layer for building graphical interface
 *	elements. This is part should not be modified when the OPaC
 *	system gets adapted to a new target.
 *	
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_ABSTRACT_CLASS (OPGrailPort);

/****************************************************************/


/*
 *	Initialises a Grail Port and allocate some of the
 *	private structure fields (eg. a region for clipping).
 */

OPGrailPort::OPGrailPort ()
{
	OP_METHOD (OPGrailPort, 0);
	
	this->surface    = 0;
	this->dash_array = 0;
	this->dash_num   = 0;
	
	this->font_man   = OPFontMan::Current ();
	this->adorner    = 0;									//	lazy allocation...
	
	OP_ASSERT (this->font_man.IsValid ());
	
	this->rect_x1 = this->rect_y1 = 0;
	this->rect_x2 = this->rect_y2 = 0;
	this->clip_x1 = this->clip_y1 = 0;
	this->clip_x2 = this->clip_y2 = 0;
	
	this->line_style = OP_LINE_SOLID;			//	default is solid line style
	this->font       = 0;
	this->fore_color = OPColorPixel (0.0, 0.0, 0.0);	//	black
	this->back_color = OPColorPixel (1.0, 1.0, 1.0);	//	white
	
	this->is_ready   = FALSE;					//	will be initialised by a subclass
	this->is_drawing = FALSE;					//	currently not drawing
}



/*
 *	Save the grail port to the archive. Is this really
 *	meaningful ?
 */

void
OPGrailPort::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to store Grail Port");
}



/*
 *	Restore the grail port from the archive. Is this
 *	really meaningful ?
 */

void
OPGrailPort::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to restore Grail Port");
}



/*
 *	Prepare the grail port for deletion.
 */

void
OPGrailPort::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->adorner.IsValid ()) {
	
		//	Just remove the adorner, because it is tightly coupled to the
		//	GrailPort. Be careful not to call Delete for the GrailPort from
		//	the adorner (in fact, it would not harm, just be useless).
		
		OPAdornerRef adorn = this->adorner;
		
		this->adorner = 0;
		adorn->Delete ();
	}
}


/*
 *	Records the adorner. Usually, this will only be called
 *	if an already existing adorner wants to subscribe.
 */

void
OPGrailPort::InitAdorner (OPAdornerIn adorner)
{
	OP_METHOD (InitAdorner, 1);
	OP_ARG (1, OPAdornerIn, adorner);
	this->adorner = adorner;
}



/*
 *	Start drawing into a specified rectangle. The subclass
 *	has to initialise properly the "info->hdc" field so
 *	that the display context has been validated.
 */

void
OPGrailPort::BeginDraw (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (BeginDraw, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	this->BeginDraw (x1, y1, x2, y2);
}

void
OPGrailPort::BeginDraw (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (BeginDraw, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->is_ready == TRUE);
	OP_ASSERT (this->is_drawing == FALSE);
	OP_ASSERT ((x2 >= x1) && (y2 >= y1));
	
	this->rect_x1 = x1;
	this->rect_y1 = y1;
	this->rect_x2 = x2;
	this->rect_y2 = y2;
	
	//	Restore standard line style, properties and font.
	
	this->is_drawing = TRUE;
	this->SetLineStyle (OP_LINE_SOLID);
	this->SetFont (this->font);
	this->ResetClipRect ();
}



/*
 *	The drawing is finished. The subclass will have to set
 *	the drawn part as dirty, if double buffering.
 */

void
OPGrailPort::EndDraw ()
{
	OP_METHOD (EndDraw, 0);
	this->is_drawing = FALSE;
}



/*
 *	Flush the output buffer to synchronize with the screen.
 */

void
OPGrailPort::Flush ()
{
	OP_METHOD (Flush, 0);
}





/*
 *	Return the current clipping rectangle. Default is same
 *	as "BeginDraw" region.
 */

void
OPGrailPort::GetClipRect (OPCoordOut x1, OPCoordOut y1,
						  OPCoordOut x2, OPCoordOut y2) const
{
	OP_METHOD (GetClipRect, 4);
	OP_ARG (1, OPCoordOut, x1);
	OP_ARG (2, OPCoordOut, y1);
	OP_ARG (3, OPCoordOut, x2);
	OP_ARG (4, OPCoordOut, y2);
	
	x1 = this->clip_x1;
	y1 = this->clip_y1;
	x2 = this->clip_x2;
	y2 = this->clip_y2;
}

void
OPGrailPort::GetClipRect (OPRectOut rect) const
{
	OP_METHOD (GetClipRect, 1);
	OP_ARG (1, OPRectOut, rect);
	
	rect = OPRect (this->clip_x1, this->clip_y1, this->clip_x2, this->clip_y2);
}



/*
 *	Return a pointer to the font manager associated to this
 *	Grail Port.
 */

const OPFontManRef&
OPGrailPort::ReturnFontMan () const
{
	OP_METHOD (ReturnFontMan, 0);
	return this->font_man;
}



/*
 *	Return (and allocate if not already done) the Adorner
 *	associated to the port. This can be used by the widgets
 *	to draw some GUI elements.
 */

const OPAdornerRef&
OPGrailPort::ReturnAdorner ()
{
	OP_METHOD (ReturnAdorner, 0);
	
	if (this->adorner.IsValid () == FALSE) {
		this->adorner = OP_NEW_FROM_ZONE (OPaC_sys_zone, OPAdorner);
		this->adorner->Initialise (this);
	}
	
	return this->adorner;
}



/*
 *	Define the foreground color used when drawing
 *	monochrome objects (lines, circles, text...) and
 *	filling.
 */

void
OPGrailPort::SetForeColor (OPColorPixelIn pixel)
{
	OP_METHOD (SetForeColor, 1);
	OP_ARG (1, OPColorPixelIn, pixel);
	
	if (pixel == this->fore_color)
		return;
	
	this->fore_color = pixel;
	this->SetLineStyle (this->line_style);
}



/*
 *	Define the background color.
 */

void
OPGrailPort::SetBackColor (OPColorPixelIn pixel)
{
	OP_METHOD (SetBackColor, 1);
	OP_ARG (1, OPColorPixelIn, pixel);
	
	this->back_color = pixel;
}


/*
 *	Set up a line style. This will (on Windows) take in account
 *	the foreground color for the line style.
 */

void
OPGrailPort::SetLineStyle (OPLineStyle style)
{
	OP_METHOD (SetLineStyle, 1);
	OP_ARG (1, OPLineStyle, style);
	OP_ASSERT (this->is_drawing);
	
	static const Card16 dash_dot[2]  = { 1, 1 };
	static const Card16 dash_dash[2] = { 3, 3 };
	
	switch (style) {
		case OP_LINE_SOLID:	this->dash_array = 0; this->dash_num = 0; break;
		case OP_LINE_DOT: this->dash_array = dash_dot; this->dash_num = 2; break;
		case OP_LINE_DASH: this->dash_array = dash_dash; this->dash_num = 2; break;
	}
	
	this->line_style = style;
}



/*
 *	Set up a font. If the specified font pointer is 0, use
 *	a default, fixed-pitch, system font.
 */

void
OPGrailPort::SetFont (OPGrailFontIn font)
{
	OP_METHOD (SetFont, 1);
	OP_ARG (1, OPGrailFontIn, font);
	
	if (font.IsValid ()) {
		this->font = font;
	}
}
