/*
 *	hashtab.cxx
 *
 *	The OPaC Hash Table provides a mean to store objects in an
 *	unordered way. Stored objects may be looked for quickly.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/object.h>
#include <OPaC/container.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPHashTable);

/****************************************************************/

#define	HASH_MIN_VALUE	17
#define	HASH_DEF_VALUE	29

/****************************************************************/

/*
 *	Initialise an empty hash table. The default hash table
 *	size will be set to HASH_DEF_VALUE, if not initialised
 *	explicitely.
 */

OPHashTable::OPHashTable ()
{
	OP_METHOD (OPHashTable, 0);
	
	this->max_hash = 0;
	this->array    = 0;
	
	this->cache_num   = 0;
	this->cache_index = 0;
	this->cache_e_ptr = 0;
	this->cache_elem  = 0;
}



/*
 *	Destroy the hash table : remove every element from the
 *	hash table, but do not delete them.
 */

OPHashTable::~OPHashTable ()
{
	OP_METHOD (~OPHashTable, 0);
	
	if (this->array) {
		this->FreeTable ();
	}
}



/*
 *	Prepare the hash table for destruction.
 */

void
OPHashTable::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->RemoveAll ();
	
	super::Delete ();
}



/*
 *	Initialise the hash table to the specified size. The larger
 *	the size, the faster the accesses will be. If you supply a
 *	"bad" hash table size, it will automatically be increased
 *	to a "good" size.
 */

void
OPHashTable::Initialise (Count max)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, Count, max);
	OP_ASSERT (this->array == 0);
	
	if (max == 0) max = HASH_DEF_VALUE;
	if (max < HASH_MIN_VALUE) max = HASH_MIN_VALUE;
	
	//	We'll try to make a good hash value : it won't be divisible
	//	by 2, 3, 5, 7, 11, 13... which is already a good starting point.
	
	while ( ((max % 2) == 0)
		 || ((max % 3) == 0)
		 || ((max % 5) == 0)
		 || ((max % 7) == 0)
		 || ((max % 11) == 0)
		 || ((max % 13) == 0) )
		max++;
	
	this->AllocTable (max);
}



/*
 *	Allocate a hash table. This will initialise it with zero.
 */

void
OPHashTable::AllocTable (Count num)
{
	OP_METHOD (AllocTable, 1);
	OP_ARG (1, Count, num);
	
	this->array      = OP_ALLOC_ARRAY_FROM_ZONE (this->ReturnMemZone (), Element*, num);
	this->max_hash   = num;
	this->cache_elem = 0;
	
	Count i;
	
	for (i = 0; i < num; i++) {
		this->array[i] = 0;
	}
}



/*
 *	Free the hash table contents and also the hash table itself.
 *	This does not free any stored object.
 */

void
OPHashTable::FreeTable ()
{
	OP_METHOD (FreeTable, 0);
	
	if (this->array) {
		this->RemoveAll ();
		OP_FREE_FROM_ZONE (this->ReturnMemZone (), this->array);
	}
	
	this->array       = 0;
	this->max_hash    = 0;
	this->cache_elem  = 0;
	this->num_element = 0;
}



/*
 *	Update the position cache corresponding to the specified
 *	position. This will be quite fast if the previous position
 *	was smaller than the new one, since we need not start from
 *	scratch again.
 */

Bool
OPHashTable::UpdateCache (Count pos)
{
	OP_METHOD (UpdateCache, 1);
	OP_ARG (1, Count, pos);
	
	if ( (this->array == 0)
	  || (pos >= this->num_element) ) {
		return FALSE;
	}
	
	if ( (this->cache_num > pos)
	  || (this->cache_elem == 0) ) {
		
		//	Moving backwards is not possible, so forget about the cache and
		//	start from the beginning.
		
		this->cache_num   = 0;
		this->cache_index = 0;
		this->cache_e_ptr = & this->array[0];
		this->cache_elem  = * this->cache_e_ptr;
		
		//	Skip empty elements (don't count them).
		
		while (this->cache_elem == 0) {
			this->cache_index++;
			this->cache_e_ptr = & this->array[this->cache_index];
			this->cache_elem  = * this->cache_e_ptr;
		}
	}
	
	if (this->cache_num == pos) {
		return TRUE;
	}
	
	//	Just move forwards, until the specified element has been found...
	
	pos -= this->cache_num;
	
	while (pos--) {
		
		//	Skip empty elements (don't count them).
		
		while (this->cache_elem == 0) {
			this->cache_index++;
			this->cache_e_ptr = & this->array[this->cache_index];
			this->cache_elem  = * this->cache_e_ptr;
		}
		
		//	Skip full elements (count them).
		
		this->cache_num++;
		this->cache_e_ptr = & this->cache_elem->next;
		this->cache_elem  = * this->cache_e_ptr;
	}
	
	//	The element which has been found may not be a null pointer, so
	//	skip (again) the empty elements.
	
	while (this->cache_elem == 0) {
		this->cache_index++;
		this->cache_e_ptr = & this->array[this->cache_index];
		this->cache_elem  = * this->cache_e_ptr;
	}
	
	return TRUE;
}



/*
 *	Save a hash table to the archive. This will save every
 *	element stored in the table. It is not guaranteed that
 *	the position in the hash table will be the same when
 *	restoring it...
 */

void
OPHashTable::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPHashTable max_hash { data } >");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->max_hash;
	
	Element** elem_ptr = this->array;
	Count     elem_num = this->max_hash;
		
	storage << OPStorageSync::repeat;
	
	while (elem_num--) {
		
		//	Save every element from every hash table entry.
		
		if (elem_ptr[0]) {
			
			Element* elem = elem_ptr[0];
			
			while (elem) {
				
				if (elem->data.IsValid ()) {
					storage << elem->data;
				}
				
				elem = elem->next;
			}
		}
			
		elem_ptr++;
	}
	
	storage << OPStorageSync::end_repeat;
	storage << OPStorageSync::end;
}



/*
 *	Restore a hash table : allocate it and then restore every
 *	object which belonged to it. This might increase the hash
 *	table size, if it is to small.
 */

void
OPHashTable::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	Count num = this->num_element;
	Count hash;
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> hash;
	storage >> OPStorageSync::repeat;
	
	//	A small hash table size will give poor results : so we decide to
	//	increase the hash size, if we judge we need to.
	
	if (hash < num) hash = num * 2;
	
	this->Initialise (hash);
	
	//	Read every value back again and re-hash everything into the hash
	//	table.
	
	while (num--) {
		OPObjectRef value;
		storage >> value;
		this->num_element--;
		this->JustInsert (value);
	}
	
	storage >> OPStorageSync::end_repeat;
	storage >> OPStorageSync::end;
}



/*
 *	Insert an object into the hash table. The hash index will
 *	be computed by the inserted object...
 *
 *	You might wonder why I didn't choose the address as a hash
 *	index. Why, that would yield a quite slower implementation
 *	of the "GetSimilar"	method.
 */

Bool
OPHashTable::JustInsert (OPObjectIn value)
{
	OP_METHOD (JustInsert, 1);
	OP_ARG (1, OPObjectIn, value);
	
	if (this->array == 0) {
		this->Initialise (0);
	}
	
	Count    index = value->ComputeHashValue (this->max_hash);
	Element* elem  = NEW_FROM_ZONE (this->ReturnMemZone ()) Element;
	
	elem->next = this->array[index];
	elem->data = value;
	
	this->array[index] = elem;
	this->num_element++;
	
	if (index <= this->cache_index) {
		
		//	Update the cache index : the index points to the same element,
		//	but the position has been incremented.
		
		if (this->cache_elem == elem->next)
			this->cache_e_ptr = & elem->next;	//	rare case : insertion replaced cache
		
		this->cache_num++;
	}
	
	return TRUE;
}



/*
 *	Public part of "JustInsert". This will tell the object
 *	to notify the hash table when it gets deleted.
 */

Bool
OPHashTable::Insert (OPObjectIn value)
{
	OP_METHOD (Insert, 1);
	OP_ARG (1, OPObjectIn, value);
	
	if (value.IsValid ()) {
		
		this->JustInsert (value);
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Remove an object from the hash table. Return TRUE if
 *	the operation has been successful.
 *
 *	Return the removed object (or 0).
 */

Bool
OPHashTable::Remove (OPObjectIn value, Count nth)
{
	OP_METHOD (Remove, 2);
	OP_ARG (1, OPObjectIn, value);
	OP_ARG (2, Count, nth);
	
	if (this->array == 0) {
		return FALSE;
	}
	
	Count     index = value->ComputeHashValue (this->max_hash);
	Element** e_ptr = & this->array[index];
	Element*  elem  = *e_ptr;
	
	while (elem) {
		
		if ( (elem->data == value)
		  && (nth-- == 0) ) {
			
			if (index <= this->cache_index) {
				
				//	Update the cache index.
				
				if (this->cache_elem == elem->next) {
					this->cache_e_ptr = e_ptr;
				}
				
				this->cache_num--;
				
				if (this->cache_elem == elem) {
					this->cache_elem  = 0;
				}
			}
			
			*e_ptr = elem->next;
			this->num_element--;
			
			DELETE (elem);
			return TRUE;
		}
		
		e_ptr = & elem->next;
		elem  = * e_ptr;
	}
	
	return FALSE;
}



/*
 *	Remove an object specified by its position in the hash
 *	table. This is usually slower than the previous method
 *	and should be used when sequential deletes are to be
 *	performed.
 *
 *	Return the removed object (or 0).
 */

Bool
OPHashTable::Remove (Count pos)
{
	OP_METHOD (Remove, 1);
	OP_ARG (1, Count, pos);
	
	if (this->UpdateCache (pos) == FALSE) {
		return FALSE;
	}
	
	//	Remove element and update the cache, so that an access to the
	//	immediately following element will be fast.
	
	OPObjectRef& value = this->cache_elem->data;
	*this->cache_e_ptr = this->cache_elem->next;
	
	//	Now, remove the referenced object from the cache, so that it
	//	will not hang around if no longer used any more...
	
	value = 0;
	
	//	Only after that, we may free the memory.
	
	DELETE (this->cache_elem);
	
	this->cache_elem = *this->cache_e_ptr;
	this->num_element--;
	
	if (this->cache_num < this->num_element) {
		while (this->cache_elem == 0) {
			this->cache_index++;
			this->cache_e_ptr = & this->array[this->cache_index];
			this->cache_elem  = * this->cache_e_ptr;
		}
	}
	
	return TRUE;
}



/*
 *	Remove every element from the hash table. The elements
 *	won't be deleted.	
 */

void
OPHashTable::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);
	
	if (this->array) {
		
		Element** elem_ptr = this->array;
		Count     elem_num = this->max_hash;
		
		while (elem_num--) {
			
			//	Remove every element from every hash table entry. This won't be really
			//	as fast as I would like, but can you think of a better way building a
			//	dynamic hash table ?
			//	
			//	Well, look at Sedgewick's if you've got spare time... and tell me !
			
			if (elem_ptr[0]) {
				
				Element* elem = elem_ptr[0];
				
				while (elem) {
					Element*     next  = elem->next;
					OPObjectRef& value = elem->data;
					
					//	Now, remove the referenced object from the element, otherwise it
					//	would hang around in memory without being referenced any more.
					
					value = 0;
					
					//	Only then, can the element be deleted.
					
					DELETE (elem);
					elem = next;
				}
				
				elem_ptr[0] = 0;
			}
			
			elem_ptr++;
		}
	}
	
	this->cache_elem  = 0;
	this->num_element = 0;
}



/*
 *	Return the nth element found in the hash table. This uses
 *	a cache to speed up sequential increasing traversals.
 *	
 *	The order of the objects stored in a hash table is not
 *	predictable (well, it could be, but it is easier to assume
 *	it is not).
 */

Bool
OPHashTable::Get (Count pos, OPObjectOut object)
{
	OP_METHOD (Get, 2);
	OP_ARG (1, Count, pos);
	OP_ARG (2, OPObjectOut, object);
	
	if (this->UpdateCache (pos)) {
		object = this->cache_elem->data;
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Search for a similar object in the hash table. This will
 *	be quite fast, since if one exists, both will have the
 *	same hash index.
 *
 *	This uses the "CompareWith" method of the Value class.
 */

Bool
OPHashTable::GetSimilar (OPObjectIn model, OPObjectOut object, Count nth)
{
	OP_METHOD (GetSimilar, 3);
	OP_ARG (1, OPObjectIn, model);
	OP_ARG (2, OPObjectOut, object);
	OP_ARG (3, Count, nth);
	
	if (this->array == 0) {
		return FALSE;
	}
	
	Count    index = model->ComputeHashValue (this->max_hash);
	Element* elem  = this->array[index];
	
	while (elem) {
		
		if ( (elem->data->CompareWith (model) == OP_COMP_THIS_SAME)
		  && (nth-- == 0) ) {
			object = elem->data;
			return TRUE;
		}
		
		elem = elem->next;
	}
	
	return FALSE;
}



/*
 *	This looks for the specified object in the hash table. If
 *	an exact match occurs (same pointers), it will return TRUE.
 */

Bool
OPHashTable::LookFor (OPObjectIn value)
{
	OP_METHOD (LookFor, 1);
	OP_ARG (1, OPObjectIn, value);
	
	if (this->array == 0) {
		return FALSE;
	}
	
	Count    index = value->ComputeHashValue (this->max_hash);
	Element* elem  = this->array[index];
	
	while (elem) {
		
		if (elem->data == value) {
			return TRUE;
		}
		
		elem = elem->next;
	}
	
	return FALSE;
}



/*
 *	Return TRUE if the object can be found in the list,
 *	starting at position "index". The index will be
 *	updated to match the position where the object was
 *	found.
 */

Bool
OPHashTable::Find (OPObjectIn value, CountInOut index)
{
	OP_METHOD (Find, 2);
	OP_ARG (1, OPObjectIn, value);
	OP_ARG (2, CountInOut, index);
	
	while (this->UpdateCache (index)) {
		if (value == this->cache_elem->data) {
			return TRUE;
		}
		index++;
	}
	
	return FALSE;
}


