/*
 *	listree.cxx
 *
 *	The List Tree represents a tree using linked lists. The tree
 *	implements multi-level push operations.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/listree.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPLisTree);

/****************************************************************/


/*
 *	Create an empty dictionary tree.
 */

OPLisTree::OPLisTree ()
{
	OP_METHOD (OPLisTree, 0);
	
	this->list = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPList);
	this->level = this;
}



/*
 *	Destroy the dictionary tree.
 */

OPLisTree::~OPLisTree ()
{
	OP_METHOD (~OPLisTree, 0);
}



/*
 *	Save the dictionary tree to the storage.
 */

void
OPLisTree::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);

	static OPStorageSync info ("<OPLisTree parent list level>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->parent;
	storage << this->list;
	storage << this->level;
	storage << OPStorageSync::end;
}



/*
 *	Restore the dictionary from the storage.
 */

void
OPLisTree::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->parent;
	storage >> this->list;
	storage >> this->level;
	storage >> OPStorageSync::end;
}


OPIteratorRef
OPLisTree::NewIterator ()
{
	OP_METHOD (NewIterator, 0);
	
	OPLisTreeIteratorRef iter = OP_NEW (OPLisTreeIterator);
	
	iter->level = this;
	iter->index = 0;
	
	return iter;
}


/*
 *	Prepare the dictionary tree for its destruction.
 */

void
OPLisTree::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->RemoveAll ();
	this->level = OPLisTreeNull;
	
	super::Delete ();
}




Bool
OPLisTree::Insert (OPObjectIn data)
{
	OP_METHOD (Insert, 1);
	OP_ARG (1, OPObjectIn, data);
	
	this->num_element++;
	return this->level->list->PushTail (data);
}

Bool
OPLisTree::Remove (OPObjectIn data, Count nth)
{
	OP_METHOD (Remove, 2);
	OP_ARG (1, OPObjectIn, data);
	OP_ARG (2, Count, nth);
	
	return FALSE;
}

Bool
OPLisTree::Remove (Count index)
{
	OP_METHOD (Remove, 1);
	OP_ARG (1, Count, index);
	
	return FALSE;
}

void
OPLisTree::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);
	
	OPObjectRef object;
	
	while (this->list->PopHead (object)) {
		if ( (object.IsValid ())
		  && (object->IsKindOf (OP_RUNTIME (OPLisTree))) ) {
			OPLisTreeRef level = object;
			if (level->parent == this) {
				level->RemoveAll ();
				level->parent = OPLisTreeNull;
			}
		}
	}
	
	this->num_element = 0;
}


Bool
OPLisTree::Get (Count index, OPObjectOut object)
{
	OP_METHOD (Get, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, object);
	
	return FALSE;
}

Bool
OPLisTree::GetSimilar (OPObjectIn model, OPObjectOut data, Count nth)
{
	OP_METHOD (GetSimilar, 3);
	OP_ARG (1, OPObjectIn, model);
	OP_ARG (2, OPObjectOut, data);
	OP_ARG (3, Count, nth);
	
	return FALSE;
}

Bool
OPLisTree::LookFor (OPObjectIn value)
{
	OP_METHOD (LookFor, 1);
	OP_ARG (1, OPObjectIn, value);
	
	return FALSE;
}

Bool
OPLisTree::Find (OPObjectIn value, CountInOut index)
{
	OP_METHOD (Find, 2);
	OP_ARG (1, OPObjectIn, value);
	OP_ARG (2, CountInOut, index);
	
	return FALSE;
}

Bool
OPLisTree::InsertNewLevel ()
{
	OP_METHOD (InsertNewLevel, 0);
	
	OPLisTreeRef level = OP_NEW (OPLisTree);
	
	level->parent = this->level;
	this->level->list->PushTail (level);
	this->level = level;
	this->num_element++;
	
	return TRUE;
}

Bool
OPLisTree::InsertEndLevel ()
{
	OP_METHOD (InsertEndLevel, 0);
	
	if (this->level == this) {
		return FALSE;
	}
	
	this->level->list->PushTail (this->level->parent);
	this->level = this->level->parent;
	this->num_element++;
	
	return TRUE;
}

Bool
OPLisTree::InsertMark ()
{
	OP_METHOD (InsertMark, 0);
	
	this->level->list->PushTail (this->level);
	this->num_element++;
	
	return TRUE;
}

Bool
OPLisTree::Rewind ()
{
	OP_METHOD (Rewind, 0);
	
	this->level = this;
	return TRUE;
}

OPLisTree::Type
OPLisTree::PeekHeadType () const
{
	OP_METHOD (PeekHeadType, 0);
	
	OPObjectRef object;
	
	if (this->level->list->Get (0, object)) {
		
		if ( (object.IsValid ())
		  && (object->IsKindOf (OP_RUNTIME (OPLisTree))) ) {
			
			OPLisTreeRef level = object;
			
			if (level == this->level) {
				return OP_LIST_MARK;
			}
			if (level == this->level->parent) {
				return OP_LIST_END_LEVEL;
			}
			if (level->parent == this->level) {
				return OP_LIST_NEW_LEVEL;
			}
		}
		
		return OP_LIST_DATA;
	}
	
	return OP_LIST_EMPTY;
}

OPLisTree::Type
OPLisTree::PopHead (OPObjectOut object)
{
	OP_METHOD (PopHead, 1);
	OP_ARG (1, OPObjectOut, object);
	
	if (this->level->list->Get (0, object)) {
		
		if ( (object.IsValid ())
		  && (object->IsKindOf (OP_RUNTIME (OPLisTree))) ) {
			
			OPLisTreeRef level = object;
			
			if (level == this->level) {
				this->level->list->PopHead (object);
				object = OPObjectNull;
				this->num_element--;
				return OP_LIST_MARK;
			}
			if (level == this->level->parent) {
				this->level->list->PopHead (object);
				this->level = this->level->parent;
				this->level->list->PopHead (object);
				object = OPObjectNull;
				this->num_element--;
				return OP_LIST_END_LEVEL;
			}
			if (level->parent == this->level) {
				this->level = level;
				object = OPObjectNull;
				this->num_element--;
				return OP_LIST_NEW_LEVEL;
			}
		}
		
		this->level->list->PopHead (object);
		this->num_element--;
		return OP_LIST_DATA;
	}
	
	object = OPObjectNull;
	
	return OP_LIST_EMPTY;
}


Bool
OPLisTree::Clone (OPLisTreeOut tree, OPLisTreeIn parent) const
{
	OP_METHOD (Clone, 2);
	OP_ARG (1, OPLisTreeOut, tree);
	OP_ARG (2, OPLisTreeIn, parent);
	
	tree = OP_NEW (OPLisTree);
	tree->parent = parent;
	
	Count        index = 0;
	OPObjectRef  object;
	
	while (this->list->Get (index++, object)) {
		if ( (object.IsValid ())
		  && (object->IsKindOf (OP_RUNTIME (OPLisTree))) ) {
			
			OPLisTreeRef level = object;
			OPLisTreeRef that  = const_cast<OPLisTree*> (this);
			
			if (level == that) {				//	mark
				tree->list->PushTail (tree);
			} else if (level == this->parent) {	//	end
				tree->list->PushTail (tree->parent);
			} else if (level->parent == that) {	//	new
				OPLisTreeRef new_tree;
				level->Clone (new_tree, tree);
				tree->list->PushTail (new_tree);
			} else {
				tree->list->PushTail (object);
			}
		} else {
			tree->list->PushTail (object);
		}
	}
	
	tree->num_element = this->num_element;
	
	return TRUE;
}
