/*
 *	object.cxx
 *
 *	This file implements the base class used by every OPaC class. The
 *	run-time information class is implemented in rtclass.cxx !
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/atom.h>
#include <OPaC/container.h>
#include <OPaC/convert.h>
#include <OPaC/dataslot.h>
#include <OPaC/dictree.h>
#include <OPaC/new.h>
#include <OPaC/object.h>
#include <OPaC/opacirt.h>
#include <OPaC/string.h>
#include <OPaC/typeinfo.h>
#include <OPaC/values.h>

#include <OPaC/notmacro.h>

#include <string.h>

OP_IMPLEMENTATION_FILE;

/****************************************************************/

/*
 *	When the following symbol is defined, it is considered
 *	to be ok to restore an atom and use it immediately
 *	afterwards, without having to wait till the "Awake"
 *	method is being called.
 */

#define	SAVE_CROSS_OBSERVER		0
#define	MAX_OBSERVERS			100

/****************************************************************/

Bool OPObjectRef::report_creation   = TRUE;

/****************************************************************/

/*
 *	Description of an observer : an instance, a name
 *	and an ID is enough to do the job.
 */

struct Observer
{
	Observer*			next;					//	link to next observer
	Observer*			prev;					//	link to previous observer
	const OPaC_Target*	target;					//	observer full description
	OPObjectRef	 		observing;				//	observer instance (the one who is observing)
	OPAtomRef			event_name;				//	observer "event" name
	OPAtomRef			actor_name;				//	name of actor/observer
	Bool				is_actor;				//	TRUE => target is an actor, otherwise an observer
	Bool				is_volatile_observer;	//	TRUE => don't save observer
	Card32		 		id;						//	observer id
	Card32				same_names;				//	observer count with the same name
	
	Observer*			x_next;					//	link to next cross-observer
	Observer*			x_prev;					//	link to previous cross-observer
	OPObjectRef			x_observed;				//	observer instance (the one who is observed)
};

/*
 *	Description of the additional data record.
 */

struct OPObject::MoreInfo
{
	Observer*	 	obs_head;					//  link to first observer
	Observer*	 	obs_tail;					//	link to last observer
	Observer*	 	obs_cache;					//	link to cached observer
	Card32			obs_count;					//	# observers currently linked
	Card32			obs_volatile_count;			//	# observers declared as volatile
	Observer*		obs_pend;					//	link to pending observer
	Observer*		x_head;						//	link to first cross-observer
	Observer*		x_tail;						//	link to last cross-observer
	Card32			x_count;					//	# cross-observers currently linked
	
	OPDictRef		attributes;					//	dictionary of attributes (if any)
	OPHashTableRef	registered;					//	registered as target of objects
	OPObjectRef		unknown;					//	tree of unknown restored data
};

/****************************************************************/

// OP_PUBLISH_SLOT_READ (OPObject, self, ReturnSelf, "self reference", OPObject);

static OPValueRef
__rd_acc_OPObject_104 (const OPObject* obj)
{
	OPObject* instance = const_cast<OPObject*>(obj);
	return (OPValueRef)(instance);
}

static void
__collect_OPObject_104 (OPRuntime* notify)
{
	OPaC_PubDataDef pdd = { 0 };
	pdd.runtime      = & OP_RUNTIME (OPObject);
	pdd.var_name     = "self";
	pdd.var_desc     = "self reference";
	pdd.var_type     = OPObjectTypeDef;
	pdd.is_read_ok   = TRUE;
	pdd.is_write_ok  = FALSE;
	pdd.access_read  = __rd_acc_OPObject_104;
	pdd.access_write = 0;
	notify->PublishData (pdd);
}

static OPaC_PubDataExport
__export_OPObject_104 (__collect_OPObject_104);

/****************************************************************/

OP_NOTIFY_EVENT (OPObject, Changed);
OP_NOTIFY_EVENT (OPObject, ObjectAlive);

OP_NOTIFY_CHANGE (OPObject, self, ObjectAlive);

/****************************************************************/

/*
 *	The OP_ABSTRACT_CLASS macro would produce the following
 *	source, but it can't be used for the base class, since
 *	it hasn't any superclass.
 */

OPRuntimeRef OPObject_runtime (0);

const OPRuntimeRef&
OPObject::ReturnRuntime () const
{
	return OPObject_runtime;
}

const OPRuntimeRef&
OPObjectRef::ReturnRuntime ()
{
	return OPObject_runtime;
}

Bool
OPObject::IsAbstractClass () const
{
	return TRUE;
}

Bool
OPObject::IsSpecialClass () const
{
	return FALSE;
}

void
OPObject::_Destroy ()
{
	this->~OPObject ();
}

void
OPObject::Construct (void* ptr, Card32 size)
{
	OP_ABORT ("Cannot instanciate OPObject");
	OP_USE (ptr);
	OP_USE (size);
}

void
OPObject::InitConstruct (void* ptr, Card32 size)
{
	OP_ABORT ("Cannot instanciate OPObject");
	OP_USE (ptr);
	OP_USE (size);
}

/****************************************************************/

Count			OPObject::global_report_id = 0;
OPaC_DebugHook*	OPObject::debug_hook       = 0;

/****************************************************************/

OPObjectRef
OPaC_ValueToOPObject (OPValueIn val)
{
	OPObjectRef _temp_;
	_temp_ = val;
	return _temp_;
}

OPStorageIn
operator >> (OPStorageIn storage, OPObjectOut ref)
{
	storage->RestoreObject (ref, OP_RUNTIME (OPObject));
	return storage;
}

static const LatinChar class_OPObject_name[] = "OPObject";
static OPaC_RuntimeInfo g_runtime_info (OPObject_runtime, 0, class_OPObject_name,
										__OPaC_src_file, sizeof (OPObject),
										OPObject::Construct, OPObject::InitConstruct,
										TRUE, FALSE);

const OPObjectRef OPObjectNull = 0;

OPaC_TypeDef  OPObject_type  = { "OPObject", 0, OP_TYPE_OBJECT, &(OPObjectRef::ReturnRuntime ()), 0 };
OPTypeDef	  OPObjectTypeDef = & OPObject_type;
OPaC_TypeInfo OPObject_reg_type (&OPObject_type);

OPaC_TypeDef  OPObjectArray_type = { "OPObjectArray", 0, OP_TYPE_ARRAY, &(OPArrayRef::ReturnRuntime ()), 0 };
OPTypeDef     OPObjectArrayTypeDef = & OPObjectArray_type;
OPaC_TypeInfo OPObjectArray_reg_type (&OPObjectArray_type);

/****************************************************************/

static void
check_consistency (OPObject::MoreInfo* more)
{
	Observer* iter = more->obs_head;
	Count     num  = more->obs_count;
	
	while (num > 1) {
		iter = iter->next;
		num--;
	}
	
	OP_ASSERT (iter == more->obs_tail);
	
	iter = more->obs_tail;
	num  = more->obs_count;
	
	while (num > 1) {
		iter = iter->prev;
		num--;
	}
	
	OP_ASSERT (iter == more->obs_head);
}

/*
 *	Remove an observer from the additional information record.
 */

static void
observer_remove (OPMemZone* zone, OPObject::MoreInfo* more, OPObject::MoreInfo* xmore, Observer* obs)
{
	OP_ASSERT (more);
	OP_ASSERT (xmore);
	OP_ASSERT (obs);
	
	if (obs == more->obs_cache) {
		more->obs_cache = 0;
	}
	
	//	The observer record is attached to the observed object, but also to the
	//	observing object. We will first remove the record from the list of that
	//	one, then only remove it from the observed object.
	
	(obs->x_next ? obs->x_next->x_prev : xmore->x_tail) = obs->x_prev;
	(obs->x_prev ? obs->x_prev->x_next : xmore->x_head) = obs->x_next;
	
	xmore->x_count--;
	
	//	Now handle the observer list of the observed object.
	
	(obs->next ? obs->next->prev : more->obs_tail) = obs->prev;
	(obs->prev ? obs->prev->next : more->obs_head) = obs->next;
	
	more->obs_count--;
	
	if (obs->is_volatile_observer) {
		more->obs_volatile_count--;
	}
	
	check_consistency (more);
	
	if (obs->same_names) {
		Observer* iter = more->obs_head;
		while (iter) {
			if (iter->event_name == obs->event_name) {
				if (iter->same_names-- == 1)
					break;
			}
			iter = iter->next;
		}
	}
	
	obs->observing  = 0;
	obs->x_observed = 0;
	
	DELETE_FROM_ZONE (zone, obs);
}



/*
 *	Allocate an additional information record...
 */

static OPObject::MoreInfo*
alloc_more_info (OPMemZone* zone)
{
	OPObject::MoreInfo* more = NEW_FROM_ZONE (zone) OPObject::MoreInfo;
	
	more->obs_head   = 0;
	more->obs_tail   = 0;
	more->obs_cache  = 0;
	more->obs_count  = 0;
	more->obs_volatile_count = 0;
	more->obs_pend   = 0;
	more->x_head     = 0;
	more->x_tail     = 0;
	more->x_count    = 0;
	more->attributes = 0;
	more->registered = 0;
	more->unknown    = 0;
	
	return more;
}

static void
free_more_info (OPObject::MoreInfo*& more, OPMemZone* zone)
{
	if (more) {
		
		if (more->attributes.IsValid ()) {
			more->attributes->Delete ();
			more->attributes = 0;
		}
		
		if (more->registered.IsValid ()) {
			more->registered->Delete ();
			more->registered = 0;
		}
		
		if (more->unknown.IsValid ()) {
			more->unknown->Delete ();
			more->unknown = 0;
		}
		
		OP_ASSERT (more->x_head == 0);
		DELETE_FROM_ZONE (zone, more);
	}

	more = 0;
}

/****************************************************************/

struct OPaC_ObjectAllocInfo
{
	OPObject*		object;
	Address			start;
	Address			end;
};

#define	MAX_OBJ_ALLOC_INFO	10

static OPaC_ObjectAllocInfo g_obj_alloc_info[MAX_OBJ_ALLOC_INFO] = { { 0 }, { 0 } };
static Count                g_obj_alloc_index = 0;

/****************************************************************/

/*
 *	This function notifies the OPRuntime class that a ref
 *	has been created.
 */

void
OPObjectRef::ReportCreation (OPObjectRef* ref, OPRuntimeIn rt)
{
	OP_MEMBER (ReportCreation, 2);
	OP_ARG (1, OPObjectRef*, ref);
	OP_ARG (2, OPRuntimeIn, rt);
	
	Address addr = (Address)(ref);
	
	if ( (g_obj_alloc_info[g_obj_alloc_index].object)
	  && (g_obj_alloc_info[g_obj_alloc_index].start <= addr)
	  && (g_obj_alloc_info[g_obj_alloc_index].end   > addr) ) {
		ref->embedded_in_object = g_obj_alloc_info[g_obj_alloc_index].object;
	}
	
	//	OPRuntime::AddOutlet (ref, rt);
}

/****************************************************************/

/*
 *	Base class constructor. This could be used to do some
 *	debugging, but now it does just initialises everything
 *	to zero.
 */

OPObject::OPObject ()
{
	OP_METHOD (OPObject, 0);
	
	//	Do not clear the memory zone ! It has been initialised by the caller.
	
	this->more      = 0;
/*	this->mem_zone  = 0;  */
	this->current_report_id = 0;
	this->ref_count = 0;
	this->ref_head  = 0;
	this->ref_tail  = 0;
	
	if (this->mem_zone) {
		
		OP_ASSERT (this->mem_zone->latest_allocation);
		
		Address a_start = (Address)(this->mem_zone->latest_allocation->ptr);
		Address a_end   = a_start + this->mem_zone->latest_allocation->size;
		Address a_this  = (Address)(this);
		
		OP_ASSERT (a_this >= a_start);
		OP_ASSERT (a_this < a_end);
		
		if (++g_obj_alloc_index == MAX_OBJ_ALLOC_INFO) {
			g_obj_alloc_index = 0;
		}
		
		g_obj_alloc_info[g_obj_alloc_index].object = this;
		g_obj_alloc_info[g_obj_alloc_index].start  = a_start;
		g_obj_alloc_info[g_obj_alloc_index].end    = a_end;
		
	} else {
		this->mem_zone = 0;
	}
}

OPObject::OPObject (OPMemZone* zone)
{
	OP_METHOD (OPObject, 1);
	OP_ARG (1, OPMemZone*, zone);
	
	this->more      = 0;
	this->mem_zone  = zone;
	this->ref_count = 0;
	this->ref_head  = 0;
	this->ref_tail  = 0;
	
	if (this->mem_zone) {
		
		OP_ASSERT (this->mem_zone->latest_allocation);
		
		Address a_start = (Address)(this->mem_zone->latest_allocation->ptr);
		Address a_end   = a_start + this->mem_zone->latest_allocation->size;
		Address a_this  = (Address)(this);
		
		OP_ASSERT (a_this >= a_start);
		OP_ASSERT (a_this < a_end);
		
		g_obj_alloc_info[g_obj_alloc_index].object = this;
		g_obj_alloc_info[g_obj_alloc_index].start  = a_start;
		g_obj_alloc_info[g_obj_alloc_index].end    = a_end;
		
		if (++g_obj_alloc_index == MAX_OBJ_ALLOC_INFO) {
			g_obj_alloc_index = 0;
		}
		
	} else {
		this->mem_zone = 0;
	}
}


/*
 *	There is some special constructor which does nothing
 *	else than implicitely call the constructor of the
 *	contained references, thus enabling OPaC to know at
 *	run-time where to look for references...
 */

OPObject::OPObject (void* ptr, Card32 size)
{
	OP_METHOD (OPObject, 2);
	OP_ARG (1, void*, ptr);
	OP_ARG (2, Card32, size);
	
	//	Just don't do anything... The instance created that way will never be
	//	used and won't ever get deleted properly.
}



/*
 *	Like the constructor, the destructor could be used to
 *	monitor and help debugging. It is actually not used.
 */

OPObject::~OPObject ()
{
	OP_METHOD (OPObject, 0);
	
	OP_ASSERT ( (this->ref_count == 0) || (this->ref_count == 0x80000000) );
	
	MoreInfo* more  = this->more;
	MoreInfo* xmore = 0;
	
	if (g_obj_alloc_info[g_obj_alloc_index].object == this) {
		g_obj_alloc_info[g_obj_alloc_index].object = 0;
		g_obj_alloc_info[g_obj_alloc_index].start  = 0;
		g_obj_alloc_info[g_obj_alloc_index].end    = 0;
	}
	
	if (more) {
		while (more->obs_head) {
			xmore = (MoreInfo*)(more->obs_head->observing->more);
			observer_remove (this->mem_zone, more, xmore, more->obs_head);
		}
	}
	
	free_more_info (this->more, this->mem_zone);
}


/*
 *	Resolution methods, which can be used to replace an object
 *	by another one when using references.
 */

OPObject*
OPObject::Resolve ()
{
//	OP_METHOD (Resolve, 0);
	return this;
}

OPObject*
OPObject::CheckAndResolve ()
{
//	OP_METHOD (CheckAndResolve, 0);
	return this;
}

Bool
OPObject::GetUnknownData (OPObjectOut unknown) const
{
	OP_METHOD (GetUnknownData, 1);
	OP_ARG (1, OPObjectOut, unknown);
	
	unknown = 0;
	
	if (this->more) {
		unknown = this->more->unknown;
	}
	
	return unknown.IsValid ();
}

Bool
OPObject::SetUnknownData (OPObjectIn unknown)
{
	OP_METHOD (SetUnknownData, 1);
	OP_ARG (1, OPObjectIn, unknown);
	
	if (this->more == 0) {
		this->more = alloc_more_info (this->mem_zone);
	}
	
	this->more->unknown = unknown;
	
	return TRUE;
}


/*
 *	Store the instance to the storage class. This must
 *	first call the super-class' `StoreInstance' method,
 *	then save the data.
 */

void
OPObject::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	static OPStorageSync info ("<OPObject"
							   " x_count { <observed event_name x_observed is_actor actor_name id> }"
							   " obs_count { <observing event_name observing is_actor actor_name id> }"
							   " dict hash>");
	
	storage << info;
	storage << OPStorageSync::begin;
	
	MoreInfo* more = this->more;
	
	{
		Observer* obs = (more) ? more->x_head : 0;
		Card32    cnt = (more) ? more->x_count : 0;
		
		storage << cnt;		//	save info on observed object
		storage << OPStorageSync::repeat;
		
		while (cnt--) {
			
			storage << OPStorageSync::begin;
			storage << obs->event_name;
			storage << obs->x_observed;
			storage << obs->is_actor;
			storage << obs->actor_name;
			storage << obs->id;
			storage << OPStorageSync::end;
			
			obs = obs->x_next;
		}
		
		storage << OPStorageSync::end_repeat;
	}
	{
		Observer* obs = (more) ? more->obs_head : 0;
		Card32    cnt = (more) ? more->obs_count : 0;
		Card32  s_num = (more) ? cnt - more->obs_volatile_count : 0;
		
		storage << s_num;
		storage << OPStorageSync::repeat;
		
		while (cnt--) {
			
			if (obs->is_volatile_observer == FALSE) {
				
				storage << OPStorageSync::begin;
				storage << obs->event_name;
				storage << obs->observing;
				storage << obs->is_actor;
				storage << obs->actor_name;
				storage << obs->id;
				storage << OPStorageSync::end;
			}
			
			obs = obs->next;
		}
		
		storage << OPStorageSync::end_repeat;
	}
	
	//	If there are some attributes, save them with the object
	//	so they can be restored later on.
	
	OPDictRef      dict = 0;
	OPHashTableRef hash = 0;
	
	if ( (more)
	  && (more->attributes.IsValid ()) ) {
		dict = more->attributes;
	}
	
	if ( (more)
	  && (more->registered.IsValid ()) ) {
		hash = more->registered;
	}
	
	storage << dict << hash;
	storage << OPStorageSync::end;
}



/*
 *	Restore the instance from the storage. This must first
 *	call the super-class' `RestoreInstance' method, then
 *	restore the data.
 *
 *	When storing/restoring data, the class should always
 *	save a local version number of itself or be smart
 *	enough to recognize an older version of its saved
 *	data...
 */

void
OPObject::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);

	Card32        cnt = 0;
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> cnt;
	storage >> OPStorageSync::repeat;
	
	while (cnt--) {
		
		OPAtomRef   name;
		OPAtomRef   actor_name;
		OPObjectRef x_observed;
		Card32      id       = 0;
		Bool        is_actor = FALSE;
		
		storage >> OPStorageSync::begin;
		storage >> name >> x_observed >> is_actor >> actor_name >> id;
		storage >> OPStorageSync::end;
		
		OPObject::AttachObserver (x_observed, this, name, actor_name, id);
	}
	
	storage >> OPStorageSync::end_repeat;
	storage >> cnt;
	storage >> OPStorageSync::repeat;
	
	while (cnt--) {
		
		OPAtomRef   name;
		OPAtomRef   actor_name;
		OPObjectRef observing;
		Card32      id       = 0;
		Bool        is_actor = FALSE;
		
		storage >> OPStorageSync::begin;
		storage >> name >> observing >> is_actor >> actor_name >> id;
		storage >> OPStorageSync::end;
		
		OPObject::AttachObserver (this, observing, name, actor_name, id);
	}
	
	storage >> OPStorageSync::end_repeat;
	
	//	Attributes can be associated to any object. If there are any,
	//	we just restore them and link them thanks to the additional
	//	information record to the current object.
	
	OPDictRef      dict = 0;
	OPHashTableRef hash = 0;
	
	storage >> dict >> hash;
	storage >> OPStorageSync::end;
	
	if (dict.IsValid ()) {
		if (this->more == 0) this->more = alloc_more_info (this->mem_zone);
		this->more->attributes = dict;
	}
	
	if (hash.IsValid ()) {
		if (this->more == 0) this->more = alloc_more_info (this->mem_zone);
		this->more->registered = hash;
	}
}



/*
 *	The awake method gets called after the instance graph
 *	has been fully restored. This will allow some more
 *	initialization.
 */

void
OPObject::Awake ()
{
	OP_METHOD (Awake, 0);
#ifdef OK_TO_RESTORE_ATOM_NAME_IMMEDIATELY
#else
	MoreInfo*  more = this->more;
	OPMemZone* zone = this->mem_zone;
	
	if (more) {
		while (more->obs_pend) {
			
			Observer* obs = more->obs_pend;
			more->obs_pend = obs->next;

			this->AttachObserver (obs->observing, obs->event_name,
								  obs->actor_name, obs->id);
			
			DELETE_FROM_ZONE (zone, obs);
		}
	}
#endif
	
	this->Report (notify_ObjectAlive, OP_REPORT_NEW);
	this->ReportChange ();
}



/*
 *	Same functionality as the "Awake" method, but
 *	specifies the application which was responsible
 *	for the restoration.
 */

void
OPObject::AwakeFromApp (OPAppIn app)
{
	OP_METHOD (AwakeFromApp, 1);
	OP_ARG (1, OPAppIn, app);

	this->Awake ();
}



/*
 *	Same functionality as the "Awake" method, but
 *	specifies the object which was responsible for the
 *	restoration.
 */

void
OPObject::AwakeFromObject (OPObjectIn object)
{
	OP_METHOD (AwakeFromObject, 1);
	OP_ARG (1, OPObjectIn, object);
	
	this->Awake ();
}



/*
 *	Look if this instance belongs to the specified class.
 *	This will just look at every superclass and compare
 *	the run-time information.
 */

Bool
OPObject::IsKindOf (OPRuntimeIn what) const
{
	OP_METHOD (IsKindOf, 1);
	OP_ARG (1, OPRuntimeIn, what);
	
	OPRuntimeRef runtime = this->ReturnRuntime ();
	
	while (runtime.IsValid ()) {
		
		if (runtime->ReturnClassName () == what->ReturnClassName ()) {
			return TRUE;
		}
		
		runtime = runtime->ReturnSuper ();
	}
	
	return FALSE;
}



/*
 *	Return TRUE if the runtime class corresponds to the
 *	specified class.
 */

Bool
OPObject::IsA (OPRuntimeIn what) const
{
	OP_METHOD (IsA, 1);
	OP_ARG (1, OPRuntimeIn, what);
	
	if (this->ReturnRuntime ()->ReturnClassName () == what->ReturnClassName ()) {
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Remove every link this object has established with
 *	other objects. This has to be called on some objects
 *	which reference themselves mutually, or they would
 *	never get destroyed.
 */

void
OPObject::Delete ()
{
	OP_METHOD (Delete, 0);
	
	MoreInfo* more = this->more;
	
	if (more) {
		
		MoreInfo*  xmore = 0;
		OPMemZone* zone = this->mem_zone;
		
		while (more->obs_head) {
			xmore = (MoreInfo*)(more->obs_head->observing->more);
			observer_remove (zone, more, xmore, more->obs_head);
		}
		
		while (more->x_head) {
			xmore = (MoreInfo*)(more->x_head->x_observed->more);
			zone = more->x_head->x_observed->mem_zone;
			observer_remove (zone, xmore, more, more->x_head);
		}
	}	
	
	free_more_info (this->more, this->mem_zone);
}



/*
 *	Mark an object as "dead" : this will automatically set
 *	every referring object to a null pointer.
 *
 *	Easy, quick and dirty :-)
 */

void
OPObject::MarkAsDead ()
{
	OP_METHOD (MarkAsDead, 0);
	
	//	The object will no longer be available to the referring references. Just
	//	detach the references from the object.
	
	if (this->ref_count) {
		
		OPObjectRef* iter = this->ref_head;
		OPObjectRef* next = 0;
		
		while (iter) {
			next = iter->next;
			iter->object = 0;
			iter->next   = 0;
			iter->prev   = 0;
			iter = next;
		}
		
		this->ref_count = 0;
		this->ref_head  = 0;
		this->ref_tail  = 0;
	}
}



/*
 *	Mark an object as "dead" : this will automatically set
 *	every referring object to a null pointer; the reference
 *	specified as "keep" will not be marked as dead.
 */

void
OPObject::MarkAsDeadButOne (OPObjectIn keep)
{
	OP_METHOD (MarkAsDeadButOne, 1);
	OP_ARG (1, OPObjectIn, keep);
	
	//	The object will no longer be available to the referring references. Just
	//	detach the references from the object.
	
	if (this->ref_count) {
		
		OPObjectRef* iter  = this->ref_head;
		OPObjectRef* next  = 0;
		OPObjectRef* hit   = 0;
		
		while (iter) {
			
			if (iter == & keep) {
				hit = iter;
			}

			next = iter->next;
			
			iter->object = 0;
			iter->next   = 0;
			iter->prev   = 0;
			
			iter = next;
		}
		
		this->ref_count = 0;
		this->ref_head  = 0;
		this->ref_tail  = 0;
		
		if (hit) {
			hit->object     = this;
			this->ref_count = 1;
			this->ref_head  = hit;
			this->ref_tail  = hit;
		}
	}
}



/*
 *	Return a hash index for this instance.
 *
 *	CAUTION: This method should always be overridden,
 *			 since using the address as a hash index
 *			 doesn't make sense !
 */

Count
OPObject::ComputeHashValue (Count num) const
{
	OP_METHOD (ComputeHashValue, 1);
	OP_ARG (1, Count, num);
	OP_ASSERT (num);
	
	Address index = (Address)(this);
	return (Count)(index % num);
}


/*
 *	Compare with the specified object. This can implement
 *	an optimised comparison method (with respect to the more
 *	complete CompareWith method).
 */

Bool
OPObject::IsEqual (OPObjectIn object) const
{
	OP_METHOD (IsEqual, 1);
	OP_ARG (1, OPObjectIn, object);
	
	return (this->CompareWith (object) == OP_COMP_THIS_SAME) ? TRUE : FALSE;
}


/*
 *	Compare "this" with another object.
 *
 *	CAUTION: This method should always be overridden,
 *			 since comparing object values using the
 *			 pointers doesn't make sense !
 */

int
OPObject::CompareWith (OPObjectIn value) const
{
	OP_METHOD (CompareWith, 1);
	OP_ARG (1, OPObjectIn, value);
	
	int result = OP_COMP_THIS_SAME;
	
	if (this == value.object) {
		result = OP_COMP_THIS_SAME;
	} else if (this < value.object) {
		result = OP_COMP_THIS_BEFORE;
	} else if (this > value.object) {
		result = OP_COMP_THIS_AFTER;
	}
	
	return result;
}

/****************************************************************/


/*
 *	Report some event to the observers.
 */

Bool
OPObject::Report (OPAtomIn event, Count id)
{
	OP_METHOD (Report, 2);
	OP_ARG (1, OPAtomIn, event);
	OP_ARG (2, Count, id);
	
	this->SetCurrentReportID (id);
	
	MoreInfo* more = this->more;
	Bool      ok   = FALSE;
	Card32    num  = 0;
	
	if (more == 0) {
		return ok;
	}
	
	OPObjectRef self = this;
	Observer*   obs;
	
	obs = more->obs_cache;
	
	if ( (obs) && (obs->event_name == event) ) {
		
		if (obs->is_actor) {
			(obs->target->target_method->a_trampo) (obs->observing.object->Resolve ());
		} else {
			(obs->target->target_method->o_trampo) (obs->observing.object->Resolve (), self, obs->id);
		}
		
		return TRUE;
	}
	
	obs = more->obs_head;
	
	//	Maintain a table of the called observers. We cannot simply walk through
	//	the observer list, since it can be modified dynamically while calling the
	//	actors and observers.
	
	Observer* called[MAX_OBSERVERS];
	Count num_called = 0;
	
	while (obs) {
		
		Bool not_called = TRUE;
		
		//	Check if this is a new observer or if it has already been called before.
		
		for (Count index = 0; index < num_called; index++) {
			if (called[index] == obs) {
				not_called = FALSE;
				break;
			}
		}
		
		if ( (not_called)
		  && (obs->event_name == event) ) {
			
			if (num-- == 0) {
				
				if (ok) {
					return ok;
				}
				
				num = obs->same_names;
				ok  = TRUE;
				
				if (num == 0) {
					
					//	If this observer is the only one with the specified name, then
					//	we may insert it into the cache, so that any further report will
					//	be executed faster.
					
					more->obs_cache = obs;
				}
			}
			
			if (obs->is_actor) {
				(obs->target->target_method->a_trampo) (obs->observing.object->Resolve ());
			} else {
				(obs->target->target_method->o_trampo) (obs->observing.object->Resolve (), self, obs->id);
			}
			
			if (this->more == 0) {
				return ok;
			}
			
			//	Remember that we already called this observer. This will avoid that
			//	we call this observer twice...
			
			called[num_called++] = obs;
			
			OP_ASSERT (num_called < MAX_OBSERVERS);
			
			//	Start at the beginning, since the observer chain might have been modified
			//	(worst case : an observer has been removed from the list).
			
			obs = more->obs_head;
			
		} else {
			
			obs = obs->next;
		}
	}

	return ok;
}

Bool
OPObject::Report (const LatinChar* event, Count id)
{
	OP_METHOD (Report, 2);
	OP_ARG (1, const LatinChar*, event);
	OP_ARG (2, Count, id);
	
	OPAtomRef atom = event;
	return this->Report (atom, id);
}


Bool
OPObject::ReportChange (Count id)
{
	OP_METHOD (ReportChange, 1);
	OP_ARG (1, Count, id);
	
	this->Report (notify_Changed, id);
	
	return TRUE;
}


/*
 *	When an embedded object changes, it signals this change
 *	to the object, so that a more general reporting can be
 *	initiated.
 */

Bool
OPObject::ReportEmbeddedChange (OPObjectRef& ref, Count id)
{
	OP_METHOD (ReportEmbeddedChange, 2);
	OP_ARG (1, OPObjectRef&, ref);
	OP_ARG (2, Count, id);
	
	if (OPClassMan::IsInitOk ()) {
		this->Report (notify_Changed, id);
	}
	
	return TRUE;
}


/*
 *	Return the object's report ID, which was defined by
 *	the last change (ReportChange, ReportEmbeddedChange).
 */

Count
OPObject::GetCurrentReportID () const
{
	return this->current_report_id;
}


/*
 *	Define the current report ID for this object. This is
 *	set both by the ReportChange and ReportEmbeddedChange
 *	methods so that circular report loops can be detected
 *	and avoided.
 */

void
OPObject::SetCurrentReportID (Count id)
{
	if (id == OP_REPORT_AUTO) {
		
		this->current_report_id = OPObject::global_report_id;
		
	} else if (id == OP_REPORT_NEW) {
		
		this->current_report_id = OPObject::NewReportID ();
		
	} else {
		
		this->current_report_id = id;
	}
}

/*
 *	Return a new unique report ID. The ID may wrap around
 *	since it is a limited 32-bit quantity. The special ID
 *	values (defined as OP_REPORT_xxx) are avoided.
 */

Count
OPObject::NewReportID ()
{
	do {
		OPObject::global_report_id++;
	} while ( (OPObject::global_report_id < OP_REPORT_MIN_ID)
		   || (OPObject::global_report_id > OP_REPORT_MAX_ID) );
	
	return OPObject::global_report_id;
}

/****************************************************************/

/*
 *	An observer will be added to this object. Whenever an event
 *	is being generated using the OPObject::Report method, the
 *	corresponding event method will be invoked.
 *
 *	The methods are exported using the OP_TARGET_OBSERVER or
 *	OP_TARGET_ACTOR macros.
 */

void
OPObject::AttachObserver (OPObjectIn obj, OPAtomIn event, Card32 id)
{
	OP_METHOD (AttachObserver, 3);
	OP_ARG (1, OPObjectIn, obj);
	OP_ARG (2, OPAtomIn, event);
	OP_ARG (3, Card32, id);
	
	this->AttachObserver (obj, event, event, id);
}

void
OPObject::AttachObserver (OPObjectIn obs, const LatinChar* event, Card32 id)
{
	OP_METHOD (AttachObserver, 3);
	OP_ARG (1, OPObjectIn, obs);
	OP_ARG (2, const LatinChar*, event);
	OP_ARG (3, Card32, id);
	
	OPAtomRef atom = event;
	this->AttachObserver (obs, atom, id);
}

void
OPObject::AttachObserver (OPObjectIn obj, OPAtomIn event,
						  OPAtomIn actor, Card32 id)
{
	OP_METHOD (AttachObserver, 4);
	OP_ARG (1, OPObjectIn, obj);
	OP_ARG (2, OPAtomIn, event);
	OP_ARG (3, OPAtomIn, actor);
	OP_ARG (4, Card32, id);
	
	OPObject::AttachObserver (this, obj, event, actor, id);
}
	
void
OPObject::AttachObserver (OPObjectIn tobj, OPObjectIn obj, OPAtomIn event, OPAtomIn actor, Card32 id)
{
	OP_MEMBER (AttachObserver, 5);
	OP_ARG (1, OPObjectIn, tobj);
	OP_ARG (2, OPObjectIn, obj);
	OP_ARG (3, OPAtomIn, event);
	OP_ARG (4, OPAtomIn, actor);
	OP_ARG (5, Card32, id);
	
	const OPRuntimeRef& runtime  = obj->ReturnRuntime ();
	const OPaC_Target*  target   = 0;
	Bool                is_actor = FALSE;
	
	target = runtime->FindActor (actor);
	
	if (target) {
		is_actor = TRUE;
	} else {
		target = runtime->FindObserver (actor);
		OP_ASSERT (target);
	}
	
	if (tobj->more == 0) {
		tobj->more = alloc_more_info (tobj->mem_zone);
	}
	
	if (obj->more == 0) {
		obj->more = alloc_more_info (obj->mem_zone);
	}
	
	//	Check that the observer is unique. Avoid multiple connections
	//	with exactly the same object (this would be meaningless).
	
	MoreInfo* imore = tobj->more;
	
	if (imore) {
		
		Observer* obs = imore->obs_head;

		while (obs) {
			
			Observer* next = obs->next;
			
			if ( (obs->id == id)
			  && (obs->observing == obj)
			  && (obs->event_name == event)
			  && (obs->actor_name == actor) ) {
				
				return;
			}

			obs = next;
		}
	}
	
	//	Allocate an observer description record. This will simply store the
	//	data which has been given by the caller.
	
	MoreInfo* more  = tobj->more;
	MoreInfo* xmore = obj->more;
	Observer* obs   = (Observer*) NEW_FROM_ZONE (tobj->mem_zone) Observer;
	Observer* iter  = more->obs_head;
	
	obs->target        = target;
	obs->observing     = obj;
	obs->id            = id;
	obs->event_name    = event;
	obs->next          = 0;
	obs->prev          = more->obs_tail;
	obs->same_names    = 0;
	obs->is_actor      = is_actor;
	obs->actor_name    = actor;
	obs->x_next        = 0;
	obs->x_prev        = xmore->x_tail;
	obs->x_observed    = tobj;
	
	obs->is_volatile_observer = obj->GetAttribute ("OPaC.volatile.observer");
	
	//	Update the count of any observer which would have a matching name,
	//	so that we know if an observer exists more than once in the list.
	
	while (iter) {
		if (iter->event_name == obs->event_name) {
			iter->same_names++;
			obs->same_names++;
		}
		iter = iter->next;
	}
	
	if (more->obs_cache && more->obs_cache->same_names) {
		
		//	An observer may only reside in the cache if it is present once
		//	in the list, or we would miss the others...

		more->obs_cache = 0;
	}
	
	(obs->next ? obs->next->prev : more->obs_tail) = obs;
	(obs->prev ? obs->prev->next : more->obs_head) = obs;
	
	(obs->x_next ? obs->x_next->x_prev : xmore->x_tail) = obs;
	(obs->x_prev ? obs->x_prev->x_next : xmore->x_head) = obs;
	
	more->obs_count++;
	xmore->x_count++;
	
	if (obs->is_volatile_observer) {
		more->obs_volatile_count++;
	}
	
	check_consistency (more);
	
	tobj->Report (notify_ObjectAlive);
}

void
OPObject::AttachObserver (OPObjectIn obs, const LatinChar* event,
						  const LatinChar* actor, Card32 id)
{
	OP_METHOD (AttachObserver, 4);
	OP_ARG (1, OPObjectIn, obs);
	OP_ARG (2, const LatinChar*, event);
	OP_ARG (3, const LatinChar*, actor);
	OP_ARG (4, Card32, id);
	
	OPAtomRef atom = event;
	this->AttachObserver (obs, atom, actor, id);
}



/*
 *	Remove an observer from the list. This will no longer get
 *	notifications when the object changes.
 */

void
OPObject::DetachObserver (OPObjectIn obj, OPAtomIn event, Card32 id)
{
	OP_METHOD (DetachObserver, 3);
	OP_ARG (1, OPObjectIn, obj);
	OP_ARG (2, OPAtomIn, event);
	OP_ARG (3, Card32, id);
	
	this->DetachObserver (obj, event, event, id);
}

void
OPObject::DetachObserver (OPObjectIn obs, const LatinChar* event, Card32 id)
{
	OP_METHOD (DetachObserver, 3);
	OP_ARG (1, OPObjectIn, obs);
	OP_ARG (2, const LatinChar*, event);
	OP_ARG (3, Card32, id);
	
	OPAtomRef atom = event;
	this->DetachObserver (obs, atom, id);
}

void
OPObject::DetachObserver (OPObjectIn obj, OPAtomIn event,
						  OPAtomIn actor, Card32 id)
{
	OP_METHOD (DetachObserver, 4);
	OP_ARG (1, OPObjectIn, obj);
	OP_ARG (2, OPAtomIn, event);
	OP_ARG (3, OPAtomIn, actor);
	OP_ARG (4, Card32, id);
	
	MoreInfo* more  = this->more;
	MoreInfo* xmore = 0;
	Observer* obs;
	
	if (more) {
		
		obs = more->obs_head;

		while (obs) {
			
			Observer* next = obs->next;
			
			if ( (obs->id == id)
			  && (obs->observing == obj)
			  && (obs->event_name == event)
			  && (obs->actor_name == actor) ) {
				
				xmore = (MoreInfo*)(obs->observing->more);
				observer_remove (this->mem_zone, more, xmore, obs);

				//	NB: This removes only the first matching observer !

				break;
			}

			obs = next;
		}
	}
}

void
OPObject::DetachObserver (OPObjectIn obs, const LatinChar* event,
						  const LatinChar* actor, Card32 id)
{
	OP_METHOD (DetachObserver, 4);
	OP_ARG (1, OPObjectIn, obs);
	OP_ARG (2, const LatinChar*, event);
	OP_ARG (3, const LatinChar*, actor);
	OP_ARG (4, Card32, id);
	
	OPAtomRef atom = event;
	this->DetachObserver (obs, atom, actor, id);
}

void
OPObject::DetachObserver (OPObjectIn obj)
{
	OP_METHOD (DetachObserver, 1);
	OP_ARG (1, OPObjectIn, obj);
	
	if (this->more) {
		
		MoreInfo* xmore = 0;
		Observer* obs   = this->more->obs_head;
		
		while (obs) {
			if (obs->observing == obj) {
				xmore = (MoreInfo*)(obs->observing->more);
				observer_remove (this->mem_zone, more, xmore, obs);
				obs = this->more ? this->more->obs_head : 0;
			} else {
				obs = obs->next;
			}
		}
	}
}

void
OPObject::DetachSelfFromAll ()
{
	OP_METHOD (DetachSelfFromAll, 0);
	
	while ( (this->more)
		 && (this->more->x_head) ) {
		this->more->x_head->x_observed->DetachObserver (this);
	}
}

/****************************************************************/


/*
 *	Count the number of the specified targets. This function
 *	is being used by some OPObject::CountXXX methods and does
 *	the real job for them.
 */

static Count
count_target_by_type (OPRuntimeIn runtime, OPTargetType match)
{
	Count index = 0;
	Count total = 0;

	OPAtomRef          name;
	OPTargetType       type = OP_TARGET_IS_INVALID;
	const OPaC_Target* target = 0;

	while (runtime->GetTarget (index++, name, type, target)) {
		if (type == match) {
			total++;
		}
	}
	
	return total;
}



/*
 *	Find the target description of the specified target. This
 *	function is being used by some OPObject::GetXXX methods.
 */

static const OPaC_Target*
find_target_by_type (Count nth, OPRuntimeIn runtime, OPTargetType match, OPAtomOut name)
{
	Count index = 0;

	OPTargetType       type = OP_TARGET_IS_INVALID;
	const OPaC_Target* target = 0;

	while (runtime->GetTarget (index++, name, type, target)) {
		if (type == match) {
			if (nth-- == 0) {
				return target;
			}
		}
	}
	
	return 0;
}

/****************************************************************/


/*
 *	Look how many observers are being made available to the
 *	rest of the world.
 *
 *	These are published using OP_TARGET_IS_OBSERVER.
 */

Count
OPObject::CountObservers () const
{
	OP_METHOD (CountObservers, 0);
	return count_target_by_type (this->ReturnRuntime (), OP_TARGET_IS_OBSERVER);
}


/*
 *	Look how many actors are being made available to the rest
 *	of the world.
 *
 *	These are published using OP_TARGET_IS_ACTOR.
 */

Count
OPObject::CountActors () const
{
	OP_METHOD (CountActors, 0);
	return count_target_by_type (this->ReturnRuntime (), OP_TARGET_IS_ACTOR);
}


/*
 *	Look how many general methods are being made available to
 *	the rest of the world.
 *
 *	These are published using OP_TARGET_IS_METHOD.
 */

Count
OPObject::CountMethods () const
{
	OP_METHOD (CountMethods, 0);
	return count_target_by_type (this->ReturnRuntime (), OP_TARGET_IS_METHOD);
}


/*
 *	Look how many data slots are being made available to the
 *	rest of the world, available through GetPublicData.
 *
 *	This has to be overridden by the subclasses in order to
 *	work properly.
 */

Count
OPObject::CountPublicData () const
{
	OP_METHOD (CountPublicData, 0);
	
	Count num = 0;
	OPRuntimeRef runtime = this->ReturnRuntime ();
	
	while (runtime.IsValid ()) {
		num    += runtime->CountPubData ();
		runtime = runtime->ReturnSuper ();
	}
	
	return num;
}


/*
 *	Look how many observers have been attached to the object
 *	and which will be able to get messages using the Report
 *	method.
 */

Count
OPObject::CountAttached () const
{
	OP_METHOD (CountAttached, 0);
	return (this->more) ? this->more->obs_count : 0;
}


/****************************************************************/

Count
OPObject::CountAttributes () const
{
	OP_METHOD (CountAttributes, 0);
	if ( (this->more)
	  && (this->more->attributes.IsValid ()) ) {
		return this->more->attributes->TotalCount ();
	}
	return 0;
}

Bool
OPObject::SetAttribute (const LatinChar* key)
{
	OP_METHOD (SetAttribute, 1);
	OP_ARG (1, const LatinChar*, key);
	
	if (this->more == 0) {
		this->more = alloc_more_info (this->mem_zone);
	}
	
	OPAtomRef atom_key = key;
	
	if (this->more->attributes.IsValid () == FALSE) {
		this->more->attributes = OP_NEW_FROM_ZONE (this->mem_zone, OPDict);
	}
	
	return this->more->attributes->Insert (atom_key, OPObjectNull);
}

Bool
OPObject::SetAttribute (OPObjectIn key)
{
	OP_METHOD (SetAttribute, 1);
	OP_ARG (1, OPObjectIn, key);
	
	if (this->more == 0) {
		this->more = alloc_more_info (this->mem_zone);
	}
	
	if (this->more->attributes.IsValid () == FALSE) {
		this->more->attributes = OP_NEW_FROM_ZONE (this->mem_zone, OPDict);
	}
	
	return this->more->attributes->Insert (key, OPObjectNull);
}

Bool
OPObject::SetAttribute (const LatinChar* key, OPObjectIn value)
{
	OP_METHOD (SetAttribute, 2);
	OP_ARG (1, const LatinChar*, key);
	OP_ARG (2, OPObjectIn, value);
	
	if (this->more == 0) {
		this->more = alloc_more_info (this->mem_zone);
	}
	
	OPAtomRef atom_key = key;
	
	if (this->more->attributes.IsValid () == FALSE) {
		this->more->attributes = OP_NEW_FROM_ZONE (this->mem_zone, OPDict);
	}
	
	return this->more->attributes->Insert (atom_key, value);
}

Bool
OPObject::SetAttribute (OPObjectIn key, OPObjectIn value)
{
	OP_METHOD (SetAttribute, 2);
	OP_ARG (1, OPObjectIn, key);
	OP_ARG (2, OPObjectIn, value);
	
	if (this->more == 0) {
		this->more = alloc_more_info (this->mem_zone);
	}
	
	if (this->more->attributes.IsValid () == FALSE) {
		this->more->attributes = OP_NEW_FROM_ZONE (this->mem_zone, OPDict);
	}
	
	return this->more->attributes->Insert (key, value);
}

Bool
OPObject::RemoveAttribute (const LatinChar* key, Count nth)
{
	OP_METHOD (RemoveAttribute, 2);
	OP_ARG (1, const LatinChar*, key);
	OP_ARG (2, Count, nth);
	
	if ( (this->more)
	  && (this->more->attributes.IsValid ()) ) {
		OPAtomRef atom_key = key;
		return more->attributes->Remove (atom_key, nth);
	}
	
	return FALSE;
}

Bool
OPObject::RemoveAttribute (OPObjectIn key, Count nth)
{
	OP_METHOD (RemoveAttribute, 2);
	OP_ARG (1, OPObjectIn, key);
	OP_ARG (2, Count, nth);
	
	if ( (this->more)
	  && (this->more->attributes.IsValid ()) ) {
		return this->more->attributes->Remove (key, nth);
	}
	
	return FALSE;
}

Bool
OPObject::RemoveAttribute (Count index)
{
	OP_METHOD (RemoveAttribute, 1);
	OP_ARG (1, Count, index);
	
	if ( (this->more)
	  && (this->more->attributes.IsValid ()) ) {
		return this->more->attributes->Remove (index);
	}
	
	return FALSE;
}

Bool
OPObject::GetAttribute (const LatinChar* key, Count nth) const
{
	OP_METHOD (GetAttribute, 2);
	OP_ARG (1, const LatinChar*, key);
	OP_ARG (2, Count, nth);
	
	OPObjectRef value;
	
	return this->GetAttribute (key, value, nth);
}

Bool
OPObject::GetAttribute (OPObjectIn key, Count nth) const
{
	OP_METHOD (GetAttribute, 2);
	OP_ARG (1, OPObjectIn, key);
	OP_ARG (2, Count, nth);
	
	OPObjectRef value;
	
	return this->GetAttribute (key, value, nth);
}

Bool
OPObject::GetAttribute (const LatinChar* key, OPObjectOut value, Count nth) const
{
	OP_METHOD (GetAttribute, 3);
	OP_ARG (1, const LatinChar*, key);
	OP_ARG (2, OPObjectOut, value);
	OP_ARG (3, Count, nth);
	
	if ( (this->more)
	  && (this->more->attributes.IsValid ()) ) {
		OPAtomRef atom_key = key;
		return this->more->attributes->GetFromKey (atom_key, value, nth);
	}
	
	return FALSE;
}

Bool
OPObject::GetAttribute (OPObjectIn key, OPObjectOut value, Count nth) const
{
	OP_METHOD (GetAttribute, 3);
	OP_ARG (1, OPObjectIn, key);
	OP_ARG (2, OPObjectOut, value);
	OP_ARG (3, Count, nth);
	
	if ( (this->more)
	  && (this->more->attributes.IsValid ()) ) {
		return this->more->attributes->GetFromKey (key, value, nth);
	}
	
	return FALSE;
}

Bool
OPObject::GetAttribute (Count index, OPObjectOut key, OPObjectOut value) const
{
	OP_METHOD (GetAttribute, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, key);
	OP_ARG (3, OPObjectOut, value);
	
	if ( (this->more)
	  && (this->more->attributes.IsValid ()) ) {
		return this->more->attributes->Get (index, key, value);
	}
	
	return FALSE;
}

/****************************************************************/

/*
 *	The Data Report class might want to be saved when this
 *	object gets saved. It has to register, otherwise, we would
 *	never know about the connection.
 */

Bool
OPObject::RegisterAsTargetOf (OPObjectIn initiator)
{
	OP_METHOD (RegisterAsTargetOf, 1);
	OP_ARG (1, OPObjectIn, initiator);
	
	//	Don't register a volatile observer...
	
	if ( (initiator.IsInvalid ())
	  || (initiator->GetAttribute ("OPaC.volatile.observer")) ) {
		return FALSE;
	}
	
	if (this->more == 0) {
		this->more = alloc_more_info (this->mem_zone);
	}
	
	if (this->more->registered.IsValid () == FALSE) {
		this->more->registered = OP_NEW_FROM_ZONE (this->mem_zone, OPHashTable);
	}
	
	return this->more->registered->Insert (initiator);
}

Bool
OPObject::UnregisterAsTargetOf (OPObjectIn initiator)
{
	OP_METHOD (UnregisterAsTargetOf, 1);
	OP_ARG (1, OPObjectIn, initiator);
	
	if ( (this->more)
	  && (this->more->registered.IsValid ()) ) {
		return this->more->registered->Remove (initiator);
	}
	
	return FALSE;
}

/****************************************************************/

/*
 *	Return the nth observer name. An observer has a standard
 *	interface, so the name is all the information that is
 *	needed.
 */

Bool
OPObject::GetObserver (Count index, OPAtomOut name) const
{
	OP_METHOD (GetObserver, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, name);

	return (find_target_by_type (index, this->ReturnRuntime (), OP_TARGET_IS_OBSERVER, name))
		 ? TRUE : FALSE;
}


/*
 *	Return the nth actor name. An actor has a standard
 *	interface, so the name is all the information that is
 *	needed.
 */

Bool
OPObject::GetActor (Count index, OPAtomOut name) const
{
	OP_METHOD (GetActor, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, name);

	return (find_target_by_type (index, this->ReturnRuntime (), OP_TARGET_IS_ACTOR, name))
		 ? TRUE : FALSE;
}


/*
 *	Return the nth method name. If you need more information
 *	about the method's interface, use the complete GetMethod
 *	method.
 */

Bool
OPObject::GetMethod (Count index, OPAtomOut name) const
{
	OP_METHOD (GetMethod, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, name);

	return (find_target_by_type (index, this->ReturnRuntime (), OP_TARGET_IS_METHOD, name))
		 ? TRUE : FALSE;
}


/*
 *	Return the nth method information, including following
 *	data: the name, the number of parameters taken by that
 *	method, the parameter descriptor array.
 */

Bool
OPObject::GetMethod (Count index, OPAtomOut name, Card32Out param_num,
					 const OPaC_ParamDesc* param_array) const
{
	OP_METHOD (GetMethod, 4);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, name);
	OP_ARG (3, Card32Out, param_num);
	OP_ARG (4, const OPaC_ParamDesc*, param_array);
	
	const OPaC_Target* target = find_target_by_type (index, this->ReturnRuntime (), OP_TARGET_IS_METHOD, name);

	if (target) {
		
		param_num   = target->target_method->num_param;
		param_array	= target->target_method->param;
		
		return TRUE;
	}

	return FALSE;
}



/*
 *	Return the nth data slot name. If you need to know more
 *	about the data type, just use the complete GetPublicData.
 */

Bool
OPObject::GetPublicData (CountIn index, OPAtomOut name) const
{
	OP_METHOD (GetPublicData, 2);
	OP_ARG (1, CountIn, index);
	OP_ARG (2, OPAtomOut, name);

	OPTypeInfoRef type;
	OPStringRef   desc;
	
	return this->GetPublicData (index, name, desc, type);
}

Bool
OPObject::GetPublicData (CountIn index,
						 OPAtomOut name,
						 OPTypeInfoOut type) const
{
	OP_METHOD (GetPublicData, 3);
	OP_ARG (1, CountIn, index);
	OP_ARG (2, OPAtomOut, name);
	OP_ARG (3, OPTypeInfoOut, type);
	
	OPStringRef desc;
	return this->GetPublicData (index, name, desc, type);
}



/*
 *	Return the nth data slot information, including following
 *	data: the name and the data type.
 *	
 *	If this method gets overridden by some subclass, it is
 *	important to override also the CountPublicData, Query...
 *	and Define... methods. Be careful to return specific
 *	shortcut values.
 */

Bool
OPObject::GetPublicData (Count index, OPAtomOut name, OPStringOut desc,
						 OPTypeInfoOut type) const
{
	OP_METHOD (GetPublicData, 4);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, name);
	OP_ARG (3, OPStringOut, desc);
	OP_ARG (4, OPTypeInfoOut, type);
	
	const OPaC_PubData* pub = 0;
	OPRuntimeRef runtime = this->ReturnRuntime ();
	
	while (runtime->CountPubData () <= index) {
		index  -= runtime->CountPubData ();
		runtime = runtime->ReturnSuper ();
		if (runtime.IsInvalid ()) return FALSE;
	}
	
	if (runtime->GetPubData (index, pub)) {
		
		type = OP_NEW (OPTypeInfo);
		
		name = pub->data_def.var_name;
		desc = pub->data_def.var_desc;
		
		type->Initialise (pub->data_def.var_type,
						  pub->bitset_info,
						  pub->data_def.is_read_ok,
						  pub->data_def.is_write_ok);
		
		return TRUE;
	}

	return FALSE;
}

Bool
OPObject::GetPublicDataNotifier (OPAtomIn name, Count index,
								 OPAtomOut notifier) const
{
	OP_METHOD (GetPublicDataNotifier, 3);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, Count, index);
	OP_ARG (3, OPAtomOut, notifier);
	
	OPRuntimeRef runtime = this->ReturnRuntime ();
	
	while (runtime->CountNotify (name) <= index) {
		index  -= runtime->CountNotify (name);
		runtime = runtime->ReturnSuper ();
		if (runtime.IsInvalid ()) return FALSE;
	}
	
	return runtime->GetNotify (name, index, notifier);
}


/*
 *	Return the nth attached observer, if any. When no more
 *	observer can be found, return FALSE.
 */

Bool
OPObject::GetAttached (Count index, OPAtomOut event_name,
					   OPObjectOut observing, Card32Out id) const
{
	OP_METHOD (GetAttached, 4);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, event_name);
	OP_ARG (3, OPObjectOut, observing);
	OP_ARG (4, Card32Out, id);

	OPAtomRef actor_name;
	
	return this->GetAttached (index, event_name, actor_name, observing, id);
}

Bool
OPObject::GetAttached (Count index,
					   OPAtomOut event_name,
					   OPAtomOut actor_name,
					   OPObjectOut observing,
					   Card32Out id) const
{
	OP_METHOD (GetAttached, 4);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, event_name);
	OP_ARG (3, OPAtomOut, actor_name);
	OP_ARG (4, OPObjectOut, observing);
	OP_ARG (5, Card32Out, id);

	if ( (this->more)
	  && (index < this->more->obs_count) ) {
		
		Observer* iter = this->more->obs_head;
		
		while (index--) {
			iter = iter->next;
		}
		
		event_name = iter->event_name;
		actor_name = iter->actor_name;
		observing  = iter->observing;
		id         = iter->id;

		return TRUE;
	}
	
	return FALSE;
}

/****************************************************************/

Bool
OPObject::GetObserver (OPContainerOut info) const
{
	OP_METHOD (GetObserver, 1);
	OP_ARG (1, OPContainerOut, info);

	info = OPContainerNull;
	
	//	NIY
	
	return FALSE;
}


Bool
OPObject::GetActor (OPContainerOut info) const
{
	OP_METHOD (GetActor, 1);
	OP_ARG (1, OPContainerOut, info);

	info = OPContainerNull;
	
	//	NIY
	
	return FALSE;
}


Bool
OPObject::GetMethod (OPContainerOut info) const
{
	OP_METHOD (GetMethod, 1);
	OP_ARG (1, OPContainerOut, info);

	info = OPContainerNull;
	
	//	NIY
	
	return FALSE;
}


Bool
OPObject::GetPublicData (OPContainerOut info) const
{
	OP_METHOD (GetPublicData, 1);
	OP_ARG (1, OPContainerOut, info);

	info = OPContainerNull;
	
	//	NIY
	
	return FALSE;
}


Bool
OPObject::GetAttached (OPContainerOut info) const
{
	OP_METHOD (GetAttached, 1);
	OP_ARG (1, OPContainerOut, info);

	info = OPContainerNull;
	
	//	NIY
	
	return FALSE;
}

/****************************************************************/


/*
 *	Perform some action. No parameter is needed, because
 *	the actor doesn't need any. Return TRUE if the actor
 *	could be properly called.
 */

Bool
OPObject::Action (OPAtomIn event)
{
	OP_METHOD (Action, 1);
	OP_ARG (1, OPAtomIn, event);
	
	const OPaC_Target* actor = this->ReturnRuntime ()->FindActor (event);

	if (actor) {
		(actor->target_method->a_trampo) (this);
		return TRUE;
	}

	return FALSE;
}

Bool
OPObject::Action (const LatinChar* event)
{
	OP_METHOD (Action, 1);
	OP_ARG (1, const LatinChar*, event);
	
	OPAtomRef atom = event;
	return this->Action (atom);
}


/****************************************************************/

static Bool
handle_query_object (const OPaC_PubDataDef& data_def, OPObjectOut data, const OPObject* that)
{
	if ( (data_def.var_type->simple == OP_TYPE_OBJECT)
	  || (data_def.var_type->simple == OP_TYPE_TEXT)
	  || (data_def.var_type->simple == OP_TYPE_ARRAY) ) {
		
		if (data_def.is_read_ok) {
			data = (*data_def.access_read)(that);
			return TRUE;
		}
	}
	
	return FALSE;
}

static Bool
handle_define_object (const OPaC_PubDataDef& data_def, OPObjectIn data, OPObject* that)
{
	OPObjectRef x = data;
	
	while ( (x.IsValid ())
		 && (x->IsKindOf (OP_RUNTIME (OPValue))) ) {
		OPValueRef value = x;
		switch (value->ReturnType ()) {
			case OP_TYPE_OBJECT:
				x = value->ReturnObject ();
				break;
			case OP_TYPE_ARRAY:
				x = value->ReturnArray ();
				break;
			case OP_TYPE_TEXT:
				x = value->ReturnText ();
				break;
			default:
				goto other;
		}
	}
	
other:
	if ( (data_def.var_type->simple == OP_TYPE_OBJECT)
	  || (data_def.var_type->simple == OP_TYPE_TEXT)
	  || (data_def.var_type->simple == OP_TYPE_ARRAY) ) {
		
		if (data_def.is_write_ok) {
			return (*data_def.access_write)(that, x);
		}
	}
	
	return FALSE;
}

/*
 *	Special case for handling conversions between basic data
 *	types and the OPValue data type.
 */

static Bool
handle_query_value (const OPaC_PubDataDef& data_def, OPValueOut data, const OPObject* that)
{
	if (data_def.is_read_ok) {
		data = (*data_def.access_read)(that);
		return TRUE;
	}
	return FALSE;
#if 0
	if (data_def.var_type == Card32TypeDef) {
		
		Card32 value = 0;
		
		if (data_def.is_function) {
			OPaC_FunctionCard32 fptr = data_def.rd.funcCard32;
			if (fptr != 0) {
				value = (*fptr)(that);
				goto ok_card;
			}
		}

		if (data_def.is_accessor) {
			OPaC_ReadAccessCard32 fptr = data_def.rd.accessCard32;
			if (fptr != 0) {
				if ((*fptr)(that, value)) {
					goto ok_card;
				}
			}
		}
		
		return FALSE;
		
ok_card:
		data = value;
		return TRUE;
	}
	
	if (data_def.var_type == Int32TypeDef) {
		
		Int32  value = 0;
		
		if (data_def.is_function) {
			OPaC_FunctionInt32 fptr = data_def.rd.funcInt32;
			if (fptr != 0) {
				value = (*fptr)(that);
				goto ok_int;
			}
		}

		if (data_def.is_accessor) {
			OPaC_ReadAccessInt32 fptr = data_def.rd.accessInt32;
			if (fptr != 0) {
				if ((*fptr)(that, value)) {
					goto ok_int;
				}
			}
		}
		
		return FALSE;
		
ok_int:
		data = value;
		return TRUE;
	}
	
	if (data_def.var_type == RealTypeDef) {
		
		Real   value = 0;
		
		if (data_def.is_function) {
			OPaC_FunctionReal fptr = data_def.rd.funcReal;
			if (fptr != 0) {
				value = (*fptr)(that);
				goto ok_real;
			}
		}

		if (data_def.is_accessor) {
			OPaC_ReadAccessReal fptr = data_def.rd.accessReal;
			if (fptr != 0) {
				if ((*fptr)(that, value)) {
					goto ok_real;
				}
			}
		}
		
		return FALSE;
		
ok_real:
		data = value;
		return TRUE;
	}
	
	if (data_def.var_type == BoolTypeDef) {
		
		Bool   value = 0;
		
		if (data_def.is_function) {
			OPaC_FunctionBool fptr = data_def.rd.funcBool;
			if (fptr != 0) {
				value = (*fptr)(that);
				goto ok_bool;
			}
		}

		if (data_def.is_accessor) {
			OPaC_ReadAccessBool fptr = data_def.rd.accessBool;
			if (fptr != 0) {
				if ((*fptr)(that, value)) {
					goto ok_bool;
				}
			}
		}
		
		return FALSE;
		
ok_bool:
		data = value;
		return TRUE;
	}
	
	return FALSE;
#endif
}

static Bool
handle_define_value (const OPaC_PubDataDef& data_def, OPValueIn data, OPObject* that)
{
	if (data_def.is_write_ok) {
		return (*data_def.access_write)(that, data);
	}
	
	return FALSE;
#if 0
	if (data_def.var_type == Card32TypeDef) {
		
		Card32 value = data;
		
		if (data_def.is_accessor) {
			OPaC_WriteAccessCard32 fptr = data_def.wr.accessCard32;
			if (fptr != 0) {
				return ((*fptr)(that, value));
			}
		}
		
		return FALSE;
	}
	
	if (data_def.var_type == RealTypeDef) {
		
		Real value = data;
		
		if (data_def.is_accessor) {
			OPaC_WriteAccessReal fptr = data_def.wr.accessReal;
			if (fptr != 0) {
				return ((*fptr)(that, value));
			}
		}
		
		return FALSE;
	}
	
	if (data_def.var_type == BoolTypeDef) {
		
		Bool value = data;
		
		if (data_def.is_accessor) {
			OPaC_WriteAccessBool fptr = data_def.wr.accessBool;
			if (fptr != 0) {
				return ((*fptr)(that, value));
			}
		}
		
		return FALSE;
	}
	
	if (data_def.var_type == OPStringTypeDef) {
		
		OPStringRef value;
		
		if (data->ConvertToString (value) == FALSE) {
			value = "?";
		}
		
		if (data_def.is_accessor) {
			OPaC_WriteAccessOPObject fptr = data_def.wr.accessOPObject;
			if (fptr != 0) {
				return ((*fptr)(that, value));
			}
		}
		
		return FALSE;
	}
	
	if (data_def.var_type == OPAtomTypeDef) {
		
		OPAtomRef value;
		
		if (data->ConvertToAtom (value) == FALSE) {
			value = "?";
		}
		
		if (data_def.is_accessor) {
			OPaC_WriteAccessOPObject fptr = data_def.wr.accessOPObject;
			if (fptr != 0) {
				return ((*fptr)(that, value));
			}
		}
		
		return FALSE;
	}
	
	if (data_def.var_type == OPValueTypeDef) {
		
		if (data_def.is_accessor) {
			OPaC_WriteAccessOPObject fptr = data_def.wr.accessOPObject;
			if (fptr != 0) {
				return ((*fptr)(that, data));
			}
		}
		
		return FALSE;
	}
	
	return FALSE;
#endif
}

/*
 *	The QueryData methods return a data record, using the
 *	specified name/index and data type. Return TRUE on
 *	success.
 *
 *	The DefineData methods are similar to QueryData, but
 *	they modify the data rather than querying the existing
 *	value. Return TRUE on success.
 *
 *	NB: Some published data values may only be queried and
 *	    will not accept to be (re)defined.
 */

Bool
OPObject::QueryData (OPAtomIn name, OPObjectOut data) const
{
	OP_METHOD (QueryData, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectOut, data);
	
	Count               nth = 0;
	const OPaC_PubData* pub = 0;
	
	for (;;) {
		
		if (pub == 0) {
			
			pub = this->ReturnRuntime ()->FindPubData (name, nth++);
			
			if (pub == 0) {
				break;
			}
		}
		
		if (pub->data_def.is_read_ok) {
			if (handle_query_object (pub->data_def, data, this)) {
				return TRUE;
			}
		}
		
		pub = pub->sim_next;
	}
	
	if (this->IsKindOf (OP_RUNTIME (OPDataSlot))) {
		OPDataSlotRef slot = (OPDataSlot*) const_cast<OPObject*> (this);
		return slot->FindField (name, data);
	}
	
	return FALSE;
}

Bool
OPObject::QueryData (OPAtomIn name, OPValueOut data) const
{
	OP_METHOD (QueryData, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPValueOut, data);
	
	Count               nth = 0;
	const OPaC_PubData* pub = 0;
	
	for (;;) {
		
		if (pub == 0) {
			
			pub = this->ReturnRuntime ()->FindPubData (name, nth++);
			
			if (pub == 0) {
				break;
			}
		}
		
		if (pub->data_def.is_read_ok) {
			OPObjectRef object;
			if (handle_query_object (pub->data_def, object, this)) {
				if (object.IsValid () == FALSE) {
					data = object;
					return TRUE;
				}
				if (object->IsKindOf (OP_RUNTIME (OPValue))) {
					data = object;
					return TRUE;
				}
				data = OP_NEW (OPValue);
				data->SetObject (object);
				return TRUE;
			}
			if (handle_query_value (pub->data_def, data, this)) {
				return TRUE;
			}
		}
		
		pub = pub->sim_next;
	}

	static OPAtomRef atom_value = "value";
	
	//	Last chance... Try to use the object itself as a data.
	
	if (name == atom_value) {
		
		if (this->IsKindOf (OP_RUNTIME (OPAtom))) {
			const OPAtom* x = (const OPAtom*)this;
			OPStringRef text = x->ReturnString ();
			data = OP_NEW (OPValue);
			data->SetText (text);
			return TRUE;
		}
		
		if (this->IsKindOf (OP_RUNTIME (OPString))) {
			const OPString* x = (const OPString*)this;
			OPStringRef text = x->ReturnString ();
			data = OP_NEW (OPValue);
			data->SetText (text);
			return TRUE;
		}
	}
	
	return FALSE;
}

Bool
OPObject::QueryData (Count index, OPObjectOut data) const
{
	OP_METHOD (QueryData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, data);
	
	const OPaC_PubData* pub = 0;
	OPRuntimeRef runtime = this->ReturnRuntime ();
	
	while (runtime->CountPubData () <= index) {
		index  -= runtime->CountPubData ();
		runtime = runtime->ReturnSuper ();
		if (runtime.IsValid () == FALSE) return FALSE;
	}
	
	if (runtime->GetPubData (index, pub)) {
		while (pub) {
			if (pub->data_def.is_read_ok) {
				if (handle_query_object (pub->data_def, data, this)) {
					return TRUE;
				}
			}
			
			pub = pub->sim_next;
		}
	}
	
	return FALSE;
}

Bool
OPObject::QueryData (Count index, OPValueOut data) const
{
	OP_METHOD (QueryData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPValueOut, data);
	
	const OPaC_PubData* pub = 0;
	OPRuntimeRef runtime = this->ReturnRuntime ();
	
	while (runtime->CountPubData () <= index) {
		index  -= runtime->CountPubData ();
		runtime = runtime->ReturnSuper ();
		if (runtime.IsValid () == FALSE) return FALSE;
	}
	
	if (runtime->GetPubData (index, pub)) {
		while (pub) {
			if (pub->data_def.is_read_ok) {
				OPObjectRef object;
				if (handle_query_object (pub->data_def, object, this)) {
					data = OP_NEW (OPValue);
					data->SetObject (object);
					return TRUE;
				}
				if (handle_query_value (pub->data_def, data, this)) {
					return TRUE;
				}
			}
			
			pub = pub->sim_next;
		}
	}
	
	return FALSE;
}


Bool
OPObject::DefineData (OPAtomIn name, OPObjectIn data)
{
	OP_METHOD (DefineData, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectIn, data);
	
	Count               nth = 0;
	const OPaC_PubData* pub = 0;
	
	for (;;) {
		
		if (pub == 0) {
			
			pub = this->ReturnRuntime ()->FindPubData (name, nth++);
			
			if (pub == 0) {
				break;
			}
		}
		
		if (pub->data_def.is_write_ok) {
			if (handle_define_object (pub->data_def, data, this)) {
				return TRUE;
			}
		}
		
		pub = pub->sim_next;
	}
	
	return FALSE;
}

Bool
OPObject::DefineData (OPAtomIn name, OPValueIn data)
{
	OP_METHOD (DefineData, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPValueIn, data);
	
	Count               nth = 0;
	const OPaC_PubData* pub = 0;
	
	for (;;) {
		
		if (pub == 0) {
			
			pub = this->ReturnRuntime ()->FindPubData (name, nth++);
			
			if (pub == 0) {
				break;
			}
		}
		
		if (pub->data_def.is_write_ok) {
			if (handle_define_value (pub->data_def, data, this)) {
				return TRUE;
			}
			if (handle_define_object (pub->data_def, data, this)) {
				return TRUE;
			}
		}
		
		pub = pub->sim_next;
	}
	
	return FALSE;
}

Bool
OPObject::DefineData (Count index, OPObjectIn data)
{
	OP_METHOD (DefineData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectIn, data);
	
	const OPaC_PubData* pub = 0;
	OPRuntimeRef runtime = this->ReturnRuntime ();
	
	while (runtime->CountPubData () <= index) {
		index  -= runtime->CountPubData ();
		runtime = runtime->ReturnSuper ();
		if (runtime.IsValid () == FALSE) return FALSE;
	}
	
	if (runtime->GetPubData (index, pub)) {
		while (pub) {
			if (pub->data_def.is_write_ok) {
				if (handle_define_object (pub->data_def, data, this)) {
					return TRUE;
				}
			}
			
			pub = pub->sim_next;
		}
	}
	
	return FALSE;
}

Bool
OPObject::DefineData (Count index, OPValueIn data)
{
	OP_METHOD (DefineData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPValueIn, data);
	
	const OPaC_PubData* pub = 0;
	OPRuntimeRef runtime = this->ReturnRuntime ();
	
	while (runtime->CountPubData () <= index) {
		index  -= runtime->CountPubData ();
		runtime = runtime->ReturnSuper ();
		if (runtime.IsValid () == FALSE) return FALSE;
	}
	
	if (runtime->GetPubData (index, pub)) {
		while (pub) {
			if (pub->data_def.is_write_ok) {
				if (handle_define_value (pub->data_def, data, this)) {
					return TRUE;
				}
				if (handle_define_object (pub->data_def, data, this)) {
					return TRUE;
				}
			}
			
			pub = pub->sim_next;
		}
	}
	
	return FALSE;
}


/****************************************************************/

static Bool
object_query_container (OPObjectIn obj, OPStringIn name, OPObjectOut data)
{
	LatinChar buffer[10];
	Count     buflen;
	name->GetStringCopy (buffer, 10);
	buflen = strlen (buffer);
	
	if ( (buffer[0] == '[')
	  && (buffer[buflen-1] == ']')
	  && (obj->IsKindOf (OP_RUNTIME (OPGenericContainer))) ) {
		
		OPGenericContainerRef container = obj;
		buffer[buflen-1] = 0;
		Count num = atoi (buffer+1);
		
		if (container->Get (num, data)) {
			return TRUE;
		}
	}
	
	return FALSE;
}

static Bool
object_parse_path (OPObjectInOut obj, OPStringInOut dir)
{
	if (obj.IsValid () == FALSE) {
		return FALSE;
	}
	
	Count pos = 0;
	
	while (dir->Find ('.', pos)) {
		
		OPStringRef name;
		OPStringRef next_dir;
		OPObjectRef next_obj;
		
		dir->MakeLeft (pos, name);
		dir->MakeMiddle (pos+1, 1000, next_dir);
		
		//	The special notation "[nnn]" is accepted and is used as an index
		//	for a generic container. It won't query the object...
		
		if (object_query_container (obj, name, obj)) {
			dir = next_dir;
			continue;
		}
		
		//	We really need to query the object.
		
		if ( (obj->QueryData (name, next_obj) == FALSE)
		  || (next_obj.IsValid () == FALSE) ) {
			return FALSE;
		}
		
		dir = next_dir;
		obj = next_obj;
	}
	
	return obj.IsValid ();
}

Bool
OPObject::Action (OPObjectIn o, OPStringIn path)
{
	OP_MEMBER (Action, 2);
	OP_ARG (1, OPObjectIn, o);
	OP_ARG (2, OPStringIn, path);
	
	OPStringRef dir = path;
	OPObjectRef obj = o;
	
	if (object_parse_path (obj, dir)) {
		return obj->Action (dir);
	}
	
	return FALSE;
}

Bool
OPObject::QueryData (OPObjectIn o, OPStringIn path, OPObjectOut data)
{
	OP_MEMBER (QueryData, 3);
	OP_ARG (1, OPObjectIn, o);
	OP_ARG (2, OPStringIn, path);
	OP_ARG (3, OPObjectOut, data);
	
	OPStringRef dir = path;
	OPObjectRef obj = o;
	
	if (object_parse_path (obj, dir)) {
		if (object_query_container (obj, dir, data)) {
			return TRUE;
		} else {
			return obj->QueryData (dir, data);
		}
	}
	
	return FALSE;
}

Bool
OPObject::QueryData (OPObjectIn o, OPStringIn path, OPValueOut data)
{
	OP_MEMBER (QueryData, 3);
	OP_ARG (1, OPObjectIn, o);
	OP_ARG (2, OPStringIn, path);
	OP_ARG (3, OPValueOut, data);
	
	OPStringRef dir = path;
	OPObjectRef obj = o;
	
	if (object_parse_path (obj, dir)) {
		OPObjectRef x;
		if (object_query_container (obj, dir, x)) {
			if (x->IsKindOf (OP_RUNTIME (OPValue))) {
				data = x;
			} else {
				data = OP_NEW (OPValue);
				data->SetObject (x);
			}
			return TRUE;
		} else {
			return obj->QueryData (dir, data);
		}
	}
	
	return FALSE;
}

Bool
OPObject::DefineData (OPObjectIn o, OPStringIn path, OPObjectIn data)
{
	OP_MEMBER (DefineData, 3);
	OP_ARG (1, OPObjectIn, o);
	OP_ARG (2, OPStringIn, path);
	OP_ARG (3, OPObjectIn, data);
	
	OPStringRef dir = path;
	OPObjectRef obj = o;
	
	if (object_parse_path (obj, dir)) {
		return obj->DefineData (dir, data);
	}
	
	return FALSE;
}

Bool
OPObject::DefineData (OPObjectIn o, OPStringIn path, OPValueIn data)
{
	OP_MEMBER (DefineData, 3);
	OP_ARG (1, OPObjectIn, o);
	OP_ARG (2, OPStringIn, path);
	OP_ARG (3, OPValueIn, data);
	
	OPStringRef dir = path;
	OPObjectRef obj = o;
	
	if (object_parse_path (obj, dir)) {
		return obj->DefineData (dir, data);
	}
	
	return FALSE;
}

OPObjectRef
OPObject::ReturnSelf ()
{
	OP_METHOD (ReturnSelf, 0);
	return this;
}


OPObjectRef
OPObject::Clone () const
{
	OP_METHOD (Clone, 0);
	OPObjectRef copy = OP_DYNAMIC_NEW (this->ReturnRuntime (), OPObject);
	
	return copy->Copy (const_cast<OPObject*>(this));
}

/*
 *	This method MUST be overridden in order to do something
 *	meaningful !
 */

OPObjectRef
OPObject::Copy (OPObjectIn copy)
{
	OP_METHOD (Copy, 1);
	OP_ARG (1, OPObjectIn, copy);
	return this;
}
