/*
 *	plotview.cxx
 *
 *	Plot View class definition.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/plotview.h>
#include <OPaC/window.h>

#include <OPaC/notmacro.h>
#include <OPaC/pubmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPPlotView);

/****************************************************************/

OP_IMPLEMENT_TYPEENUM (OPPlotViewType)
{
	OP_IMPLEMENT_TYPEENUM_DEF (OP_PLOT_VIEW_NONE),
	OP_IMPLEMENT_TYPEENUM_DEF (OP_PLOT_VIEW_Y_T),
	OP_IMPLEMENT_TYPEENUM_DEF (OP_PLOT_VIEW_X_Y),
	OP_IMPLEMENT_TYPEENUM_END ()
};

/****************************************************************/

OP_PUBLISH_SLOT (OPPlotView, h_zero, ReturnHorizontalZero, SetHorizontalZero, "horizontal zero", Real);
OP_PUBLISH_SLOT (OPPlotView, h_size, ReturnHorizontalSize, SetHorizontalSize, "horizontal size", Real);
OP_PUBLISH_SLOT (OPPlotView, v_zero, ReturnVerticalZero,   SetVerticalZero,   "vertical zero",   Real);
OP_PUBLISH_SLOT (OPPlotView, v_size, ReturnVerticalSize,   SetVerticalSize,   "vertical size",   Real);

/****************************************************************/

/*
 *	Create the plot view object.
 */

OPPlotView::OPPlotView ()
{
	OP_METHOD (OPPlotView, 0);
	
	this->h_zero = 0.0;
	this->v_zero = 0.5;
	this->h_size = 1.0;
	this->v_size = 1.0;
	
	this->default_color = OP_COLOR_BLACK;
	
	this->data_type = OP_PLOT_VIEW_NONE;
	this->data_set  = 0;
	this->data_num  = 0;
	this->data_zone = 0;
	this->data_org  = 0;
}



/*
 *	Dispose of any cloned data before destroying the plot
 *	view object.
 */

OPPlotView::~OPPlotView ()
{
	OP_METHOD (~OPControl, 0);
	
	this->FlushDataSet ();
}



/*
 *	Mark the object for deletion
 */

void
OPPlotView::Delete ()
{
	OP_METHOD (Delete, 0);
	
	super::Delete ();
}



/*
 *	Save the control to the storage.
 */

void
OPPlotView::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);

	static OPStorageSync info ("<OPPlotView h_zero h_size : v_zero v_size :"
							   " default_color data_num data_type data_org :"
							   " { data } >");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->h_zero << this->h_size << OPStorageSync::mark;
	storage << this->v_zero << this->v_size << OPStorageSync::mark;
	storage << this->default_color;
	storage << this->data_num;
	storage << this->data_type;
	storage << this->data_org;
	storage << OPStorageSync::mark;
	
	this->StoreDataSet (storage);
	
	storage << OPStorageSync::end;
}



/*
 *	Restore the control from the storage.
 */

void
OPPlotView::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	this->FlushDataSet ();
	
	Card32 type = 0;
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->h_zero >> this->h_size >> OPStorageSync::mark;
	storage >> this->v_zero >> this->v_size >> OPStorageSync::mark;
	storage >> this->default_color;
	storage >> this->data_num;
	storage >> this->data_type;
	storage >> this->data_org;
	storage >> OPStorageSync::mark;
	
	this->RestoreDataSet (storage);
	
	storage >> OPStorageSync::end;
}


/*
 *	Store the data set to the storage. The type and the
 *	number of elements has already been stored.
 */

void
OPPlotView::StoreDataSet (OPStorageIn storage)
{
	OP_METHOD (StoreData, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	storage << OPStorageSync::repeat;
	
	for (Count i = 0; i < this->data_num; i++) {
		storage << this->data_set[i];
	}
	
	storage << OPStorageSync::end_repeat;
}


/*
 *	Restore the data set to the storage. The type and the
 *	number of elements has already been restored.
 */

void
OPPlotView::RestoreDataSet (OPStorageIn storage)
{
	OP_METHOD (RestoreData, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	Real* set = 0;
	this->data_zone = OPaC_data_zone;
	this->data_set  = set = OP_ALLOC_ARRAY_FROM_ZONE (this->data_zone, Real, this->data_num);
	
	storage >> OPStorageSync::repeat;
	
	for (Count i = 0; i < this->data_num; i++) {
		storage >> set[i];
	}
	
	storage >> OPStorageSync::end_repeat;
}


/*
 *	Dispose of any cloned data. If the data was provided
 *	as external data, simply forget about its pointer.
 */

void
OPPlotView::FlushDataSet ()
{
	OP_METHOD (FlushDataSet, 0);
	
	if (this->data_zone) {
		OP_FREE_FROM_ZONE (this->data_zone, this->data_set);
		this->data_zone = 0;
	}
	
	this->data_set = 0;
	this->data_num = 0;
	this->data_org = 0;
}


/*
 *	Define the data set which will be used for displaying
 *	information in the plot view. The data will not be
 *	copied unless "CloneDataSet" is called. If the data
 *	has to be freed by the plot view, the caller must
 *	specify its memory zone.
 */

void
OPPlotView::DefineDataSet (const Real* set, Count num, OPPlotViewType type, OPMemZone* zone)
{
	OP_METHOD (DefineDataSet, 4);
	OP_ARG (1, const Real*, set);
	OP_ARG (2, Count, num);
	OP_ARG (3, OPPlotViewType, type);
	OP_ARG (4, OPMemZone*, zone);
	
	this->FlushDataSet ();
	
	this->data_type = type;
	this->data_num  = num;
	this->data_zone = zone;
	this->data_set  = const_cast<Real*> (set);
	this->data_org  = 0;
}


/*
 *	Copy the contents of the data set. This cloning is only
 *	needed if the data is provided by an external memory
 *	pointer. If the data already belongs to the plot view,
 *	there is no need to clone it.
 */

void
OPPlotView::CloneDataSet ()
{
	OP_METHOD (CloneDataSet, 0);
	
	if ( (this->data_zone == 0)
	  && (this->data_set) ) {
		
		const Real* model = this->data_set;
		Real* set = 0;
		
		this->data_zone = OPaC_data_zone;
		this->data_set  = set = OP_ALLOC_ARRAY_FROM_ZONE (this->data_zone, Real, this->data_num);
		
		for (Count i = 0; i < this->data_num; i++) {
			set[i] = model[i];
		}
	}
}


/*
 *	Define data sample, depending on the type of internal
 *	representation (1D or 2D).
 */

Bool
OPPlotView::SetData (Count index, Real d1)
{
	OP_METHOD (SetData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, Real, d1);
	
	if ( (this->data_num)
	  && (this->data_zone)
	  && (this->data_type == OP_PLOT_VIEW_Y_T) ) {
		
		index += this->data_org;
		index %= this->data_num;
		
		this->data_set[index] = d1;
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPPlotView::SetData (Count index, Real d1, Real d2)
{
	OP_METHOD (SetData, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, Real, d1);
	OP_ARG (3, Real, d2);
	
	if ( (this->data_num)
	  && (this->data_zone)
	  && (this->data_type == OP_PLOT_VIEW_X_Y) ) {
		
		index += this->data_org*2;
		index %= this->data_num/2;
		
		this->data_set[index*2+0] = d1;
		this->data_set[index*2+1] = d2;
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Access data sample, depending on the type of internal
 *	representation (1D or 2D).
 */

Bool
OPPlotView::GetData (Count index, RealOut d1) const
{
	OP_METHOD (GetData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, RealOut, d1);
	
	if ( (this->data_num)
	  && (this->data_type == OP_PLOT_VIEW_Y_T) ) {
		
		index += this->data_org;
		index %= this->data_num;
		
		d1 = this->data_set[index];
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPPlotView::GetData (Count index, RealOut d1, RealOut d2) const
{
	OP_METHOD (GetData, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, RealOut, d1);
	OP_ARG (3, RealOut, d2);
	
	if ( (this->data_num)
	  && (this->data_type == OP_PLOT_VIEW_X_Y) ) {
		
		index += this->data_org*2;
		index %= this->data_num/2;
		
		d1 = this->data_set[index*2+0];
		d2 = this->data_set[index*2+1];
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	Return the number of useful points decribed by the
 *	data set (this depends on the data type).
 */

Count
OPPlotView::TotalCount () const
{
	OP_METHOD (TotalCount, 0);
	
	switch (this->data_type) {
		
		case OP_PLOT_VIEW_NONE:
			return 0;
		
		case OP_PLOT_VIEW_Y_T:
			return this->data_num;
		
		case OP_PLOT_VIEW_X_Y:
			return this->data_num / 2;
	}
	
	return 0;
}

Bool
OPPlotView::GetDataPoint (Count index, RealOut x, RealOut y, OPColorPixelOut color)
{
	OP_METHOD (GetDataPoint, 4);
	OP_ARG (1, Count, index);
	OP_ARG (2, RealOut, x);
	OP_ARG (3, RealOut, y);
	OP_ARG (4, OPColorPixelOut, color);
	
	switch (this->data_type) {
		
		case OP_PLOT_VIEW_NONE:
			return FALSE;
		
		case OP_PLOT_VIEW_Y_T:
			
			if (index < this->data_num) {
				
				Count pos = index;
				
				index += this->data_org;
				index %= this->data_num;
				
				x     = (Real)(pos) / (Real)(this->data_num);
				y     = this->data_set[index];
				color = this->default_color;
				
				return TRUE;
			}
			break;
		
		case OP_PLOT_VIEW_X_Y:
			
			if (index < this->data_num/2) {
				
				index += this->data_org*2;
				index %= this->data_num/2;
				
				x     = this->data_set[index*2+0];
				y     = this->data_set[index*2+1];
				color = this->default_color;
				
				return TRUE;
			}
			break;
	}
	
	return FALSE;
}


Bool
OPPlotView::SetHorizontalZero (Real zero)
{
	OP_METHOD (SetHorizontalZero, 1);
	OP_ARG (1, Real, zero);
	
	this->h_zero = zero;
	
	return TRUE;
}

Bool
OPPlotView::SetHorizontalSize (Real size)
{
	OP_METHOD (SetHorizontalSize, 1);
	OP_ARG (1, Real, size);
	
	this->h_size = size;
	
	return TRUE;
}

Bool
OPPlotView::SetVerticalZero (Real zero)
{
	OP_METHOD (SetVerticalZero, 1);
	OP_ARG (1, Real, zero);
	
	this->v_zero = zero;
	
	return TRUE;
}

Bool
OPPlotView::SetVerticalSize (Real size)
{
	OP_METHOD (SetVerticalSize, 1);
	OP_ARG (1, Real, size);
	
	this->v_size = size;
	
	return TRUE;
}

Real
OPPlotView::ReturnHorizontalZero () const
{
	OP_METHOD (ReturnHorizontalZero, 0);
	return this->h_zero;
}

Real
OPPlotView::ReturnHorizontalSize () const
{
	OP_METHOD (ReturnHorizontalSize, 0);
	return this->h_size;
}

Real
OPPlotView::ReturnVerticalZero () const
{
	OP_METHOD (ReturnVerticalZero, 0);
	return this->v_zero;
}

Real
OPPlotView::ReturnVerticalSize () const
{
	OP_METHOD (ReturnVerticalSize, 0);
	return this->v_size;
}


/*
 *	Define the origin of the current data set. A positive
 *	origin offsets the whole data set to the left. This is
 *	useful when using "wrap-around" data sets.
 */

Bool
OPPlotView::SetDataOrigin (Count origin)
{
	OP_METHOD (SetDataOrigin, 1);
	OP_ARG (1, Count, origin);
	
	if (this->data_num == 0) {
		return FALSE;
	}
	
	this->data_org = origin % this->data_num;
	
	return TRUE;
}

Count
OPPlotView::ReturnDataOrigin () const
{
	OP_METHOD (ReturnDataOrigin, 0);
	return this->data_org;
}


/*
 *	Draw the frame of the control. This is has been
 *	implemented in this class, since the frame is used
 *	to show which view is the main responder.
 */

void
OPPlotView::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	const OPLookManRef& lookman = OPLookMan::Current ();
	
	if (what & OP_DRAW_BACKGROUND) {
		
	}
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		Count max   = this->TotalCount ();
		Count index = 0;
		
		OPColorPixel color;
		OPCoord dx;
		OPCoord dy;
		OPCoord ox;
		OPCoord oy;
		
		this->GetOrigin (ox, oy);
		this->GetSize (dx, dy);
		
		oy = oy + dy;
		dy = -dy;
		
		Real x;
		Real y;
		
		for (index = 0; index < max; index++) {
			this->GetDataPoint (index, x, y, color);
			x /= this->h_size;
			x += this->h_zero;
			y /= this->v_size;
			y += this->v_zero;
			if ( (x >= 0.0) && (x < 1.0)
			  && (y >= 0.0) && (y < 1.0) ) {
				x *= dx;
				y *= dy;
				port->DrawDot (ox+x, oy+y, color);
			}
		}
	}
}
