/*
 *	random.cxx
 *
 *	Encapsulate values (numbers, integers, booleans, reals, etc.)
 *	so that they can be handled as standard OPaC objects.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/random.h>
#include <OPaC/pubmacro.h>

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPRandom);

/****************************************************************/

OP_PUBLISH_SLOT (OPRandom, maximum, ReturnMaximum, SetMaximum, "Maximum value",  Card32);
OP_PUBLISH_SLOT (OPRandom, scale,   ReturnScale,   SetScale,   "Scale factor",   Real);
OP_PUBLISH_SLOT (OPRandom, offset,  ReturnOffset,  SetOffset,  "Offset of zero", Real);

/****************************************************************/

#define	MAX_VALUE		((Card32)(100000000))
#define	MAX_VALUE_ROOT	((Card32)(10000))

/****************************************************************/

/*
 *	Initialise the random object to produce 0 or 1 in a
 *	(pseudo-)random way.
 */

OPRandom::OPRandom ()
{
	OP_METHOD (OPRandom, 0);
	
	this->SetMaximum (1);
	this->SetScaleOffset (1.0, 0.0);
	
	this->type  = OP_TYPE_REAL;
	this->value.the_real = 0.0;
	
	this->Initialise ();
}


/*
 *	Override if you want to provide a really random seed
 *	(e.g. by looking at the clock or processor time).
 */

Card32
OPRandom::ReturnInitialSeed ()
{
	OP_METHOD (ReturnInitialSeed, 0);
	
	//	This method can be overridden if the seed has to be generated
	//	in some smarter way. By default, this fixed seed is used if the
	//	user does not call Initialise after construction.
	
	return 21987;
}


/*
 *	Set up the accumulation array with the initial seed-
 *	generated values.
 */

void
OPRandom::Initialise ()
{
	OP_METHOD (Initialise, 0);
	
	static const Card32 base = 31415821;		//	from Sedgewick's
	Card32 seed = this->ReturnInitialSeed ();	//	any seed will do...
	
	this->index = 0;
	this->accumulate[0] = seed;
	
	for (Count j = 1; j < 55; j++) {
		this->accumulate[j] = (OPRandom::Multiply (this->accumulate[j-1], base) + 1) % MAX_VALUE;
	}
}



/*
 *	Nothing to clean-up...
 */

OPRandom::~OPRandom ()
{
	OP_METHOD (~OPRandom, 0);
	
}


/*
 *	Store the parameters of the random generator, but not
 *	its seed.
 */

void
OPRandom::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPRandom max scale offset>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->max << this->scale << this->offset;
	storage << OPStorageSync::end;
}

void
OPRandom::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	Card32 max    = 0;
	Real   scale  = 0;
	Real   offset = 0;
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> max >> scale >> offset;
	storage >> OPStorageSync::end;
	
	this->SetMaximum (max);
	this->SetScaleOffset (scale, offset);
}


/*
 *	Generate a random value.
 */

void
OPRandom::RefreshValue () const
{
	OP_METHOD (RefreshValue, 0);
	
	Card32 x;
	
	do {
		
		//	We assume that Sedgewick and Knuth know what they are doing.
		//	This should produce some evenly distributed number in the
		//	range 0..99'999'999 (MAX_VALUE-1).
		
		Card32 j = this->index++ % 55;
		
		this->accumulate[j] = (this->accumulate[(j+23) % 55]
							 + this->accumulate[(j+54) % 55]) % MAX_VALUE;
		
		x = this->accumulate[j];
		
	} while (x >= this->max_round);
	
	//	The result we got can now be reduced to the interval of interest.
	
	x %= this->max_modulo;
	
	this->type = OP_TYPE_REAL;
	this->value.the_real = (Real)(x) * this->scale + this->offset;
}


/*
 *	The maximum must be in the range [1..99'999'999].
 */

Bool
OPRandom::SetMaximum (Card32 max)
{
	OP_METHOD (SetMaximum, 1);
	OP_ARG (1, Card32, max);
	
	if (max >= MAX_VALUE) max = MAX_VALUE-1;
	
	Card32 num = MAX_VALUE / (max+1);
	
	this->max        = max;
	this->max_round  = num * (max+1);
	this->max_modulo = max+1;
	
	return TRUE;
}


/*
 *	The scale and offset are by default set to 1.0 and 0.0
 *	respectively.
 */

Bool
OPRandom::SetScaleOffset (Real scale, Real offset)
{
	OP_METHOD (SetScaleOffset, 2);
	OP_ARG (1, Real, scale);
	OP_ARG (2, Real, offset);
	
	this->scale  = scale;
	this->offset = offset;
	
	return TRUE;
}

Bool
OPRandom::SetScale (Real scale)
{
	OP_METHOD (SetScale, 1);
	OP_ARG (1, Real, scale);
	
	this->scale  = scale;
	
	return TRUE;
}

Bool
OPRandom::SetOffset (Real offset)
{
	OP_METHOD (SetOffset, 1);
	OP_ARG (1, Real, offset);
	
	this->offset = offset;
	
	return TRUE;
}


/*
 *	Return the random generator's parameters.
 */

Card32
OPRandom::ReturnMaximum () const
{
	OP_METHOD (ReturnMaximum, 0);
	return this->max;
}

Real
OPRandom::ReturnScale () const
{
	OP_METHOD (ReturnScale, 0);
	return this->scale;
}

Real
OPRandom::ReturnOffset () const
{
	OP_METHOD (ReturnOffset, 0);
	return this->offset;
}


/*
 *	Special modulo 100'000'000 multiplication which takes
 *	care of overflows.
 */

Card32
OPRandom::Multiply (Card32 p, Card32 q)
{
	OP_MEMBER (Multiply, 2);
	OP_ARG (1, Card32, p);
	OP_ARG (2, Card32, q);
	
	static const Card32	m  = MAX_VALUE;
	static const Card32 m1 = MAX_VALUE_ROOT;
	
	Card32 p1 = p / m1;
	Card32 p0 = p % m1;
	Card32 q1 = q / m1;
	Card32 q0 = q % m1;
	
	return (((p0*q1+p1*q0) % m1)*m1 + p0*q0) % m;
}

