/*
 *	rect.cxx
 *
 *	Rectangle related operations. This implements only the ones
 *	which are not being defined as inline methods.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/rect.h>
#include <OPaC/values.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_TYPECLASS_VALUE (OPRect, OP_TYPE_RECT);

/****************************************************************/

const OPRect		OPRect::none;


OPRect::OPRect (OPValueIn value)
{
	const OPRect& rect = value->ReturnRect ();
	this->x1 = rect.x1;
	this->y1 = rect.y1;
	this->x2 = rect.x2;
	this->y2 = rect.y2;
	this->is_empty = rect.is_empty;
}

OPRect&
OPRect::operator = (OPValueIn value)
{
	const OPRect& rect = value->ReturnRect ();
	this->x1 = rect.x1;
	this->y1 = rect.y1;
	this->x2 = rect.x2;
	this->y2 = rect.y2;
	this->is_empty = rect.is_empty;
	return (*this);
}

/*
 *	Compute the union of the specified rectangle with the
 *	current one.
 */

OPRect&
OPRect::operator += (OPRectIn rect)
{
	if (this->IsEmpty () || rect.IsEmpty ()) return *this;
	if (rect.x1 < this->x1) this->x1 = rect.x1;
	if (rect.x2 > this->x2) this->x2 = rect.x2;
	if (rect.y1 < this->y1) this->y1 = rect.y1;
	if (rect.y2 > this->y2) this->y2 = rect.y2;
	return *this;
}


	
/*
 *	Compute the intersection of the specified rectangle
 *	with the current one.
 */

OPRect&
OPRect::operator *= (OPRectIn rect)
{
	if (rect.IsEmpty ()) {
		return this->MakeEmpty ();
	}

	if (this->IsEmpty ()) {
		return *this;
	}

	if (rect.x1 > this->x1) this->x1 = rect.x1;
	if (rect.x2 < this->x2) this->x2 = rect.x2;
	if (rect.y1 > this->y1) this->y1 = rect.y1;
	if (rect.y2 < this->y2) this->y2 = rect.y2;

	return (this->x1 > this->x2) || (this->y1 > this->y2)
		 ? (this->MakeEmpty ())
		 : (*this);
}



/*
 *	Save the rectangle to the storage.
 */

OPStorageIn
operator << (OPStorageIn storage, OPRectIn rect)
{
	if (rect.is_empty) {
		
		storage << (OPCoord)(0);
		storage << (OPCoord)(0);
		storage << (OPCoord)(0);
		storage << (OPCoord)(0);
		
	} else {
		
		storage << rect.x1;
		storage << rect.y1;
		storage << rect.x2;
		storage << rect.y2;
	}
	
	return storage;
}



/*
 *	Restore the rectangle from the storage.
 */

OPStorageIn
operator >> (OPStorageIn storage, OPRectOut rect)
{
	storage >> rect.x1;
	storage >> rect.y1;
	storage >> rect.x2;
	storage >> rect.y2;
	
	rect.is_empty = ( (rect.x1 == rect.y1)
				   && (rect.x2 == rect.y2)
				   && (rect.x1 == (OPCoord)(0))
				   && (rect.x2 == (OPCoord)(0)) ) ? TRUE : FALSE;
	
	return storage;
}

