/*
 *	scrlistpane.cxx
 *
 *	...
 *
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/grail.h>
#include <OPaC/new.h>
#include <OPaC/cellgrid.h>
#include <OPaC/scrlistpane.h>
#include <OPaC/scroller.h>
#include <OPaC/targetmacro.h>
#include <OPaC/textcaption.h>
#include <OPaC/values.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPScrollListPane);
	
/****************************************************************/

OP_TARGET_OBSERVER (OPScrollListPane, HandleScroll);

/****************************************************************/

/*
 *	Initialise an empty scroll list.
 */

OPScrollListPane::OPScrollListPane ()
{
	OP_METHOD (OPScrollListPane, 0);
	
	this->scroller       = 0;
	this->spacing        = 4;
	this->caption_height = 24;
	
	this->view_array     = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPArray);
	this->caption_array  = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPArray);
	
	this->total_num      = 0;
	this->visible_num    = 0;
	this->left_index     = 0;
}


/*
 *	Free a scroll list.
 */

OPScrollListPane::~OPScrollListPane ()
{
	OP_METHOD (~OPScrollListPane, 0);

	this->view_array->SetSize (0);
	this->caption_array->SetSize (0);
	
	this->view_array    = 0;
	this->caption_array = 0;
}


/*
 *
 */

void
OPScrollListPane::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPScrollListPane scroller spacing caption_height"
							   " total_num visible_num left_index view_array caption_array>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->scroller;
	storage << this->spacing    << this->caption_height;
	storage << this->total_num  << this->visible_num << this->left_index;
	storage << this->view_array << this->caption_array;
	storage << OPStorageSync::end;
}

void
OPScrollListPane::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->scroller;
	storage >> this->spacing    >> this->caption_height;
	storage >> this->total_num  >> this->visible_num >> this->left_index;
	storage >> this->view_array >> this->caption_array;
	storage >> OPStorageSync::end;
	
	this->is_geom_ok = FALSE;
}


/*
 *	Prepare the scroll list for its deletion.
 */

void
OPScrollListPane::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->view_array->Delete ();
	this->caption_array->Delete ();
	
	super::Delete ();
}



/*
 *	Called when the geometry changes.
 */

void
OPScrollListPane::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
	
	this->UpdateGeometry ();
}

void
OPScrollListPane::UpdateGeometry ()
{
	OP_METHOD (UpdateGeometry, 0);
	
	if (this->is_geom_ok == FALSE) {
		
		this->UpdateScroller ();
		this->UpdateColumns ();
		
		this->is_geom_ok = TRUE;
	}
}


/*
 *	Compute the new position and parameters of the scroller.
 *	This must be called when the number of columns changes,
 *	the shown columns scroll or when the geometry changes.
 */

void
OPScrollListPane::UpdateScroller ()
{
	OP_METHOD (UpdateScroller, 0);
	
	if (this->scroller.IsValid ()) {
		
		OPCoord dx = this->box.DX ();
		OPCoord dy = this->scroller->ReturnBreadth ();
		OPCoord oy = this->box.DY () - dy;
		
		OPCoord x  = this->box.X1 ();
		OPCoord y  = this->box.Y1 () + oy;
		
		this->scroller->Initialise (x, y, dx, TRUE);
		
		Real pos   = 0.0;
		Real ratio = 1.0;
		
		if (this->visible_num < this->total_num) {
			pos   = (Real)(this->left_index) / (Real)(this->total_num - this->visible_num);
			ratio = (Real)(this->visible_num) / (Real)(this->total_num);
		}
		
		this->scroller->SetVisibleValue (pos);
		this->scroller->SetVisibleRatio (ratio);
	}
}


/*
 *	Update the columns and position them at the proper place
 *	in the scroll pane.
 */

void
OPScrollListPane::UpdateColumns ()
{
	OP_METHOD (UpdateColumns, 0);
	
	if (this->visible_num == 0) {
		return;
	}
	
	Count index = 0;
	
	OPCoord wx = (this->box.DX () - (this->visible_num-1) * this->spacing) / this->visible_num;
	OPCoord ox = this->box.X1 ();
	OPCoord oy = this->caption_height;
	OPCoord dy = this->box.DY ();
	OPCoord sy = (this->scroller.IsValid ()) ? this->scroller->ReturnBreadth () : 0;
	
	if (oy != 0) oy += this->spacing;
	if (sy != 0) sy += this->spacing;
	
	dy -= oy + sy;

	
	for (index = 0; index < this->total_num; index++) {
		
		if ( (index < this->left_index)
		  || (index >= this->left_index + this->visible_num) ) {
			
			OPTextCaptionRef& caption = this->caption_array[index];
			OPViewRef& view = this->view_array[index];
			
			if (caption.IsValid ()) {
				caption->SetParent (OPViewNull);
			}
			
			if (view.IsValid ()) {
				view->SetParent (OPViewNull);
			}
		}
	}
		
	for (index = 0; index < this->visible_num; index++) {
		
		OPTextCaptionRef& caption = this->caption_array[this->left_index + index];
		OPViewRef& view = this->view_array[this->left_index + index];
		
		if (caption.IsValid ()) {
			caption->SetParent (this);
			caption->SetWindow (this->window);
			caption->SetOrigin (ox, this->box.Y1 ());
			caption->SetSize (wx, this->caption_height);
		}
		
		if (view.IsValid ()) {
			view->SetParent (this);
			view->SetWindow (this->window);
			view->SetOrigin (ox, this->box.Y1 () + oy);
			view->SetSize (wx, dy);
		}

		ox += wx;
		ox += this->spacing;
	}
}



/*
 *	Redraw the pane. This involves drawing of the scroller, the
 *	lists and the captions (they are not registered as children
 *	so they won't be drawn automatically).
 */

void
OPScrollListPane::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	if (this->is_geom_ok == FALSE) {
		this->UpdateGeometry ();
	}
	
	super::Draw (port, rect, what);
	
	Count index = 0;
	
	while (index < this->visible_num) {
		
		OPTextCaptionRef& caption = this->caption_array[this->left_index + index];
		OPViewRef& view = this->view_array[this->left_index + index];
		
		if (caption.IsValid ()) caption->Draw (port, rect);
		if (view.IsValid ()) view->Draw (port, rect);
		
		index++;
	}
}


/*
 *	...
 */

Bool
OPScrollListPane::Detect (OPCoord x, OPCoord y, OPViewOut detected)
{
	OP_METHOD (Detect, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPViewOut, detected);
	
	if (this->is_geom_ok == FALSE) {
		this->UpdateGeometry ();
	}
	
	Count index = 0;
	
	while (index < this->visible_num) {
		
		OPTextCaptionRef& caption = this->caption_array[this->left_index + index];
		OPViewRef& view = this->view_array[this->left_index + index];
		
		if (caption.IsValid () && caption->Detect (x, y, detected)) return TRUE;
		if (view.IsValid () && view->Detect (x, y, detected)) return TRUE;
		
		index++;
	}
	
	return super::Detect (x, y, detected);
}


/*
 *	Define the scroller to be used in order to specify the
 *	range of cells to be displayed.
 */

Bool
OPScrollListPane::SetScroller (OPScrollerIn scroller)
{
	OP_METHOD (SetScroller, 1);
	OP_ARG (1, OPObjectIn, scroller);
	
	if (this->scroller.IsValid ()) {
		this->scroller->DetachObserver (this, "Scroll", "HandleScroll");
		this->scroller->SetParent (OPViewNull);
	}
	
	if (scroller.IsValid ()) {
		scroller->AttachObserver (this, "Scroll", "HandleScroll");
		scroller->SetParent (this);
	}
	
	this->scroller = scroller;
	this->is_geom_ok = FALSE;
	
	return TRUE;
}

Bool
OPScrollListPane::SetContents (Count index, OPViewIn view)
{
	OP_METHOD (SetScrollList, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPViewIn, view);
	
	if (index < this->total_num) {
		
		if (this->view_array[index].IsValid ()) {
			this->view_array[index]->SetParent (OPViewNull);
		}
		
		this->is_geom_ok = FALSE;
		this->view_array[index] = view;
		
		if (view.IsValid ()) {
			if ( (index >= this->left_index)
			  && (index < this->left_index + this->visible_num) ) {
				view->SetParent (this);
			} else {
				view->SetParent (OPViewNull);
			}
		}
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPScrollListPane::SetCaption (Count index, OPTextCaptionIn caption)
{
	OP_METHOD (SetCaption, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPTextCaptionIn, caption);
	
	if (index < this->total_num) {
		
		if (this->caption_array[index].IsValid ()) {
			this->caption_array[index]->SetParent (OPViewNull);
		}
		
		this->is_geom_ok = FALSE;
		this->caption_array[index] = caption;
		
		if (caption.IsValid ()) {
			if ( (index >= this->left_index)
			  && (index < this->left_index + this->visible_num) ) {
				caption->SetParent (this);
			} else {
				caption->SetParent (OPViewNull);
			}
		}
		
		return TRUE;
	}
	
	return FALSE;
}


Bool
OPScrollListPane::SetVisibleColumns (Count num)
{
	this->visible_num = num;
	this->is_geom_ok  = FALSE;
	return TRUE;
}

Bool
OPScrollListPane::SetTotalColumns (Count max)
{
	if (max < this->total_num) {
		Count index = max;
		while (index < this->total_num) {
			if (this->caption_array[index].IsValid ()) {
				this->caption_array[index]->SetParent (OPViewNull);
				this->caption_array[index]->Delete ();
			}
			if (this->view_array[index].IsValid ()) {
				this->view_array[index]->SetParent (OPViewNull);
				this->view_array[index]->Delete ();
			}
			index++;
		}
	}
	
	this->caption_array->SetSize (max);
	this->view_array->SetSize (max);
	this->total_num  = max;
	this->is_geom_ok = FALSE;
	return TRUE;
}

Bool
OPScrollListPane::ShowColumn (Count index, int relative)
{
	Count start = index;
	Count left  = this->left_index;
	Count right = this->left_index + this->visible_num - 1;
	
	if (index < this->total_num) {
		if (relative == 0) {
			
			//	Never mind where the column is. All we want is have it
			//	displayed in the pane...
			
			if ( (index >= left)
			  && (index <= right) ) {
				return TRUE;
			}
			
			relative = left - index;
		}

		if (relative < 0) {
			start = index;
		} else {
			start = index - this->visible_num + 1;
			if (start < 0) start = 0;
		}
		
		this->left_index = start;
		this->is_geom_ok = FALSE;
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Scroll the pane as specified by the scroller.
 */

void
OPScrollListPane::HandleScroll (OPObjectIn object, Card32 id)
{
	OP_METHOD (HandleScroll, 2);
	OP_ARG (1, OPObjectIn, object);
	OP_ARG (2, Card32, id);
	
	static OPAtomRef name_value = "visible_value";
	
	//	Ask the scroll bar about the current visible value. Show the
	//	specified cell.
	
	OPValueRef value;
	
	if ( (id == 0)
	  && (object->QueryData (name_value, value)) ) {
		
		Real v  = value;
		int pos = (int)(v * (this->total_num - this->visible_num));
		if (pos < 0) pos = 0;

		this->ShowColumn (pos, -1);
		this->Display ();
	}
}

