/*
 *	selview.cxx
 *
 *	The Selection View is used to highilight the views which
 *	have been selected by the user interface builder.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/app.h>
#include <OPaC/buildmisc.h>
#include <OPaC/event.h>
#include <OPaC/grail.h>
#include <OPaC/gridview.h>
#include <OPaC/lookman.h>
#include <OPaC/message.h>
#include <OPaC/new.h>
#include <OPaC/selview.h>
#include <OPaC/springboard.h>
#include <OPaC/targetmacro.h>
#include <OPaC/window.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPSelView);

/****************************************************************/

OP_TARGET_CLASS (OPSelView);
OP_TARGET_ARG (OPObjectRef, object, "selected view");
OP_TARGET_CALL (CommitUIEdit (object));
OP_TARGET_END ();

/****************************************************************/

#define	NO_CORNER			0
#define	TOP_LEFT_CORNER		1
#define	TOP_RIGHT_CORNER	2
#define	BOTTOM_RIGHT_CORNER	3
#define	BOTTOM_LEFT_CORNER	4

#define	CORNER_WIDTH		4

/****************************************************************/

/*
 *	Setup the selection view.
 */

OPSelView::OPSelView ()
{
	OP_METHOD (OPSelView, 0);
	
	this->state |= OP_VIEW_OK_UI_EDIT | OP_VIEW_OK_UI_MOVE;
}



/*
 *	End of selection. If still some view selected, first
 *	release it !
 */

OPSelView::~OPSelView ()
{
	OP_METHOD (~OPSelView, 0);
	OP_ASSERT (this->view.IsInvalid ());
	OP_ASSERT (this->object.IsInvalid ());
}                             




/*
 *	Store the selection to the archive. But why ?
 */

void
OPSelView::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPSelView view object>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->view << this->object;
	storage << OPStorageSync::end;
}



/*
 *	Restore the selection from the archive.
 */

void
OPSelView::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->view >> this->object;
	storage >> OPStorageSync::end;
}


void
OPSelView::NotifySelectBuilder (OPAtomIn action)
{
	OP_METHOD (NotifySelectBuilder, 1);
	OP_ARG (1, OPAtomIn, action);
	
	if ( (this->state & OP_VIEW_IS_ACTIVE)
	  && (this->builder.IsValid ()) ) {
		
		static OPAtomRef sel_object = OP_BUILDER_OBJECT;
		static OPAtomRef sel_view   = OP_BUILDER_SELVIEW;
		
		this->builder->DefineData (sel_object, this->object);
		this->builder->DefineData (sel_view, this);
		
		if (action.IsValid ()) {
			this->builder->Action (action);
		}
	}
}

void
OPSelView::NotifyDeselectBuilder ()
{
	OP_METHOD (NotifyDeselectBuilder, 0);
	
	if (this->builder.IsValid ()) {
		
		OPObjectRef view;
		OPObjectRef self = this;
		
		static OPAtomRef sel_object = OP_BUILDER_OBJECT;
		static OPAtomRef sel_view   = OP_BUILDER_SELVIEW;
		
		this->builder->QueryData (sel_view, view);
		
		if (view == self) {
			this->builder->DefineData (sel_object, OPObjectNull);
			this->builder->DefineData (sel_view, OPSelViewNull);
		}
	}
}


/*
 *	Prepare the instance for the automatic deletion.
 */

void
OPSelView::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->view.IsValid ()) {
		this->view->Delete ();
		this->view = OPViewNull;
	}
	
	if (this->object.IsValid ()) {
		this->object->Delete ();
		this->object = OPViewNull;
	}
	
	super::Delete ();
}



/*
 *	Select the specified view.
 */

Bool
OPSelView::BeginSelection (OPViewIn view)
{
	OP_METHOD (BeginSelection, 1);
	OP_ARG (1, OPViewIn, view);
	
	if ( (this->view.IsValid ())
	  || (view.IsInvalid ()) ) {
		return FALSE;
	}
	
	OPWindowRef window;
	OPRect      box;
	
	view->GetBox (box);
	view->GetWindow (window);
	
	if (view->DragAndDrop (box, this, OP_DDP_UI_EDIT_CHECK)) {
		
		//	Remember about the view which will be selected...
		
		this->builder = OPApp::CurrentApp ()->ReturnBuilder ();
		this->view    = view;
		this->object  = view;
		
		if (view->DragAndDrop (box, this, OP_DDP_UI_EDIT_COMMIT)) {
			
			OPMessageRef doNotify;
			
			doNotify = OP_NEW_FROM_ZONE (OPaC_temp_zone, OPMessage);
			doNotify->SetMethod (OP_BUILDER_SELECT_BEGIN);
			
			this->builder <- doNotify << this;
			
			return TRUE;
		}
	}
	
	this->view   = OPViewNull;
	this->object = OPObjectNull;
	
	return FALSE;
}


/*
 *	Call-back for the view which has just been selected by
 *	the BeginSelection method. It can provide another object
 *	instance than itself...
 */

Bool
OPSelView::CommitUIEdit (OPObjectIn object)
{
	OP_METHOD (CommitUIEdit, 1);
	OP_ARG (1, OPObjectIn, object);
	
	//	Select the view.
	
	this->box    = view->ReturnBox ();
	this->object = object;
	
	OPViewRef parent;
	OPViewRef sibling_prev;
	Bool      active = FALSE;
	
	OPViewRef active_prev = view->ReturnPrevActive ();
	OPViewRef active_next = view->ReturnNextActive ();
	
	active |= active_prev.IsValid ();
	active |= active_next.IsValid ();
	
	view->GetParent (parent);
	view->GetPrevSibling (sibling_prev);
	
	OPSpringBoardRef spring = parent->ReturnSpringBoard ();
	
	if ( (spring.IsValid ())
	  && (spring->FindView (this->view)) ) {
		spring->ReplaceView (this->view, this);
	}
	
	this->SetParent (parent);
	this->view->SetParent (this);
	
	//	Maintain the position in the active list and in the viewing list.
	//	Just remember that inserting after a "null" view inserts at the
	//	end of the list, which is what we want.
	
	if (active) this->SetActiveAfter (active_prev);
	this->ReorderAfter (sibling_prev);
	
	//	Handle the keyboard focus.
	
	if (this->view->ReturnState () & OP_VIEW_IS_ACTIVE) {
		
		//	If the view was currently the main responder -- it had the keyboard
		//	focus -- it will loose it !
		
		this->window->RemoveMainResponder ();
	}
	
	this->Display ();
	this->NotifySelectBuilder ();
	
	//	Restore a normal cursor, in case the object the user selected had
	//	changed the mouse cursor shape...
	
	OPLookMan::Current ()->SelectCursor (OP_CURSOR_NORMAL);
	
	return TRUE;
}


/*
 *	End of current selection.
 */

void
OPSelView::EndSelection ()
{
	OP_METHOD (EndSelection, 0);
	
	if (this->builder.IsInvalid ()) {
		this->builder = OPApp::CurrentApp ()->ReturnBuilder ();
	}
	
	if (this->view.IsValid ()) {
		
		OPWindowRef window = this->window;
		
		OPViewRef sibling_prev;
		Bool      active = FALSE;
		
		OPViewRef active_prev = view->ReturnPrevActive ();
		OPViewRef active_next = view->ReturnNextActive ();
		
		active |= active_prev.IsValid ();
		active |= active_next.IsValid ();
		
		OPSpringBoardRef spring = this->parent->ReturnSpringBoard ();
		
		if ( (spring.IsValid ())
		  && (spring->FindView (this)) ) {
			spring->ReplaceView (this, this->view);
		}
		
		this->GetPrevSibling (sibling_prev);
		this->view->SetParent (this->parent);
		
		this->SetParent (OPViewNull);
		this->SetWindow (OPWindowNull);
		
		if (active) this->view->SetActiveAfter (active_prev);
		this->view->ReorderAfter (sibling_prev);
		
		this->view   = OPViewNull;
		this->object = OPObjectNull;
		
		window->Redraw (this->ReturnBox ());
		
		OPMessageRef doNotify;
		
		doNotify = OP_NEW_FROM_ZONE (OPaC_temp_zone, OPMessage);
		doNotify->SetMethod (OP_BUILDER_SELECT_END);
		
		this->builder <- doNotify << this;
	}
}



/*
 *	Return reference to selected view or selected object.
 */

const OPViewRef& 
OPSelView::ReturnSelectedView () const
{
	OP_METHOD (ReturnSelectedView, 0);
	
	return this->view;
}

const OPObjectRef& 
OPSelView::ReturnSelectedObject () const
{
	OP_METHOD (ReturnSelectedObject, 0);
	
	return this->object;
}




/*
 *	Notify the view that it has become the main responder
 *	and redraw it.
 */

void
OPSelView::BecomeMain (Bool clicked)
{
	OP_METHOD (BecomeMain, 1);
	OP_ARG (1, Bool, clicked);
	
	super::BecomeMain (clicked);
	
	this->NotifySelectBuilder ();
}



/*
 *	Notify the view that it no longer is the main responder
 *	and redraw it.
 */

void
OPSelView::ResignMain ()
{
	OP_METHOD (ResignMain, 0);
	
	super::ResignMain ();
	
	this->NotifyDeselectBuilder ();
}




/*
 *	Detect the mouse clicks.
 */

Bool
OPSelView::Detect (OPCoord x, OPCoord y, OPViewOut view)
{
	OP_METHOD (Detect, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPViewOut, view);
	
	if (this->box.IsInside (x, y)) {
		view = this;
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Draw the frame of the selection.
 */

void
OPSelView::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		const OPAdornerRef& adorner = port->ReturnAdorner ();
		Bool active = (this->state & OP_VIEW_IS_ACTIVE) ? TRUE : FALSE;
		OPLookColor color = active ? OP_COLOR_UI_SEL_ACTIVE : OP_COLOR_UI_SELECT;
		
		adorner->SetForeColor (color);
		
		port->DrawRect (this->box);
		
		OPCoord x1 = this->box.X1 ();
		OPCoord x2 = this->box.X2 ();
		OPCoord y1 = this->box.Y1 ();
		OPCoord y2 = this->box.Y2 ();
		
		//	The UI edition frame has small handles in each corner :
		
		port->FillRect (x1, y1, x1+CORNER_WIDTH, y1+CORNER_WIDTH);
		port->FillRect (x2-CORNER_WIDTH, y1, x2, y1+CORNER_WIDTH);
		port->FillRect (x1, y2-CORNER_WIDTH, x1+CORNER_WIDTH, y2);
		port->FillRect (x2-CORNER_WIDTH, y2-CORNER_WIDTH, x2, y2);
	}
}



/*
 *	When a view is being selected for UI edition, it will
 *	have to track the frame's handles :
 */

void
OPSelView::ButtonPress (OPEventIn event)
{
	OP_METHOD (ButtonPress, 1);
	OP_ARG (1, OPEventIn, event);
	
	this->window->SetMainResponder (this, TRUE);
	
	//	Control-clicking starts a connection.
	
	if ( (this->state & OP_VIEW_IS_ENABLED)
	  && (this->is_being_clicked == FALSE)
	  && ((event->ReturnKey ().modifiers & OP_EV_MOD_MASK) == OP_EV_CONTROL)
	  && (event->ReturnChange () == OP_EV_BUTTON_LEFT) ) {
		this->NotifySelectBuilder (OP_BUILDER_CONNECT);
		return;
	}
	
	
	//	Shift-clicking shows the edition menu.
	
	if ( (this->state & OP_VIEW_IS_ENABLED)
	  && (this->is_being_clicked == FALSE)
	  && ((event->ReturnKey ().modifiers & OP_EV_MOD_MASK) == OP_EV_SHIFT)
	  && (event->ReturnChange () == OP_EV_BUTTON_LEFT) ) {
		this->NotifySelectBuilder (OP_BUILDER_MENU);
		return;
	}	
	
	
	//	Clicking on a handle or on the body starts a dragging operation.
	
	if ( (this->state & OP_VIEW_IS_ENABLED)
	  && (this->is_being_clicked == FALSE)
	  && (event->ReturnChange () == OP_EV_BUTTON_LEFT) ) {
		
		OPRelCoord x  = event->ReturnX ();
		OPRelCoord y  = event->ReturnY ();
		
		OPRelCoord x1 = (OPRelCoord)(this->box.X1 () + CORNER_WIDTH + 1);
		OPRelCoord y1 = (OPRelCoord)(this->box.Y1 () + CORNER_WIDTH + 1);
		OPRelCoord x2 = (OPRelCoord)(this->box.X2 () - CORNER_WIDTH - 1);
		OPRelCoord y2 = (OPRelCoord)(this->box.Y2 () - CORNER_WIDTH - 1);
		
		if ( (x < x1) && (y < y1) ) {
			
			this->corner_num = TOP_LEFT_CORNER;
			this->dx = x - this->box.X1 ();
			this->dy = y - this->box.Y1 ();
			
		} else if ( (x < x1) && (y > y2) ) {
			
			this->corner_num = BOTTOM_LEFT_CORNER;
			this->dx = x - this->box.X1 ();
			this->dy = y - this->box.Y2 ();
			
		} else if ( (x > x2) && (y < y1) ) {
			
			this->corner_num = TOP_RIGHT_CORNER;
			this->dx = x - this->box.X2 ();
			this->dy = y - this->box.Y1 ();
			
		} else if ( (x > x2) && (y > y2) ) {
			
			this->corner_num = BOTTOM_RIGHT_CORNER;
			this->dx = x - this->box.X2 ();
			this->dy = y - this->box.Y2 ();
			
		} else {
			
			this->corner_num = NO_CORNER;
			this->dx = x - this->box.X1 ();
			this->dy = y - this->box.Y1 ();
		}
		
		OPViewRef self = this;
		
		this->window->SetDraggingView (self);
		this->ActionOn ();
	}
}



/*
 *	Drag a handle while the user moves around the mouse
 *	with the left button pressed.
 */

void
OPSelView::DragNotify (OPEventIn event)
{
	if (this->is_being_clicked) {
		
		OPRect old_box = this->box;
		
		OPCoord cx = event->ReturnX ();
		OPCoord cy = event->ReturnY ();
		
		OPCoord wx = (OPCoord)(0);
		OPCoord wy = (OPCoord)(0);
		
		this->window->GetSize (wx, wy);
		
		if (cx < (OPCoord)(0)) cx = (OPCoord)(0);
		if (cy < (OPCoord)(0)) cy = (OPCoord)(0);
		if (cx > wx) cx = wx;
		if (cy > wy) cy = wy;
		
		OPRelCoord xe = cx - this->dx;
		OPRelCoord ye = cy - this->dy;
		
		OPCoord max_dx = 0, max_dy = 0;
		OPCoord min_dx = 0, min_dy = 0;
		
		this->view->GetMaxSize (max_dx, max_dy);
		this->view->GetMinSize (min_dx, min_dy);
		
		OPCoord x  = this->box.X1 ();
		OPCoord y  = this->box.Y1 ();
		OPCoord dx = this->box.DX ();
		OPCoord dy = this->box.DY ();
		
		if ( (this->corner_num == TOP_LEFT_CORNER)
		  || (this->corner_num == TOP_RIGHT_CORNER) ) {
			
			//	Handle vertical movement & resizing of top object side
			
			OPRelCoord delta = y - ye + dy;
			OPRelCoord pos   = ye;
			OPRelCoord min_y = (OPRelCoord)(min_dy);
			OPRelCoord max_y = (OPRelCoord)(max_dy);
			
			if (delta < min_y) {
				pos  -= (min_dy - delta);
				delta = min_y;
			}
			
			if (delta > max_y) {
				pos  -= (max_dy - delta);
				delta = max_y;
			}
			
			y  = pos;
			dy = delta;
		}
		
		if ( (this->corner_num == BOTTOM_LEFT_CORNER)
		  || (this->corner_num == BOTTOM_RIGHT_CORNER) ) {
			
			//	Handle vertical movement & resizing of bottom object side
			
			OPRelCoord delta = ye - y;
			OPRelCoord min_y = (OPRelCoord)(min_dy);
			OPRelCoord max_y = (OPRelCoord)(max_dy);
			
			if (delta < min_y) delta = min_y;
			if (delta > max_y) delta = max_y;
			
			dy = delta;
		}
		
		if ( (this->corner_num == TOP_LEFT_CORNER)
		  || (this->corner_num == BOTTOM_LEFT_CORNER) ) {
			
			//	Handle horizontal movement & resizing of left object side
			
			OPRelCoord delta = x - xe + dx;
			OPRelCoord pos   = xe;
			OPRelCoord min_x = (OPRelCoord)(min_dx);
			OPRelCoord max_x = (OPRelCoord)(max_dx);
			
			if (delta < min_x) {
				pos  -= (min_dx - delta);
				delta = min_x;
			}
			
			if (delta > max_x) {
				pos  -= (max_dx - delta);
				delta = max_x;
			}
			
			x  = pos;
			dx = delta;
		}
		
		if ( (this->corner_num == TOP_RIGHT_CORNER)
		  || (this->corner_num == BOTTOM_RIGHT_CORNER) ) {
			
			//	Handle vertical movement & resizing of bottom object side
			
			OPRelCoord delta = xe - x;
			OPRelCoord min_x = (OPRelCoord)(min_dx);
			OPRelCoord max_x = (OPRelCoord)(max_dx);
			
			if (delta < min_x) delta = min_x;
			if (delta > max_x) delta = max_x;
			
			dx = delta;
		}
		
		if (this->corner_num == NO_CORNER) {
			
			OPGridViewRef grid;
			
			if (this->window->GetAttribute ("BuilderGrid", grid)) {
				grid->AlignOnGrid (xe, ye);
			}
			
			this->SetOrigin (xe, ye);
			
		} else {
			
			OPGridViewRef grid;
			
			if (this->window->GetAttribute ("BuilderGrid", grid)) {
				
				OPCoord x1 = x;
				OPCoord y1 = y;
				OPCoord x2 = x + dx;
				OPCoord y2 = y + dy;
				
				grid->AlignOnGrid (x1, y1);
				grid->AlignOnGrid (x2, y2);
				
				x  = x1;
				y  = y1;
				dx = x2 - x1; if (dx < min_dx) dx = min_dx;
				dy = y2 - y1; if (dy < min_dy) dy = min_dy;
			}
			
			this->SetOrigin (x, y);
			this->SetSize (dx, dy);
			this->view->SetSize (dx, dy);
		}
		
		old_box += this->box;
		
		this->window->Redraw (old_box);
		this->ReturnAdorner ()->Flush (OP_FLUSH_SYNC);
	}
}



/*
 *	The user released a mouse button. If it is the left
 *	one, finish the current frame dragging operation.
 */

void
OPSelView::ButtonRelease (OPEventIn event)
{
	OP_METHOD (ButtonRelease, 1);
	OP_ARG (1, OPEventIn, event);
	
	if ( (this->state & OP_VIEW_IS_ENABLED)
	  && (this->is_being_clicked == TRUE)
	  && (event->ReturnChange () == OP_EV_BUTTON_LEFT) ) {
		
		this->window->SetDraggingView (OPSelViewNull);
		this->ActionOff ();
	}
}



/*
 *	If DELETE or BACKSPACE have been pressed, then delete
 *	the selected view. This will be done by the garbage
 *	collector...
 */

void
OPSelView::KeyDown (OPEventIn event)
{
	OP_METHOD (KeyDown, 1);
	OP_ARG (1, OPEventIn, event);
	
	const OPKey& key = event->ReturnKey ();
	
	if ( (key.type == OP_KEY_DELETE)
	  || (key.type == OP_KEY_BACKSPACE) ) {
		
		this->NotifySelectBuilder (OP_BUILDER_DO_DELETE);
		return;
	}
	
	if (key.type == OP_KEY_RETURN) {
		this->NotifySelectBuilder (OP_BUILDER_DO_INSPECT);
		return;
	}
	
	super::KeyDown (event);
}
