/*
 *	tabview.cxx
 *
 *	The Tabbed View holds many views of which only one can be
 *	visible at a given time.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/tablist.h>
#include <OPaC/tabview.h>
#include <OPaC/window.h>
#include <OPaC/lookman.h>
#include <OPaC/container.h>
#include <OPaC/grail.h>
#include <OPaC/new.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPTabView);

/****************************************************************/

OP_TARGET_OBSERVER (OPTabView, OnTabChange);

/****************************************************************/


/*
 *	Initialise an empty tabbed list.
 */

OPTabView::OPTabView ()
{
	OP_METHOD (OPTabView, 0);
	
	this->height = 24;
	this->tabs   = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPTabList);
	
	OPObjectRef self = this;
	this->tabs->AttachObserver (self, "OnTabChange");
	
	this->need_update = TRUE;
}


/*
 *	Destroy the tabbed list. If we still have some tabbed texts
 *	tell them that we no longer exist.
 */

OPTabView::~OPTabView ()
{
	OP_METHOD (~OPTabView, 0);

	if (this->tabs.IsValid ()) {
		this->tabs->Delete ();
		this->tabs = OPTabListNull;
	}
}


/*
 *	Prepare the tabbed view for its destruction. Remove any
 *	links to other objects, such as the tabbed texts.
 */

void
OPTabView::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->tabs->Delete ();
	
	super::Delete ();
}



/*
 *	When the view's size changes, we will have to update
 *	the position and size of its subviews.
 */

void
OPTabView::UpdateGeometry (OPRectIn box)
{
	OP_METHOD (UpdateGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	OPRect rect = box;
	
	rect.MoveOrigin (2, this->height+1);
	rect.ChangeSize (-4, -this->height-3);
	
	super::UpdateGeometry (rect);
	
	this->need_update = TRUE;
}


/*
 *	Overridden to take into account the subviews stored in
 *	the tabs, which are not considered to be normal child
 *	views.
 */

Bool
OPTabView::SetWindow (OPWindowIn window)
{
	OP_METHOD (SetWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	
	//	If the window which has been specified is equal to the current window,
	//	don't do anything...
	
	if (this->window == window) {
		return TRUE;
	}
	
	if (super::SetWindow (window)) {
		
		this->tabs->SetWindow (window);
		this->MakeViewActive (this->active_view, this->active_index);
		
		return TRUE;
	}

	return FALSE;
}


/*
 *	Store the tabs.
 */

void
OPTabView::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPTabView tabs height>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->tabs;
	storage << this->height;
	storage << OPStorageSync::end;
}


/*
 *	Restore the tabs.
 */

void
OPTabView::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->tabs;
	storage >> this->height;
	storage >> OPStorageSync::end;
	
	this->need_update = TRUE;
}


/*
 *	When a tab has been activated, one of the views will
 *	have to become the main responder.
 */

void
OPTabView::NotifyTabActivated (Count active)
{
	OP_METHOD (NotifyTabActivated, 1);
	OP_ARG (1, Count, active);
	
	if (active) {
		
		//	Make the first active view of the new tab the main
		//	responder.
		
		OPViewRef   view;
		OPWindowRef window;
		
		if ( (this->views->Get (active-1, view))
		  && (this->GetWindow (window)) ) {
			
			view = view->ReturnFirstActive ();
			
			if (view.IsValid ()) {
				window->SetMainResponder (view, FALSE);
			}
		}
	}
}


/*
 *	This observer gets called when the tab changed.
 */

void
OPTabView::OnTabChange (OPObjectIn source, Card32 id)
{
	OP_METHOD (OnTabChange, 2);
	OP_ARG (1, OPObjectIn, source);
	OP_ARG (2, Card32, id);
	
	OPTabListRef tab = source;
	
	this->ActivateView (tab->ReturnActiveTab ());
	this->Display ();
}


/*
 *	Update the geometry of the tab list.
 */

void
OPTabView::UpdateTabs ()
{
	OP_METHOD (UpdateTabs, 0);
	
	this->tabs->SetParent (this);
	this->tabs->SetOrigin (this->box.X1 ()+1, this->box.Y1 ());
	this->tabs->SetSize (this->box.DX ()-1, this->height);
	
	this->MakeViewActive (this->active_view, this->active_index);
	
	this->need_update = FALSE;
}


/*
 *	Define the active tab. The tab # 0 specifies that no
 *	tab is active at all... We don't display anything in
 *	the tab view in that case.
 */

Bool
OPTabView::MakeViewActive (OPViewIn view, Count index)
{
	OP_METHOD (MakeViewActive, 2);
	OP_ARG (1, OPViewIn, view);
	OP_ARG (2, Count, index);
	
	if (this->active_index == index) {
		return TRUE;
	}
	
	this->tabs->SetActiveTab (index);
	super::MakeViewActive (view, index);
	this->NotifyTabActivated (index);
	this->UpdateGeometry (this->box);
	
	return TRUE;
}


/*
 *	Change the height of the tab list.
 */

Bool
OPTabView::SetHeight (OPCoord height)
{
	OP_METHOD (SetHeight, 1);
	OP_ARG (1, OPCoord, height);

	this->height = height;
	this->need_update = TRUE;
	
	return TRUE;
}


/*
 *	Return the heigth reserved for the tab list.
 */

OPCoord
OPTabView::ReturnHeight () const
{
	OP_METHOD (ReturnHeight, 0);
	
	return this->height;
}


/*
 *	Return the tab list. This has to be used in order to define
 *	the text tags of the tab list.
 */

Bool
OPTabView::GetTabs (OPTabListOut tabs) const
{
	OP_METHOD (GetTabs, 1);
	OP_ARG (1, OPTabListOut, tabs);

	tabs = this->tabs;
	return tabs.IsValid ();
}


/*
 *	Just draw...
 */

void
OPTabView::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	if (this->need_update) {
		this->UpdateTabs ();
	}

	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		const OPAdornerRef& adorner = port->ReturnAdorner ();
		
		adorner->SetForeColor (OP_COLOR_WHITE);
		port->DrawLineDY (this->box.X1 ()+1, this->box.Y1 ()+this->height-1, this->box.DX ()-this->height-4);
		
		adorner->SetForeColor (OP_COLOR_LGRAY);
		port->DrawLineDY (this->box.X1 (), this->box.Y1 ()+this->height-1, this->box.DX ()-this->height-3);
		
		adorner->SetForeColor (OP_COLOR_DGRAY);
		port->DrawLineDX (this->box.X1 ()+1, this->box.Y2 ()-1, this->box.DX ()-2);
		port->DrawLineDY (this->box.X2 ()-1, this->box.Y1 ()+this->height-2, this->box.DX ()-this->height-2);

		adorner->SetForeColor (OP_COLOR_BLACK);
		port->DrawLineDX (this->box.X1 (), this->box.Y2 (), this->box.DX ());
		port->DrawLineDY (this->box.X2 (), this->box.Y1 ()+this->height-3, this->box.DX ()-this->height-1);
	}
}

