/*
 *	texttag.cxx
 *
 *	Simple Text Tag class, used for standard messages in a user
 *	interface. A text tag will usually be bound into another
 *	widget (eg. a button).
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/texttag.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/pubmacro.h>
#include <OPaC/string.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPTextTag);

/****************************************************************/

static const OPaC_BitsetDataDef local_h_mode_info[] = {
	{ OP_TEXT_HPOS_MASK,   "Horizontal", OP_BITSET_DATA_MASK },
	{ OP_TEXT_HPOS_LEFT,   "left",		 OP_BITSET_DATA_VALUE },
	{ OP_TEXT_HPOS_CENTER, "center",	 OP_BITSET_DATA_VALUE },
	{ OP_TEXT_HPOS_RIGHT,  "right",		 OP_BITSET_DATA_VALUE },
	{ OP_TEXT_HPOS_ALIGN,  "align",		 OP_BITSET_DATA_VALUE }
};

static const OPaC_BitsetDataDef local_v_mode_info[] = {
	{ OP_TEXT_VPOS_MASK,   "Vertical",	 OP_BITSET_DATA_MASK },
	{ OP_TEXT_VPOS_BASE,   "baseline",	 OP_BITSET_DATA_VALUE },
	{ OP_TEXT_VPOS_TOP,    "top",		 OP_BITSET_DATA_VALUE },
	{ OP_TEXT_VPOS_CENTER, "center",	 OP_BITSET_DATA_VALUE },
	{ OP_TEXT_VPOS_BOTTOM, "bottom",	 OP_BITSET_DATA_VALUE }
};

/****************************************************************/

OP_PUBLISH_SLOT (OPTextTag, alignment,     ReturnMode,         SetMode, "Alignment", Card32);
OP_PUBLISH_SLOT (OPTextTag, left_space,    ReturnLeftSpace,    SetLeftSpace,  "Left space",  Real);
OP_PUBLISH_SLOT (OPTextTag, right_space,   ReturnRightSpace,   SetRightSpace, "Right space", Real);
OP_PUBLISH_SLOT (OPTextTag, space_enabled, ReturnSpaceEnabled, EnableSpace, "Enable space", Bool);

OP_PUBLISH_BITSET_INFO (OPTextTag, alignment, local_h_mode_info);
OP_PUBLISH_BITSET_INFO (OPTextTag, alignment, local_v_mode_info);

/****************************************************************/

/*
 *	Create an empty text tag. No text and no font are
 *	currently defined.
 */

OPTextTag::OPTextTag ()
{
	OP_METHOD (OPTextTag, 0);
	
	this->font = OPGrailFontNull;
	this->mode = OP_TEXT_HPOS_CENTER | OP_TEXT_VPOS_CENTER;
	
	this->text_color     = OPColorPixel (0.0, 0.0, 0.0);
	this->use_auto_color = TRUE;
	this->use_space      = FALSE;
	this->left_space     = 0;
	this->right_space    = 0;
}



/* 
 *	Destroy a text tag. This is easy.
 */

OPTextTag::~OPTextTag ()
{
	OP_METHOD (~OPTextTag, 0);
	
	this->font = OPGrailFontNull;
}



/*
 *	Delete a text tag : prepare for automatic destruction
 *	by the garbage collector.
 */

void
OPTextTag::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->font = OPGrailFontNull;
	
	super::Delete ();
}



/*
 *	Store the text tag : string, font and mode plus the
 *	view's definition.
 */

void
OPTextTag::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	static OPStorageSync info ("<OPTextTag font mode use_space left_space right_space>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->font << this->mode;
	storage << this->use_space;
	storage << this->left_space;
	storage << this->right_space;
	storage << OPStorageSync::end;
}



/*
 *	Restore the text tag : string, font and mode plus the
 *	view's definition.
 */

void
OPTextTag::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->font >> this->mode;
	storage >> this->use_space;
	storage >> this->left_space;
	storage >> this->right_space;
	storage >> OPStorageSync::end;
}



/*
 *	Define the font to use with the text tag. This will
 *	not redraw the text tag. You have to do it yourself.
 */

Bool
OPTextTag::SetFont (OPGrailFontIn font)
{
	OP_METHOD (SetFont, 1);
	OP_ARG (1, OPGrailFontIn, font);
	
	this->font = font;
	return TRUE;
}



Bool
OPTextTag::SetColor (OPColorPixel pixel)
{
	OP_METHOD (SetColor, 1);
	OP_ARG (1, OPColorPixel, pixel);
	
	this->text_color     = pixel;
	this->use_auto_color = FALSE;
	
	return TRUE;
}


Bool
OPTextTag::SetAutomaticColor (Bool yes)
{
	OP_METHOD (SetAutomaticColor, 1);
	OP_ARG (1, Bool, yes);
	
	this->use_auto_color = yes;
	
	return TRUE;
}


/*
 *	Define the mode to be used to display the text :
 *
 *	- OP_TEXT_HPOS_.. horizontal alignment.
 *	- OP_TEXT_VPOS_.. vertical alignment.
 *
 *	This will not redraw the text tag...
 */

Bool
OPTextTag::SetMode (Card32 mode)
{
	OP_METHOD (SetMode, 1);
	OP_ARG (1, Card32, mode);
	
	this->mode = mode;
	return TRUE;
}


Bool
OPTextTag::SetLeftSpace (OPCoord space)
{
	OP_METHOD (SetLeftSpace, 1);
	OP_ARG (1, OPCoord, space);
	
	this->left_space = space;
	this->use_space  = TRUE;
	
	return TRUE;
}

Bool
OPTextTag::SetRightSpace (OPCoord space)
{
	OP_METHOD (SetRightSpace, 1);
	OP_ARG (1, OPCoord, space);
	
	this->right_space = space;
	this->use_space   = TRUE;
	
	return TRUE;
}

Bool
OPTextTag::EnableSpace (Bool yes)
{
	OP_METHOD (EnableSpace, 1);
	OP_ARG (1, Bool, yes);
	
	this->use_space  = yes;
	
	return TRUE;
}

/*
 *	Return the font used by the text tag. Default is the
 *	system font.
 */

Bool
OPTextTag::GetFont (OPGrailFontOut font) const
{
	OP_METHOD (GetFont, 1);
	OP_ARG (1, OPGrailFontOut, font);
	
	font = (this->font.IsValid ()) ? this->font
								   : OPLookMan::Current ()->ReturnSysFont ();
	return font.IsValid ();
}



OPColorPixel
OPTextTag::ReturnColor () const
{
	OP_METHOD (ReturnColor, 0);
	return this->text_color;
}


Bool
OPTextTag::ReturnAutomaticColor () const
{
	OP_METHOD (ReturnAutomaticColor, 0);
	return this->use_auto_color;
}


/*
 *	Return the mode to be used to display the text.
 */

Card32
OPTextTag::ReturnMode () const
{
	OP_METHOD (ReturnMode, 0);
	return this->mode;
}


OPCoord
OPTextTag::ReturnLeftSpace () const
{
	OP_METHOD (ReturnLeftSpace, 0);
	return this->left_space;
}


OPCoord
OPTextTag::ReturnRightSpace () const
{
	OP_METHOD (ReturnRightSpace, 0);
	return this->right_space;
}


Bool
OPTextTag::ReturnSpaceEnabled () const
{
	OP_METHOD (ReturnSpaceEnabled, 0);
	return this->use_space;
}


/*
 *	Return the text's minimal size.
 */

void
OPTextTag::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	OPGrailFontRef font;
	OPStringRef    text;

	this->GetText (text);
	this->GetFont (font);
	
	dx = font->ReturnWidth (text);
	dy = font->ReturnAscender () - font->ReturnDescender ();
	
	if (this->use_space) {
		dx += this->left_space;
		dx += this->right_space;
	}
	
	OPCoord dx1 = 0;
	OPCoord dy1 = 0;
	
	super::GetMinSize (dx1, dy1);
	
	if (dx1 > dx) dx = dx1;
	if (dy1 > dy) dy = dy1;
}



/*
 *	Draw the text tag.
 */

void
OPTextTag::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		const OPAdornerRef& adorner = port->ReturnAdorner ();
		
		OPStringRef    text;
		OPGrailFontRef font;
		
		this->GetFont (font);
		this->GetText (text);
		
		if (text.IsValid ()) {
			Count       length = text->CountCharacters ();
			OPLookColor color  = (this->state & OP_VIEW_IS_ENABLED)
								 ? OP_COLOR_VIEW_NORMAL : OP_COLOR_VIEW_DIM;
			
			//	Now show the text... There will never be a frame around a text tag,
			//	so there will be no margins between the text and the bounding-box...
			
			if (this->use_auto_color) {
				adorner->SetForeColor (color);
			} else {
				port->SetForeColor (this->text_color);
			}
			
			OPRect box = this->box;
			
			if (this->use_space) {
				box.X1 () += this->left_space;
				box.X2 () -= this->right_space;
			}
			
			port->SetFont (font);
			port->ShowText (box, text, 0, length, (Card8)(this->mode));
		}
	}
}
