/*
 *	typeinfo.cxx
 *
 *	Implements a run-time type information classes.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/new.h>
#include <OPaC/string.h>
#include <OPaC/typeinfo.h>

#include <string.h>

/****************************************************************/

OP_IMPLEMENTATION_FILE;
OP_IMPLEMENT_CLASS (OPTypeInfo);

/****************************************************************/

OPTypeDef			type;
Bool				is_read_ok;
Bool				is_write_ok;

OPTypeInfo::OPTypeInfo ()
{
	OP_METHOD (OPTypeInfo, 0);
	
	this->type   = 0;
	this->bitset = OPBitsetInfoNull;
	
	this->is_read_ok  = FALSE;
	this->is_write_ok = FALSE;
}


OPTypeInfo::~OPTypeInfo ()
{
	OP_METHOD (~OPTypeInfo, 0);
}

void
OPTypeInfo::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	OPAtomRef name;
	
	if (this->type) {
		name = this->type->type;
	}
	
	static OPStorageSync info ("<OPTypeInfo name bitset is_read_ok is_write_ok>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << name << this->bitset << this->is_read_ok << this->is_write_ok;
	storage << OPStorageSync::end;
}


void
OPTypeInfo::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	OPAtomRef name;
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> name >> this->bitset >> this->is_read_ok >> this->is_write_ok;
	storage >> OPStorageSync::end;
	
	this->type = 0;
	
	if (name.IsValid ()) {
		OPTypeInfo::FindType (name, this->type);
	}
}

void
OPTypeInfo::Initialise (OPTypeDef type, OPBitsetInfoIn bitset, Bool rd, Bool wr)
{
	OP_METHOD (Initialise, 4);
	OP_ARG (1, OPTypeDef, type);
	OP_ARG (2, OPBitsetInfoIn, bitset);
	OP_ARG (3, Bool, rd);
	OP_ARG (4, Bool, wr);
	
	this->type   = type;
	this->bitset = bitset;
	
	this->is_read_ok  = rd;
	this->is_write_ok = wr;
}

void
OPTypeInfo::Initialise (OPRuntimeIn runtime)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, OPRuntimeIn, runtime);
	
	OPTypeDef type;
	
	if (OPTypeInfo::FindType (runtime->ReturnClassName (), type)) {
		this->Initialise (type);
	}
}


/****************************************************************/

OPaC_TypeInfo*  OPaC_TypeInfo::head  = 0;
OPaC_TypeInfo*  OPaC_TypeInfo::tail  = 0;
OPaC_TypeInfo** OPaC_TypeInfo::array = 0;
Count           OPaC_TypeInfo::num   = 0;

/****************************************************************/

typedef OPaC_TypeInfo* OPaC_TypeInfoPtr;

/****************************************************************/

void
OPaC_TypeInfo::Initialise ()
{
	if (OPaC_TypeInfo::array == 0) {
		OPaC_TypeInfo::array = OP_ALLOC_ARRAY (OPaC_TypeInfoPtr, OPaC_TypeInfo::num);
		OPaC_TypeInfo* iter = OPaC_TypeInfo::head;
		for (Count index = 0; index < OPaC_TypeInfo::num; index++) {
			OPaC_TypeInfo::array[index] = iter;
			iter = iter->next;
		}
	}
	
	this->name = this->type->type;
}

OPaC_TypeInfo::OPaC_TypeInfo (OPTypeDef type)
{
	this->type = type;
	this->next = 0;
	this->prev = OPaC_TypeInfo::head;
	(OPaC_TypeInfo::tail ? OPaC_TypeInfo::tail->next : OPaC_TypeInfo::head) = this;
	OPaC_TypeInfo::tail = this;
	OPaC_TypeInfo::num++;
}


Bool
OPTypeInfo::GetRuntimeInfo (OPRuntimeOut runtime) const
{
	OP_METHOD (GetRuntimeInfo, 1);
	OP_ARG (1, OPRuntimeOut, runtime);
	
	if ( (this->type)
	  && ( (this->type->simple == OP_TYPE_OBJECT)
	    || (this->type->simple == OP_TYPE_TEXT)
		|| (this->type->simple == OP_TYPE_ARRAY) ) ) {
		
		runtime = ((const OPObjectRef*)(this->type->more))[0];
		return TRUE;
	}
	
	return FALSE;
}


Count
OPTypeInfo::ReturnNumTypes ()
{
	return OPaC_TypeInfo::num;
}

OPTypeDef
OPTypeInfo::ReturnNthType (Count index)
{
	if (index < OPaC_TypeInfo::num) {
		return OPaC_TypeInfo::array[index]->type;
	}
	
	return 0;
}

const OPAtomRef&
OPTypeInfo::ReturnNthName (Count index)
{
	if (index < OPaC_TypeInfo::num) {
		return OPaC_TypeInfo::array[index]->name;
	}
	
	return OPAtomNull;
}

Bool
OPTypeInfo::FindType (const LatinChar* name, OPTypeDefOut type)
{
	char buffer[100];
	int len = strlen (name);
	
	nstrcpy (buffer, name, sizeof (buffer));
	
	if (len >= 3) {
		if (strcmp (name+len-3, "Ref") == 0) {
			buffer[len-3] = 0;
		}
	}
	
	for (Count index = 0; index < OPaC_TypeInfo::num; index++) {
		if (strcmp (OPaC_TypeInfo::array[index]->type->type, buffer) == 0) {
			type = OPaC_TypeInfo::array[index]->type;
			return TRUE;
		}
	}
	
	return FALSE;
}

Bool
OPTypeInfo::FindType (OPAtomIn name, OPTypeDefOut type)
{
	if (name.IsValid ()) {
		LatinChar buffer[100];
		name->ReturnString ()->GetStringCopy (buffer, sizeof (buffer));
		return OPTypeInfo::FindType (buffer, type);
	}

	return FALSE;
}

/****************************************************************/
