/*
 *	xtsmartsrc.cxx
 *
 *	OPaC Extension: Title Widget helper.
 *
 *	(C) Copyright 1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/new.h>
#include <OPaC/string.h>
#include <OPaC/values.h>

#include <OPaC/xtsmartsrc.h>

OP_IMPLEMENTATION_FILE;

/****************************************************************/

OPxSmartSource::OPxSmartSource ()
{
	OP_METHOD (OPxSmartSource, 0);
	
	this->query_roots = OP_NEW (OPArray);
	this->query_names = OP_NEW (OPArray);
	this->values      = OP_NEW (OPArray);
	
	this->is_cache_ok = FALSE;
	this->is_cache_volatile = FALSE;
	
	this->max_num = 0;
	this->num = 0;
}



/*
 *	
 */

OPxSmartSource::~OPxSmartSource ()
{
	OP_METHOD (~OPxSmartSource, 0);
	
	this->query_roots->SetSize (0);
	this->query_names->SetSize (0);
	this->values->SetSize (0);
}



/*
 *
 */

void
OPxSmartSource::Cleanup ()
{
	OP_METHOD (Cleanup, 0);
	
	this->query_roots->SetSize (0);
	this->query_names->SetSize (0);
	this->values->SetSize (0);
}

/*
 *	Save 
 */

void
OPxSmartSource::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	static OPStorageSync info ("<OPxSmartSource query_roots query_names"
							   " is_cache_volatile max_num num>");
	
	storage << info;
	storage << OPStorageSync::begin;
	storage << this->query_roots << this->query_names;
	storage << this->is_cache_volatile << this->max_num << this->num;
	storage << OPStorageSync::end;
}



/*
 *	Restore 
 */

void
OPxSmartSource::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	storage >> OPStorageSync::type;
	storage >> OPStorageSync::begin;
	storage >> this->query_roots >> this->query_names;
	storage >> this->is_cache_volatile >> this->max_num >> this->num;
	storage >> OPStorageSync::end;
	
	this->is_cache_ok = FALSE;
}



/*
 *	Update the data stored in the cache.
 */

void
OPxSmartSource::UpdateInternal ()
{
	OP_METHOD (UpdateInternal, 0);
	
	if ( (! this->is_cache_ok)
	  || (this->is_cache_volatile) ) {
		
		this->values->SetSize (this->num);
		this->is_cache_ok = TRUE;
		
		Count index = 0;
		
		while (index < this->num) {
			
			OPValueRef& value = this->values[index];
			
			if (this->query_roots[index].IsValid ()) {
				
				if (this->query_names[index].IsValid ()) {
					
					//	Query value either by its direct name or by its path, depending on the
					//	nature of the query name.
					
					OPStringRef name;
					Count pos = 0;
					
					if (this->query_names[index]->IsKindOf (OP_RUNTIME (OPString))) {
						name = this->query_names[index];
					} else {
						name = this->query_names[index]->ReturnString ();
					}
					
					if (name->Find ('.', pos)) {
						OPObject::QueryData (this->query_roots[index], name, value);
					} else {
						OPAtomRef atom = name;
						this->query_roots[index]->QueryData (atom, value);
					}
					
				} else {
					
					//	This is not a real query... Return directly the object instead of
					//	looking for its properties.
					
					if (this->query_roots[index]->IsKindOf (OP_RUNTIME (OPValue))) {
						value = this->query_roots[index];
					} else {
						value.NewValue ();
						value->SetObject (this->query_roots[index]);
					}
				}
			}
			
			index++;
		}

		this->DoUpdate ();
	}
}


void
OPxSmartSource::DoUpdate ()
{
	OP_METHOD (DoUpdate, 0);
}


Bool
OPxSmartSource::InitValueByName (Count index, OPObjectIn obj, OPAtomIn name)
{
	OP_METHOD (InitValueByName, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectIn, obj);
	OP_ARG (3, OPAtomIn, name);
	
	if ( (this->max_num)
	  && (this->max_num <= index) ) {
		return FALSE;
	}
	
	if (index >= this->num) {
		this->num = index+1;
		this->query_roots->SetSize (this->num);
		this->query_names->SetSize (this->num);
	}
	
	this->is_cache_ok = FALSE;
	this->query_roots[index] = obj;
	this->query_names[index] = name;
	
	return TRUE;
}


Bool
OPxSmartSource::InitValueByPath (Count index, OPObjectIn root, OPStringIn path)
{
	OP_METHOD (InitValueByPath, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectIn, root);
	OP_ARG (3, OPStringIn, path);
	
	if ( (this->max_num)
	  && (this->max_num <= index) ) {
		return FALSE;
	}
	
	if (index >= this->num) {
		this->num = index+1;
		this->query_roots->SetSize (this->num);
		this->query_names->SetSize (this->num);
	}
	
	this->is_cache_ok = FALSE;
	this->query_roots[index] = root;
	this->query_names[index] = path;
	
	return TRUE;
}



/*
 *	Change the nth value in the cache. Update the "real" data
 *	source behind it (if there is one).
 */

Bool
OPxSmartSource::SetNthValue (Count index, OPValueIn value)
{
	OP_METHOD (SetNthValue, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPValueIn, value);
	
	if (index >= this->num) {
		return FALSE;
	}
	
	this->values->SetSize (this->num);
	this->values[index] = value;
	this->is_cache_ok = FALSE;
	
	if ( (this->query_roots[index].IsValid ())
	  && (this->query_names[index].IsValid ()) ) {
		
		//	Define value either by its direct name or by its path, depending on the
		//	nature of the query name.
		
		OPStringRef name;
		Count pos = 0;
		
		if (this->query_names[index]->IsKindOf (OP_RUNTIME (OPString))) {
			name = this->query_names[index];
		} else {
			name = this->query_names[index]->ReturnString ();
		}
		
		if (name->Find ('.', pos)) {
			OPObject::DefineData (this->query_roots[index], name, value);
		} else {
			OPAtomRef atom = name;
			this->query_roots[index]->DefineData (atom, value);
		}
		
	} else {
		this->query_roots[index] = value;
	}
	
	return TRUE;
}



/*
 *	Return the nth value in the cache. Maybe update it if
 *	needed.
 */

Bool
OPxSmartSource::GetNthValue (Count index, OPValueOut value) const
{
	OP_METHOD (GetNthValue, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPValueOut, value);
	
	if (index >= this->num) {
		return FALSE;
	}
	
	//	This is an infringement to the C++ ethics... but the user has
	//	not to know about the cache and it being updated by a read-only
	//	access.
	
	OPxSmartSource* that = const_cast<OPxSmartSource*> (this);
	that->UpdateInternal ();
	
	value = this->values[index];
	
	return TRUE;
}


/*
 *	Flush the internal representations (values) and compute
 *	new ones. This is required if the data has changed and
 *	the values have not been declared as volatile using the
 *	EnableCache method and mode=FALSE.
 */

void
OPxSmartSource::FlushCache ()
{
	OP_METHOD (FlushCache, 0);

	this->is_cache_ok = FALSE;
}



/*
 *	If the mode is FALSE, then the values are considered to
 *	be volatile and will be reevaluated whenever needed.
 */

void
OPxSmartSource::EnableCache (Bool mode)
{
	OP_METHOD (EnableCache, 1);
	OP_ARG (1, Bool, mode);
	
	this->is_cache_volatile = (mode) ? FALSE : TRUE;
}



void
OPxSmartSource::UpdateCache ()
{
	OP_METHOD (UpdateCache, 0);
	
	if (this->is_cache_ok) {
		this->FlushCache ();
	}
	
	this->UpdateInternal ();
}


