/*
 *	gfxsurface.cxx
 *
 *	Graphic library.
 *
 *	The GfxSurface provides an interface to a true color pixmap. The
 *	pixels are always represented as alpha-RGB (8-8-8-8). They are
 *	converted internally to a more compact format if 8-bit or 16-bit
 *	pixmaps are used.
 *
 *	The  8-bit pixmap is just a monochrome, gray-level pixmap.
 *	The 16-bit pixmap is organised as 5-6-5, which is an RGB standard.
 *	The 32-bit pixmap is organised as alpha-RGB.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "gfxtypes.h"
#include "gfxsurface.h"
#include "gfxhelpers.h"

/*****************************************************************************/

GfxSurface::GfxSurface (Card32 bpp)
	: lines (0), dots (0), dx (0), dy (0), pitch (0), alpha (0),
	  buf_dx (0), buf_dy (0),
	  base_allocated (FALSE), is_transparent (FALSE),
	  clip_x1 (0), clip_y1 (0), clip_x2 (0), clip_y2 (0),
	  depth (0), pixel_bytes (0), pixel_trans (0)
{
	this->base.mem_any        = 0;
	this->line_buffer.mem_any = 0;
	
	const Card32 next[33] =
	{
		0,
		1, 1, 1, 1, 1, 1, 1, 1,
		2, 2, 2, 2, 2, 2, 2, 2,
		4, 4, 4, 4, 4, 4, 4, 4,
		4, 4, 4, 4, 4, 4, 4, 4
	};
	
	const Card32 shift[5] = { 0, 0, 1, 0, 2 };
	
	GFX_ASSERT (bpp < 33);
	
	this->depth       = bpp;
	this->pixel_bytes = next[bpp];
	this->pixel_shift = shift[this->pixel_bytes];
}

GfxSurface::~GfxSurface ()
{
	if (this->lines || this->base_allocated || this->line_buffer.mem_any || this->dots) {
		GfxSurface::FreeSurfaceAndLines ();
	}
}

/*****************************************************************************/

void
GfxSurface::Initialise (Card32 dx, Card32 dy)
{
	this->FreeSurfaceAndLines ();
	
	this->dx    = this->buf_dx = dx;
	this->dy    = this->buf_dy = dy;
	this->pitch = q_gfx_align (dx * this->pixel_bytes);
	
	this->ResetClipRect ();
	this->AllocSurface ();
	this->AllocLinesForSurface ();
}

void
GfxSurface::Initialise (Card32 dx, Card32 dy, void* memory, Card32 pitch)
{
	this->FreeSurfaceAndLines ();
	
	this->dx    = this->buf_dx = dx;
	this->dy    = this->buf_dy = dy;
	this->pitch = pitch;
	
	this->base.mem_any = memory;
	
	this->ResetClipRect ();
	this->AllocLinesForSurface ();
}

void
GfxSurface::SetNewSize (Card32 dx, Card32 dy)
{
	if ( (dx <= this->buf_dx)
	  && (dy <= this->buf_dy) ) {
		this->dx = dx;
		this->dy = dy;
	} else {
		Card32 new_dx = (dx + 16) & -16;
		Card32 new_dy = (dy + 16) & -16;
		this->Initialise (new_dx, new_dy);
		this->dx = dx;
		this->dy = dy;
	}
}

void
GfxSurface::OptimiseSize ()
{
	if ( (this->dx < this->buf_dx/2)
	  || (this->dy < this->buf_dy/2) ) {
		
		Card32 dx = (this->dx + 16) & -16;
		Card32 dy = (this->dy + 16) & -16;
		
		if ( (dx < this->buf_dx)
		  && (dy < this->buf_dy) ) {
			this->Initialise (dx, dy);
		}
	}
}

/*****************************************************************************/

void
GfxSurface::SetClipRect (Int32 ox, Int32 oy, Card32 dx, Card32 dy)
{
	Card32 x = (ox < 0) ? 0 : (Card32)(ox);
	Card32 y = (oy < 0) ? 0 : (Card32)(oy);
	if (ox < 0) { Card32 sx = (Card32)(-ox); dx = (sx < dx) ? dx - sx : 0; }
	if (oy < 0) { Card32 sy = (Card32)(-oy); dy = (sy < dy) ? dy - sy : 0; }
	if (x >= this->dx) { dx = 0; x = 0; }
	if (y >= this->dy) { dy = 0; y = 0; }
	if (x+dx > this->dx) { dx = this->dx - x; }
	if (y+dy > this->dy) { dy = this->dy - y; }
	
	this->clip_x1 = x;
	this->clip_y1 = y;
	this->clip_x2 = x+dx;
	this->clip_y2 = y+dy;
}

void
GfxSurface::ResetClipRect ()
{
	this->clip_x1 = 0;
	this->clip_y1 = 0;
	this->clip_x2 = this->dx;
	this->clip_y2 = this->dy;
	
}

/*****************************************************************************/

void
GfxSurface::AllocSurface ()
{
	this->base.mem_any   = NEW char[this->dy * this->pitch];
	this->base_allocated = TRUE;
}

void
GfxSurface::AllocLinesForSurface ()
{
	this->lines = NEW MemoryPtr[this->dy];
	
	Card8*     base = this->base.mem_8;
	MemoryPtr* iter = this->lines;
	
	for (Card32 num = this->dy; num; num--) {
		iter[0].mem_8 = base;
		iter += 1;
		base += this->pitch;
	}
	
	this->line_buffer.mem_32 = NEW Card32[this->dx];
}

void
GfxSurface::FreeSurfaceAndLines ()
{
	if (this->lines) {
		DELETE (this->lines);
		this->lines = 0;
	}
	if (this->base_allocated) {
		DELETE (this->base.mem_any);
		this->base_allocated = FALSE;
	}
	if (this->line_buffer.mem_32) {
		DELETE (this->line_buffer.mem_32);
		this->line_buffer.mem_32 = 0;
	}
	if (this->dots) {
		DELETE (this->dots);
		this->dots = 0;
	}
	this->base.mem_any = 0;
}

/*****************************************************************************/

/*
 *	Fill the specified rectangle with the pixel value. Clip to the
 *	bounding box.
 */

void
GfxSurface::RectFill (Int32 x, Int32 y, Card32 dx, Card32 dy, Card32 rgb)
{
	Card32 pixel = rgb;
	
	switch (this->pixel_bytes) {
		case 1:	pixel = q_gfx_rgb_to_gray (rgb); break;
		case 2: pixel = q_gfx_rgb_to_565 (rgb);  break;
	}
	
	q_gfx_cliprect (x, y, dx, dy, this->clip_x1, this->clip_y1, this->clip_x2, this->clip_y2);
	q_gfx_replicatepixel (pixel, this->pixel_bytes);
	
	Card32 pixels[2] = { pixel, pixel };
	
	if (q_gfx_testrect (dx, dy)) {
		
		if (this->pixel_bytes == 4) {
			
			//	Optimised 32-bit version of the fill...
			
			Card32 copy16 = dx / 16;
			Card32 copy1  = dx & 15;
			
			while (dy--) {
				
				Card32* line;
				Card32  copy;
				
				line = this->lines[y++].mem_32 + x;
				copy = copy16;
				
				while (copy--) {
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
					*line++ = pixel;
				}
				
				copy = copy1;
				
				while (copy--) {
					*line++ = pixel;
				}
			}
			
			return;
		}
		
		Card32 offset_bytes =  x << this->pixel_shift;		//	offset in bytes
		Card32 dx_bytes     = dx << this->pixel_shift;		//	width in bytes
		
		while (dy--) {
			
			MemoryPtr    line = this->lines[y++];
			const Card8* ppix = (Card8*)(pixels);
			Card32 copy_bytes = dx_bytes;
			
			line.mem_8 += offset_bytes;
			
			//	Fill the line : first align to a 32-bit boundary, then fill in an aligned
			//	loop and finally fill the remaining bytes.
			
			while ( ((Card32)(line.mem_any) & 0x03)
				 && (copy_bytes) )	{ *line.mem_8++  = *ppix++; copy_bytes -= 1; }
			while (copy_bytes >= 4)	{ *line.mem_32++ = pixel;   copy_bytes -= 4; }
			while (copy_bytes)		{ *line.mem_8++  = *ppix++; copy_bytes -= 1; }
		}
	}
}


/*
 *	Blend the specified rectangle with the pixel value. Clip to the
 *	bounding box.
 */

void
GfxSurface::RectFillBlend (Int32 x, Int32 y, Card32 dx, Card32 dy, Card32 rgba)
{
	Card32 alpha = rgba >> 24;
	
	if (alpha == 0) {
		this->RectFill (x, y, dx, dy, rgba);
		return;
	}
	
	q_gfx_cliprect (x, y, dx, dy, this->clip_x1, this->clip_y1, this->clip_x2, this->clip_y2);
	
	//	Poor implementation... Will have to optimise this piece of code !
	
	if (q_gfx_testrect (dx, dy)) {
		
		Card32 x2 = x+dx;
		Card32 y2 = y+dy;
		
		if (this->pixel_bytes == 4) {
			for (Card32 y1 = y; y1 < y2; y1++) {
				
				Card32* ptr = this->lines[y1].mem_32 + x;
				Card32 opix = *ptr;
				Card32 npix = q_gfx_blend_rgb_to_rgb (opix, rgba);
				
				*ptr++ = npix;
				
				//	Optimise the case where the current pixel is the same as the previous
				//	one : no new merging is needed, just use the previously computed value.
				
				for (Card32 x1 = x+1; x1 < x2; x1++) {
					
					Card32 cpix = *ptr;
					
					if (cpix != opix) {
						opix = cpix;
						npix = q_gfx_blend_rgb_to_rgb (opix, rgba);
					}
					
					*ptr++ = npix;
				}
			}
			
		} else {
			for (Card32 y1 = y; y1 < y2; y1++) {
				for (Card32 x1 = x; x1 < x2; x1++) {
					this->SetPixelRGBBlend (x1, y1, rgba);
				}
			}
		}
	}
}

/*
 *	Blend with the specified pixel color; use the surface as a source
 *	for alpha-values (either 8-bit surface or 32-bit surface). The alpha
 *	value is combined with the pixel transparency.
 */

void
GfxSurface::RectFillBlend (Int32 x, Int32 y, Card32 dx, Card32 dy, Card32 pixel,
						   Int32 sx, Int32 sy, GfxSurface* surface)
{
	Int32 ox = x;
	Int32 oy = y;
	
	q_gfx_cliprect (x, y, dx, dy, this->clip_x1, this->clip_y1, this->clip_x2, this->clip_y2);
	
	sx += x - ox;
	sy += y - oy;
	
	if (surface == 0) {
		this->RectFill (x, y, dx, dy, pixel);
		return;
	}
	
	q_gfx_cliprect (sx, sy, dx, dy, surface->dx, surface->dy);
	
	//	If the pixel color alpha value is set to 0 (opaque), use 100% of
	//	the transparency information from the input surface.
	
	Card32 compose = pixel >> 24;
	if (compose > 127) compose++;
	compose = 256-compose;
	pixel  &= 0x00FFFFFF;
	
	while (sx < 0) sx += surface->dx;
	while (sy < 0) sy += surface->dy;
	
	if (q_gfx_testrect (dx, dy)) {
		
		while (sx >= (Int32)surface->dx) sx -= surface->dx;
		
		if (surface->depth == 8) {
			
			while (dy--) {
				
				while (sy >= (Int32)surface->dy) sy -= surface->dy;
				
				MemoryPtr dst_line = this->lines[y++];
				MemoryPtr tmp_line = this->line_buffer;
				MemoryPtr src_line = surface->lines[sy++];
				
				Card32 num = dx;
				Card32 ox  = sx;
				
				while (num--) {
					while (ox >= surface->dx) ox -= surface->dx;
					Card32 alpha = (src_line.mem_8[ox++] * compose) >> 8;
					*tmp_line.mem_32++ = (alpha << 24) | (pixel);
				}
				
				this->BlendWithRGB (this->line_buffer, dst_line, x, dx);
			}
			
		} else if (surface->depth == 32) {
			
			while (dy--) {
				
				while (sy >= (Int32)surface->dy) sy -= surface->dy;
				
				MemoryPtr dst_line = this->lines[y++];
				MemoryPtr tmp_line = this->line_buffer;
				MemoryPtr src_line = surface->lines[sy++];
				
				Card32 num = dx;
				Card32 ox  = sx;
				
				while (num--) {
					while (ox >= surface->dx) ox -= surface->dx;
					Card32 alpha = ((src_line.mem_32[ox++] & 0xFF000000) >> 8) * compose;
					*tmp_line.mem_32++ = alpha | pixel;
				}
				
				this->BlendWithRGB (this->line_buffer, dst_line, x, dx);
			}
		}
	}
}

/*****************************************************************************/

void
GfxSurface::RectCopy (Int32 x, Int32 y, Card32 dx, Card32 dy,
					  Int32 sx, Int32 sy, GfxSurface* surface,
					  Bool tile)
{
	if (surface == 0) surface = this;
	
	Int32 ox = x;
	Int32 oy = y;
	
	q_gfx_cliprect (x, y, dx, dy, this->clip_x1, this->clip_y1, this->clip_x2, this->clip_y2);
	
	sx += x - ox;
	sy += y - oy;
	
	if (tile) {
		
		//	Don't tile if the surface does not wrap around in the source.
		
		tile = (sx+dx > surface->dx) | (sy+dy > surface->dy);
		
		while (sx < 0) sx += surface->dx;
		while (sy < 0) sy += surface->dy;
		
	} else {
		
		q_gfx_cliprect (sx, sy, dx, dy, surface->dx, surface->dy);
	}
	
	if (q_gfx_testrect (dx, dy)) {
		
		if (tile) while (sx >= (Int32)surface->dx) sx -= surface->dx;
		
		if ((!tile) && (surface == this) && (sy < y) && (sy+(Int32)dy > y)) {
			
			//	Source and destination memory overlap in such a way that the copy must
			//	be done from the bottom to the top (source is above the destination).
			
			y  += dy;
			sy += dy;
			
			while (dy--) {
				
				MemoryPtr dst_line = this->lines[--y];
				MemoryPtr src_line = surface->lines[--sy];
				
				surface->ConvertToRGB (src_line, this->line_buffer, sx, dx);
				
				if (this->is_transparent) {
					this->ConvertFromRGB (this->line_buffer, dst_line, x, dx, this->pixel_trans);
				} else {
					this->ConvertFromRGB (this->line_buffer, dst_line, x, dx);
				}
			}
			
		} else if ((!tile) && (surface->depth == this->depth) && (this->is_transparent == FALSE)) {
			
			//	Specially optimised case : both source and destination have exactly
			//	the same depth. Simply copy the lines without using the intermediate
			//	buffer.
			
			switch (this->pixel_bytes) {
					
				case 1:
					while (dy--) {
						
						Card8* dst_line = this->lines[y++].mem_8 + x;
						Card8* src_line = surface->lines[sy++].mem_8 + sx;
						Card32 num = dx;
						
						while (num--) {
							*dst_line++ = *src_line++;
						}
					}
					break;
			
				case 2:
					while (dy--) {
						
						Card16* dst_line = this->lines[y++].mem_16 + x;
						Card16* src_line = surface->lines[sy++].mem_16 + sx;
						Card32 num = dx;
						
						while (num--) {
							*dst_line++ = *src_line++;
						}
					}
					break;
			
				case 4:
					while (dy--) {
						
						Card32* dst_line = this->lines[y++].mem_32 + x;
						Card32* src_line = surface->lines[sy++].mem_32 + sx;
						Card32 num = dx;
						
						while (num--) {
							*dst_line++ = *src_line++;
						}
					}
					break;
			}
			
		} else {
			
			//	Copy without memory overlap or with a destination below of the source.
			//	This might be a bit faster if the memory prefetches data.
			
			if (tile) {
				
				while (dy--) {
					
					while (sy >= (Int32)surface->dy) sy -= surface->dy;
					
					MemoryPtr dst_line = this->lines[y++];
					MemoryPtr src_line = surface->lines[sy++];
					
					surface->ConvertTileToRGB (src_line, this->line_buffer, sx, dx);
					
					if (this->is_transparent) {
						this->ConvertFromRGB (this->line_buffer, dst_line, x, dx, this->pixel_trans);
					} else {
						this->ConvertFromRGB (this->line_buffer, dst_line, x, dx);
					}
				}
				
			} else {
				
				while (dy--) {
					
					MemoryPtr dst_line = this->lines[y++];
					MemoryPtr src_line = surface->lines[sy++];
					
					surface->ConvertToRGB (src_line, this->line_buffer, sx, dx);
					
					if (this->is_transparent) {
						this->ConvertFromRGB (this->line_buffer, dst_line, x, dx, this->pixel_trans);
					} else {
						this->ConvertFromRGB (this->line_buffer, dst_line, x, dx);
					}
				}
			}
		}
	}
}

void
GfxSurface::RectBlend (Int32 x, Int32 y, Card32 dx, Card32 dy,
					   Int32 sx, Int32 sy, GfxSurface* surface,
					   Bool tile)
{
	if (surface == 0) surface = this;
	
	Int32 ox = x;
	Int32 oy = y;
	
	q_gfx_cliprect (x, y, dx, dy, this->clip_x1, this->clip_y1, this->clip_x2, this->clip_y2);
	
	sx += x - ox;
	sy += y - oy;
	
	if (tile) {
		
		//	Don't tile if the surface does not wrap around in the source.
		
		tile = (sx+dx > surface->dx) | (sy+dy > surface->dy);
		
		while (sx < 0) sx += surface->dx;
		while (sy < 0) sy += surface->dy;
		
	} else {
		
		q_gfx_cliprect (sx, sy, dx, dy, surface->dx, surface->dy);
	}
	
	if (q_gfx_testrect (dx, dy)) {
		
		if (tile) while (sx >= (Int32)surface->dx) sx -= surface->dx;
		
		if ((!tile) && (surface == this) && (sy < y) && (sy+(Int32)dy > y)) {
			
			//	Source and destination memory overlap in such a way that the copy must
			//	be done from the bottom to the top (source is above the destination).
			
			y  += dy;
			sy += dy;
			
			while (dy--) {
				
				MemoryPtr dst_line = this->lines[--y];
				MemoryPtr src_line = surface->lines[--sy];
				
				surface->ConvertToRGB (src_line, this->line_buffer, sx, dx);
				this->BlendWithRGB (this->line_buffer, dst_line, x, dx);
			}
			
		} else {
			
			//	Copy without memory overlap or with a destination below of the source.
			//	This might be a bit faster if the memory prefetches data.
			
			if (tile) {
				
				while (dy--) {
					
					if (tile) while (sy >= (Int32)surface->dy) sy -= surface->dy;
					
					MemoryPtr dst_line = this->lines[y++];
					MemoryPtr src_line = surface->lines[sy++];
					
					surface->ConvertTileToRGB (src_line, this->line_buffer, sx, dx);
					this->BlendWithRGB (this->line_buffer, dst_line, x, dx);
				}
				
			} else {
				
				while (dy--) {
					
					MemoryPtr dst_line = this->lines[y++];
					MemoryPtr src_line = surface->lines[sy++];
					
					surface->ConvertToRGB (src_line, this->line_buffer, sx, dx);
					this->BlendWithRGB (this->line_buffer, dst_line, x, dx);
				}
			}
		}
	}
}

/*****************************************************************************/

const Card32 g_map_5_to_8[256] =
{
	0x00, 0x08, 0x10, 0x18,  0x21, 0x29, 0x31, 0x39,
	0x42, 0x4A, 0x52, 0x5A,  0x63, 0x6B, 0x63, 0x6B,
	0x84, 0x8C, 0x94, 0x9C,  0xA5, 0xAD, 0xB5, 0xBD,
	0xC6, 0xCE, 0xD6, 0xDE,  0xE7, 0xEF, 0xF7, 0xFF,
	
	0x00, 0x08, 0x10, 0x18,  0x21, 0x29, 0x31, 0x39,
	0x42, 0x4A, 0x52, 0x5A,  0x63, 0x6B, 0x63, 0x6B,
	0x84, 0x8C, 0x94, 0x9C,  0xA5, 0xAD, 0xB5, 0xBD,
	0xC6, 0xCE, 0xD6, 0xDE,  0xE7, 0xEF, 0xF7, 0xFF,
	
	0x00, 0x08, 0x10, 0x18,  0x21, 0x29, 0x31, 0x39,
	0x42, 0x4A, 0x52, 0x5A,  0x63, 0x6B, 0x63, 0x6B,
	0x84, 0x8C, 0x94, 0x9C,  0xA5, 0xAD, 0xB5, 0xBD,
	0xC6, 0xCE, 0xD6, 0xDE,  0xE7, 0xEF, 0xF7, 0xFF,
	
	0x00, 0x08, 0x10, 0x18,  0x21, 0x29, 0x31, 0x39,
	0x42, 0x4A, 0x52, 0x5A,  0x63, 0x6B, 0x63, 0x6B,
	0x84, 0x8C, 0x94, 0x9C,  0xA5, 0xAD, 0xB5, 0xBD,
	0xC6, 0xCE, 0xD6, 0xDE,  0xE7, 0xEF, 0xF7, 0xFF,
	
	0x00, 0x08, 0x10, 0x18,  0x21, 0x29, 0x31, 0x39,
	0x42, 0x4A, 0x52, 0x5A,  0x63, 0x6B, 0x63, 0x6B,
	0x84, 0x8C, 0x94, 0x9C,  0xA5, 0xAD, 0xB5, 0xBD,
	0xC6, 0xCE, 0xD6, 0xDE,  0xE7, 0xEF, 0xF7, 0xFF,
	
	0x00, 0x08, 0x10, 0x18,  0x21, 0x29, 0x31, 0x39,
	0x42, 0x4A, 0x52, 0x5A,  0x63, 0x6B, 0x63, 0x6B,
	0x84, 0x8C, 0x94, 0x9C,  0xA5, 0xAD, 0xB5, 0xBD,
	0xC6, 0xCE, 0xD6, 0xDE,  0xE7, 0xEF, 0xF7, 0xFF,
	
	0x00, 0x08, 0x10, 0x18,  0x21, 0x29, 0x31, 0x39,
	0x42, 0x4A, 0x52, 0x5A,  0x63, 0x6B, 0x63, 0x6B,
	0x84, 0x8C, 0x94, 0x9C,  0xA5, 0xAD, 0xB5, 0xBD,
	0xC6, 0xCE, 0xD6, 0xDE,  0xE7, 0xEF, 0xF7, 0xFF,
	
	0x00, 0x08, 0x10, 0x18,  0x21, 0x29, 0x31, 0x39,
	0x42, 0x4A, 0x52, 0x5A,  0x63, 0x6B, 0x63, 0x6B,
	0x84, 0x8C, 0x94, 0x9C,  0xA5, 0xAD, 0xB5, 0xBD,
	0xC6, 0xCE, 0xD6, 0xDE,  0xE7, 0xEF, 0xF7, 0xFF
};

const Card32 g_map_6_to_8[256] =
{
	0x00, 0x04, 0x08, 0x0C,  0x10, 0x14, 0x18, 0x1C,
	0x20, 0x24, 0x28, 0x2C,  0x30, 0x34, 0x38, 0x3C,
	0x41, 0x45, 0x49, 0x4D,  0x51, 0x55, 0x59, 0x5D,
	0x61, 0x65, 0x69, 0x6D,  0x71, 0x75, 0x79, 0x7D,
	0x82, 0x86, 0x8A, 0x8E,  0x92, 0x96, 0x9A, 0x9E,
	0xA2, 0xA6, 0xAA, 0xAE,  0xB2, 0xB6, 0xBA, 0xBE,
	0xC3, 0xC7, 0xCB, 0xCF,  0xD3, 0xD7, 0xDB, 0xDF,
	0xE3, 0xE7, 0xEB, 0xEF,  0xF3, 0xF7, 0xFB, 0xFF,
	
	0x00, 0x04, 0x08, 0x0C,  0x10, 0x14, 0x18, 0x1C,
	0x20, 0x24, 0x28, 0x2C,  0x30, 0x34, 0x38, 0x3C,
	0x41, 0x45, 0x49, 0x4D,  0x51, 0x55, 0x59, 0x5D,
	0x61, 0x65, 0x69, 0x6D,  0x71, 0x75, 0x79, 0x7D,
	0x82, 0x86, 0x8A, 0x8E,  0x92, 0x96, 0x9A, 0x9E,
	0xA2, 0xA6, 0xAA, 0xAE,  0xB2, 0xB6, 0xBA, 0xBE,
	0xC3, 0xC7, 0xCB, 0xCF,  0xD3, 0xD7, 0xDB, 0xDF,
	0xE3, 0xE7, 0xEB, 0xEF,  0xF3, 0xF7, 0xFB, 0xFF,
	
	0x00, 0x04, 0x08, 0x0C,  0x10, 0x14, 0x18, 0x1C,
	0x20, 0x24, 0x28, 0x2C,  0x30, 0x34, 0x38, 0x3C,
	0x41, 0x45, 0x49, 0x4D,  0x51, 0x55, 0x59, 0x5D,
	0x61, 0x65, 0x69, 0x6D,  0x71, 0x75, 0x79, 0x7D,
	0x82, 0x86, 0x8A, 0x8E,  0x92, 0x96, 0x9A, 0x9E,
	0xA2, 0xA6, 0xAA, 0xAE,  0xB2, 0xB6, 0xBA, 0xBE,
	0xC3, 0xC7, 0xCB, 0xCF,  0xD3, 0xD7, 0xDB, 0xDF,
	0xE3, 0xE7, 0xEB, 0xEF,  0xF3, 0xF7, 0xFB, 0xFF,
	
	0x00, 0x04, 0x08, 0x0C,  0x10, 0x14, 0x18, 0x1C,
	0x20, 0x24, 0x28, 0x2C,  0x30, 0x34, 0x38, 0x3C,
	0x41, 0x45, 0x49, 0x4D,  0x51, 0x55, 0x59, 0x5D,
	0x61, 0x65, 0x69, 0x6D,  0x71, 0x75, 0x79, 0x7D,
	0x82, 0x86, 0x8A, 0x8E,  0x92, 0x96, 0x9A, 0x9E,
	0xA2, 0xA6, 0xAA, 0xAE,  0xB2, 0xB6, 0xBA, 0xBE,
	0xC3, 0xC7, 0xCB, 0xCF,  0xD3, 0xD7, 0xDB, 0xDF,
	0xE3, 0xE7, 0xEB, 0xEF,  0xF3, 0xF7, 0xFB, 0xFF
};

/*****************************************************************************/

/*
 *	Convert a line (the specified width can be larger than the real
 *	line width, it will wrap to the beginning of the line) of the
 *	current format (src belongs to our surface) to alpha-RGB.
 */

void
GfxSurface::ConvertToRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx)
{
	switch (this->pixel_bytes) {
		
		case 1:
			
			//	Convert monochrome source to alpha-RGB.
			
			while (dx--) {
				Card32 pixel  = src.mem_8[x++];
				*dst.mem_32++ = q_gfx_gray_to_rgb (pixel, this->alpha);
			}
			
			break;
		
		case 2:
			
			//	Convert 16-bit/pixel (5-6-5 representation) to alpha-RGB.
			
			while (dx--) {
				Card32 pixel  = src.mem_16[x++];
				*dst.mem_32++ = q_gfx_565_to_rgb (pixel, this->alpha);
			}
			
			break;

		case 4:
			
			//	Convert 32-bit/pixel to alpha-RGB (no conversion needed)
			
			while (dx--) {
				Card32 pixel  = src.mem_32[x++];
				*dst.mem_32++ = pixel;
			}
			
			break;
	}
}

void
GfxSurface::ConvertTileToRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx)
{
	switch (this->pixel_bytes) {
		
		case 1:
			
			//	Convert monochrome source to alpha-RGB.
			
			while (dx--) {
				while (x >= this->dx) x -= this->dx;
				Card32 pixel  = src.mem_8[x++];
				*dst.mem_32++ = q_gfx_gray_to_rgb (pixel, this->alpha);
			}
			
			break;
		
		case 2:
			
			//	Convert 16-bit/pixel (5-6-5 representation) to alpha-RGB.
			
			while (dx--) {
				while (x >= this->dx) x -= this->dx;
				Card32 pixel  = src.mem_16[x++];
				*dst.mem_32++ = q_gfx_565_to_rgb (pixel, this->alpha);
			}
			
			break;

		case 4:
			
			//	Convert 32-bit/pixel to alpha-RGB (no conversion needed)
			
			while (dx--) {
				while (x >= this->dx) x -= this->dx;
				Card32 pixel  = src.mem_32[x++];
				*dst.mem_32++ = pixel;
			}
			
			break;
	}
}

/*
 *	Convert the source alpha-RGB line into a line stored in our own
 *	representation. Beware that no checking is done on x and dx, while
 *	ConvertToRGB wraps when dx exceeds the real line width !
 */

void
GfxSurface::ConvertFromRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx)
{
	switch (this->pixel_bytes) {
		
		case 1:
			
			//	Convert RGB to monochrome (256 gray levels).
			
			dst.mem_8 += x;
			
			while (dx--) {
				Card32 pixel = *src.mem_32++;
				*dst.mem_8++ = (Card8) q_gfx_rgb_to_gray (pixel);
			}
			
			break;
		
		case 2:
			
			//	Convert RGB to 16-bit/pixel (5-6-5 representation).
			
			dst.mem_16 += x;
			
			while (dx--) {
				Card32 pixel = *src.mem_32++;
				*dst.mem_16++ = (Card16) q_gfx_rgb_to_565 (pixel);
			}
			
			break;

		case 4:
			
			//	Convert RGB to 32-bit/pixel (no conversion needed)
			
			dst.mem_32 += x;
			
			while (dx--) {
				*dst.mem_32++ = *src.mem_32++;
			}
			
			break;
	}
}

/*
 *	Same as the standard ConvertFromRGB method above, but the pixels
 *	matching the transparent color won't be written.
 */

void
GfxSurface::ConvertFromRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx, Card32 trans)
{
	switch (this->pixel_bytes) {
		
		case 1:
			
			//	Convert RGB to monochrome (256 gray levels).
			
			dst.mem_8 += x;
			
			while (dx--) {
				Card32 pixel = *src.mem_32++;
				if (pixel == trans) {
					dst.mem_8++;
				} else {
					*dst.mem_8++ = (Card8) q_gfx_rgb_to_gray (pixel);
				}
			}
			
			break;
		
		case 2:
			
			//	Convert RGB to 16-bit/pixel (5-6-5 representation).
			
			dst.mem_16 += x;
			
			while (dx--) {
				Card32 pixel = *src.mem_32++;
				if (pixel == trans) {
					dst.mem_16++;
				} else {
					*dst.mem_16++ = (Card16) q_gfx_rgb_to_565 (pixel);
				}
			}
			
			break;

		case 4:
			
			//	Convert RGB to 32-bit/pixel (no conversion needed)
			
			dst.mem_32 += x;
			
			while (dx--) {
				Card32 pixel = *src.mem_32++;
				if (pixel == trans) {
					dst.mem_32++;
				} else {
					*dst.mem_32++ = pixel;
				}
			}
			
			break;
	}
}

void
GfxSurface::BlendWithRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx)
{
	switch (this->pixel_bytes) {
		
		case 1:
			
			//	Blend RGB with monochrome (256 gray levels).
			
			dst.mem_8 += x;
			
			while (dx--) {
				*dst.mem_8 = q_gfx_blend_rgb_to_gray (*dst.mem_8, *src.mem_32++);
				dst.mem_8++;
			}
			break;
		
		case 2:
			
			//	Blend RGB with 16-bit/pixel (5-6-5 representation).
			
			dst.mem_16 += x;
			
			while (dx--) {
				*dst.mem_16 = q_gfx_blend_rgb_to_565 (*dst.mem_16, *src.mem_32++);
				dst.mem_16++;
			}
			
			break;

		case 4:
			
			//	Blend RGB with 32-bit/pixel.
			
			dst.mem_32 += x;
			
			while (dx--) {
				*dst.mem_32 = q_gfx_blend_rgb_to_rgb (*dst.mem_32, *src.mem_32++);
				dst.mem_32++;
			}
			
			break;
	}
}

/*****************************************************************************/

/*
 *	Extract 8-bit pixels quickly from the source (the 32-bit offset is
 *	specified) which is stored as alpha-RGB. Store the 8-bit result
 *	at the specified x offset. Start reading from the beginning of the
 *	source line.
 */

void
GfxSurface::ExtractFromRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx, Card32 shift)
{
	dst.mem_8 += x;
	
	while (dx--) {
		Card32 pixel = *src.mem_32++;
		*dst.mem_8++ = (Card8) (pixel >> shift);
	}
}

/*
 *	Insert 8-bit pixels quickly into the destination, which is stored
 *	as an alpha-RGB, starting from the beginning of the destination
 *	buffer !
 */

void
GfxSurface::InsertIntoRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx, Card32 shift, Card32 mask)
{
	src.mem_8 += x;
	
	while (dx--) {
		Card8  s_pix = *src.mem_8++;
		Card32 d_pix = *dst.mem_32;
		d_pix &= mask;
		d_pix |= s_pix << shift;
		*dst.mem_32++ = d_pix;
	}
}



/*
 *	Extract monochrome pixels quickly from the source. Store the result
 *	at the specified x offset. Start reading from the beginning of the
 *	source line.
 */

void
GfxSurface::ExtractMonoFromRGB (MemoryPtr src, Card8* mono, Card32 x, Card32 dx)
{
	Card8 bit = 0x80 >> (x & 7);
	int index = x / 8;
	
	while (dx--) {
		
		Card32 pixel = *src.mem_32++;
		Card8  alpha = (Card8)(pixel >> 24);
		
		if (alpha < 128) {
			Card32 r     = (Card8)(pixel >> 16);
			Card32 g     = (Card8)(pixel >>  8);
			Card32 b     = (Card8)(pixel >>  0);
			if (((r*307+g*614+b*103) >> 10) > 128) {
				mono[index] |= bit;
			} else {
				mono[index] &= ~bit;
			}
		}
		
		bit = bit >> 1;
		
		if (bit == 0) {
			bit = 0x80;
			index++;
		}
	}
}


/*
 *	Insert monochrome pixels quickly into the destination, which is
 *	stored as an alpha-RGB, starting from the beginning of the destination
 *	buffer !
 */

void
GfxSurface::InsertMonoIntoRGB (const Card8* mono, MemoryPtr dst, Card32 x, Card32 dx)
{
	Card8 bit = 0x80 >> (x & 7);
	int index = x / 8;
	
	while (dx--) {
		
		*dst.mem_32++ = (mono[index] & bit) ? 0x00000000 : 0x00FFFFFF;
		
		bit = bit >> 1;
		
		if (bit == 0) {
			bit = 0x80;
			index++;
		}
	}
}

/*****************************************************************************/

/*
 *	Set a pixel value (the alpha channel of the pixel is used to blend
 *	the color with the destination).
 */

void
GfxSurface::SetPixelRGBBlend (Int32 x, Int32 y, Card32 pixel)
{
	if (x < (Int32)this->clip_x1) return;
	if (y < (Int32)this->clip_y1) return;
	if (x >= (Int32)this->clip_x2) return;
	if (y >= (Int32)this->clip_y2) return;
	
	Card32 alpha = pixel >> 24;
	
	if (alpha) {
		
		//	Don't modify the destination pixel if the alpha channel specifies a transparent
		//	source pixel color.
		
		if (alpha == 255) return;
		
		//	Handle the color merging.
		
		switch (this->pixel_bytes) {
			
			case 1:
				this->lines[y].mem_8[x] = q_gfx_blend_rgb_to_gray (this->lines[y].mem_8[x], pixel);
				break;
			
			case 2:
				this->lines[y].mem_16[x] = q_gfx_blend_rgb_to_565 (this->lines[y].mem_16[x], pixel);
				break;
			case 4:
				this->lines[y].mem_32[x] = q_gfx_blend_rgb_to_rgb (this->lines[y].mem_32[x], pixel);
				break;
		}
		
	} else {
		
		//	The source pixel is opaque : just overwrite the destination.
		
		switch (this->pixel_bytes) {
			case 1:	this->lines[y].mem_8[x]  = (Card8)  q_gfx_rgb_to_gray (pixel); break;
			case 2: this->lines[y].mem_16[x] = (Card16) q_gfx_rgb_to_565 (pixel);  break;
			case 4: this->lines[y].mem_32[x] = pixel; break;
		}
	}
}

/*
 *	Put a pixel value, without interpreting the alpha channel. The
 *	destination gets overwritten.
 */

void
GfxSurface::SetPixelRGB (Int32 x, Int32 y, Card32 pixel)
{
	if (x < (Int32)this->clip_x1) return;
	if (y < (Int32)this->clip_y1) return;
	if (x >= (Int32)this->clip_x2) return;
	if (y >= (Int32)this->clip_y2) return;
	
	//	Overwrite the destination; don't take into account the source alpha channel
	//	for any blending.
	
	switch (this->pixel_bytes) {
		case 1:	this->lines[y].mem_8[x]  = (Card8)  q_gfx_rgb_to_gray (pixel); break;
		case 2: this->lines[y].mem_16[x] = (Card16) q_gfx_rgb_to_565 (pixel);  break;
		case 4: this->lines[y].mem_32[x] = pixel; break;
	}
}


void
GfxSurface::SetPixelLineBGR24 (Int32 x, Int32 y, Card32 dx, const Card8* rgb)
{
	if (x < (Int32)this->clip_x1) return;
	if (y < (Int32)this->clip_y1) return;
	if (x >= (Int32)this->clip_x2) return;
	if (y >= (Int32)this->clip_y2) return;
	
	if (x+(Int32)dx > (Int32)this->clip_x2) dx = (Card32)(this->clip_x2 - x);
	
	switch (this->pixel_bytes) {
		case 1:	/* NIY */ break;
		case 2: /* NIY */ break;
		case 4:
			{
				Card32* dst = & this->lines[y].mem_32[x];
				while (dx--) {
//					*dst++ = (rgb[2] << 16) | (rgb[1] << 8) | (rgb[0]);
					*dst++ = ((const Card32*)(rgb))[0] & 0x00FFFFFF;
					rgb += 3;
				}
			}
			break;
	}
}

void
GfxSurface::SetPixelLineMono (Int32 x, Int32 y, Card32 dx, Card32 pixel, const Card8* bits)
{
	if (x < (Int32)this->clip_x1) return;
	if (y < (Int32)this->clip_y1) return;
	if (x >= (Int32)this->clip_x2) return;
	if (y >= (Int32)this->clip_y2) return;
	
	if (x+(Int32)dx > (Int32)this->clip_x2) dx = (Card32)(this->clip_x2 - x);
	
	int mask = 0x80;
	int num  = dx / 8;
	int xor  = 0xFF;
	int byte = xor ^ *bits++;
	
	dx &= 7;
	
	switch (this->pixel_bytes) {
		case 1:	/* NIY */ break;
		case 2: /* NIY */ break;
		case 4:
			{
				Card32* dst = & this->lines[y].mem_32[x];
				
				while (num--) {
					if (byte) {
						while (mask) {
							if (byte & mask) {
								*dst++ = pixel;
							} else {
								dst++;
							}
							mask = mask >> 1;
						}
						mask = 0x80;
					} else {
						dst += 8;
					}
					
					byte = xor ^ *bits++;
				}
				
				while (dx--) {
					if (byte & mask) {
						*dst++ = pixel;
					} else {
						dst++;
					}
					mask = mask >> 1;
				}
			}
			break;
	}
}


/*
 *	Return the value of the specified pixel.
 */

Card32
GfxSurface::GetPixelRGB (Int32 x, Int32 y)
{
	if (x < (Int32)this->clip_x1) return 0;
	if (y < (Int32)this->clip_y1) return 0;
	if (x >= (Int32)this->clip_x2) return 0;
	if (y >= (Int32)this->clip_y2) return 0;
	
	switch (this->pixel_bytes) {
		case 1:	return q_gfx_gray_to_rgb (this->lines[y].mem_8[x], this->alpha); break;
		case 2:	return q_gfx_565_to_rgb (this->lines[y].mem_16[x], this->alpha); break;
		case 4:	return this->lines[y].mem_32[x]; break;
	}
	
	return 0;
}

/*****************************************************************************/

/*
 *	Extract a single color plane (or alpha plane) from the current
 *	surface. The destination surface must be 8-bit deep. No tiling
 *	is done (while RectCopy/RectBlend do tiling).
 */

void
GfxSurface::RectExtract (Int32 x, Int32 y, Card32 dx, Card32 dy,
						 Int32 px, Int32 py, GfxSurface* surface, ColorPlane plane)
{
	if (surface == 0) return;
	if (surface->depth != 8) return;
	
	Int32 ox = x;
	Int32 oy = y;
	
	q_gfx_cliprect (x, y, dx, dy, this->clip_x1, this->clip_y1, this->clip_x2, this->clip_y2);
	
	px += x - ox;
	py += y - oy;
	
	q_gfx_cliprect (px, py, dx, dy, surface->dx, surface->dy);
	
	if (q_gfx_testrect (dx, dy)) {
		
		Card32 shift = 0;
		
		switch (plane) {
			case PLANE_ALPHA: shift = 24; break;
			case PLANE_RED:   shift = 16; break;
			case PLANE_GREEN: shift =  8; break;
			case PLANE_BLUE:  shift =  0; break;
		}
		
		//	Extract the specified plane (either alpha, red, green or blue) from the
		//	current surface into the specified, destination surface.
		
		while (dy--) {
			
			MemoryPtr src_line = this->lines[y++];
			MemoryPtr dst_line = surface->lines[py++];
			
			this->ConvertToRGB (src_line, this->line_buffer, x, dx);
			surface->ExtractFromRGB (this->line_buffer, dst_line, px, dx, shift);
		}
	}
}



/*
 *	Extract a monochrome bitmap from the current surface. The bitmap
 *	is organised with the leftmost pixel the most significant bit of
 *	the first byte. The pitch specifies the distance between two
 *	consecutive lines.
 */

void
GfxSurface::RectExtractMono (Int32 x, Int32 y, Card32 dx, Card32 dy,
							 Int32 px, Int32 py, Card8* mono, Card32 pitch)
{
	if (mono == 0) return;
	if (pitch == 0) return;
	
	Int32 ox = x;
	Int32 oy = y;
	
	q_gfx_cliprect (x, y, dx, dy, this->clip_x1, this->clip_y1, this->clip_x2, this->clip_y2);
	
	px += x - ox;
	py += y - oy;
	
	if (q_gfx_testrect (dx, dy)) {
		
		while (dy--) {
			
			MemoryPtr src_line = this->lines[y++];
			
			this->ConvertToRGB (src_line, this->line_buffer, x, dx);
			this->ExtractMonoFromRGB (this->line_buffer, mono, px, dx);
			
			mono += pitch;
		}
	}
}


/*
 *	Insert a color plane originating from the specified 8-bit surface.
 *	No tiling is done (while RectCopy/RectBlend do tiling).
 */

void
GfxSurface::RectInsert (Int32 x, Int32 y, Card32 dx, Card32 dy, Int32 sx, Int32 sy, GfxSurface* surface, ColorPlane plane)
{
	if (surface == 0) return;
	if (surface->depth != 8) return;
	
	Int32 ox = x;
	Int32 oy = y;
	
	q_gfx_cliprect (x, y, dx, dy, this->clip_x1, this->clip_y1, this->clip_x2, this->clip_y2);
	
	sx += x - ox;
	sy += y - oy;
	
	q_gfx_cliprect (sx, sy, dx, dy, surface->dx, surface->dy);
	
	if (q_gfx_testrect (dx, dy)) {
		
		Card32 shift = 0;
		Card32 mask  = 0xFFFFFFFF;
		
		switch (plane) {
			case PLANE_ALPHA: shift = 24; mask = 0x00FFFFFF; break;
			case PLANE_RED:   shift = 16; mask = 0xFF00FFFF; break;
			case PLANE_GREEN: shift =  8; mask = 0xFFFF00FF; break;
			case PLANE_BLUE:  shift =  0; mask = 0xFFFFFF00; break;
		}
		
		//	Insert the specified plane (either alpha, red, green or blue) from the
		//	8-bit source surface into the current surface.
		
		while (dy--) {
			
			MemoryPtr src_line = surface->lines[sy++];
			MemoryPtr dst_line = this->lines[y++];
			
			//	Following steps are required :
			//
			//	1. Convert the destination into alpha-RGB (into the temporary buffer).
			//	2. Insert the 8-bit color plane into one of the alpha-RGB planes (still
			//	   in the temporary buffer).
			//	3. Copy the merged alpha-RGB temporary buffer back to the destination.
			
			this->ConvertToRGB (dst_line, this->line_buffer, x, dx);
			surface->InsertIntoRGB (src_line, this->line_buffer, sx, dx, shift, mask);
			this->ConvertFromRGB (this->line_buffer, dst_line, x, dx);
		}
	}
}

/*****************************************************************************/

GfxSurface*
GfxSurface::New (Card32 bpp)
{
	return new GfxSurface (bpp);
}

void
GfxSurface::Delete (GfxSurface*& surface)
{
	if (surface) {
		delete surface;
		surface = 0;
	}
}

/*****************************************************************************/
