/*
 *	conviewer.cxx
 *
 *	The connection viewer is a tool which allows to inspect and
 *	re-order connections between events and actors.
 *	
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "buildedit.h"
#include "conviewer.h"

#include <OPaC/boxview.h>
#include <OPaC/button.h>
#include <OPaC/cellgrid.h>
#include <OPaC/database.h>
#include <OPaC/datarep.h>
#include <OPaC/new.h>
#include <OPaC/object.h>
#include <OPaC/pubmacro.h>
#include <OPaC/scrgrid.h>
#include <OPaC/selgrid.h>
#include <OPaC/switch.h>
#include <OPaC/targetmacro.h>
#include <OPaC/tablist.h>
#include <OPaC/tabview.h>
#include <OPaC/textcaption.h>
#include <OPaC/textline.h>
#include <OPaC/textstyle.h>
#include <OPaC/texttag.h>
#include <OPaC/values.h>
#include <OPaC/view.h>
#include <OPaC/window.h>

/****************************************************************/

#define	MAX_CONNECT		100

/****************************************************************/

OP_IMPLEMENTATION_FILE;
APP_IMPLEMENT_CLASS (OPBuildConViewer);

OP_TARGET_ACTOR (OPBuildConViewer, ActorCellActivated);
OP_TARGET_ACTOR (OPBuildConViewer, ActorCellDragged);
OP_TARGET_ACTOR (OPBuildConViewer, DisconnectActor);
OP_TARGET_ACTOR (OPBuildConViewer, EventCellActivated);
OP_TARGET_ACTOR (OPBuildConViewer, Update);

/****************************************************************/

OPBuildConViewer::OPBuildConViewer ()
{
	OP_METHOD (OPBuildConViewer, 0);
	
	this->SetAttribute ("OPaC.volatile.observer");
	this->source_selected = 0;
}

OPBuildConViewer::~OPBuildConViewer ()
{
	OP_METHOD (~OPBuildConViewer, 0);
}

/****************************************************************/

/*
 *	Create a connection viewer for the specified object. This
 *	will handle two scroll lists and a few buttons.
 */

void
OPBuildConViewer::Initialise (OPObjectIn object, OPStringIn title_events, OPStringIn title_actors)
{
	this->object = object;
	
	this->view_events = OP_NEW (OPDBScrollList);
	this->view_actors = OP_NEW (OPDBScrollList);
	
	OPCoord x, y;
	OPCoord dx, dy;
	
	this->GetOrigin (x, y);
	this->GetSize (dx, dy);
	
	x  += 5;
	y  += 5;
	dx -= 20;
	dx /= 2;
	dy -= 40;
	
	this->view_events->SetParent (this);
	this->view_events->SetOrigin (x, y);
	this->view_events->SetSize (dx, dy);
	this->view_events->SetupDataBase (title_events);
	this->view_events->AttachObserver (this, "SelectionChanged", "EventCellActivated");
	this->view_events->ReturnGrid ()->SetSelectAtLeastOne (TRUE);
	this->view_events->Update ();
	
	this->view_actors->SetParent (this);
	this->view_actors->SetOrigin (x+dx+10, y);
	this->view_actors->SetSize (dx, dy);
	this->view_actors->SetupDataBase (title_actors, FALSE);
	this->view_actors->AttachObserver (this, "SelectionChanged", "ActorCellActivated");
	this->view_actors->Update ();
	
	OPSelGridRef grid = this->view_actors->ReturnGrid ();
	
	grid->EnableDragging (TRUE);
	grid->SetInsertBetweenLines (TRUE);
	grid->SetInsertBetweenColumns (FALSE);
	grid->SetInsertOverCell (FALSE);
	grid->AttachObserver (this, "CellDragged", "ActorCellDragged");
	
	OPBuildEdit::MakeEventList (this->view_events, this->object);
	
	this->button_disconnect = OP_NEW (OPButton);
	this->button_disconnect->Initialise (x+2*dx+10-80, y+dy+5, 80, 24, "Disconnect", this);
	this->button_disconnect->AttachObserver (this, "OnClick", "DisconnectActor");
	
	this->view_events->SetSelectedCell (0);
}


/*
 *	Update the actor list according to the selected event
 *	in the left scroll list.
 */

Bool
OPBuildConViewer::Update ()
{
	OPAtomRef find;
	
	this->view_actors->RemoveAll ();
	
	if (this->view_events->GetSelectedName (find)) {
		
		Count num   = this->object->CountAttached ();
		Count index = 0;
		
		while (index < num) {
			
			OPAtomRef   event;
			OPAtomRef   actor;
			OPObjectRef observer;
			Card32      id;
			
			this->object->GetAttached (index, event, actor, observer, id);
			
			if ( (event == find)
			  && (observer->IsKindOf (OP_RUNTIME (OPDataReport)) == FALSE) ) {
				OPStringRef text = "";
				text->AppendString (observer->ReturnRuntime ()->ReturnClassName ());
				text->AppendString ("::");
				text->AppendString (actor);
				
				OPValueRef idx = index;
				OPAtomRef name = text;
				
				idx->SetAttribute ("Actor", actor);
				idx->SetAttribute ("ID", (OPValueRef)id);
				
				this->view_actors->Insert (idx, name, observer, text);
			}
			
			index++;
		}
	}
	
	OPBuildEdit::MakeEventList (this->view_events, this->object);
	
	this->view_events->SetSelectedCell (this->source_selected, FALSE);
	this->view_events->Display ();
	this->view_actors->Update ();
	this->view_actors->Display ();
	
	return TRUE;
}


/*
 *	A cell has been highlighted in the event list. Update
 *	the actor list accordingly.
 */

Bool
OPBuildConViewer::EventCellActivated ()
{
	Count selected = this->source_selected;
	this->view_events->GetSelectedCell (selected);
	
	if (selected == this->source_selected) {
		return TRUE;
	}
	
	this->source_selected = selected;
	this->Update ();
	return TRUE;
}


/*
 *	A cell has been highlighted in the actor list. This
 *	might display a description of the selected observer.
 */

Bool
OPBuildConViewer::ActorCellActivated ()
{
	this->UpdateSelectedInfo ();
	
	//	NIY
	
	return TRUE;
}


/*
 *	An actor cell has been dragged from a position to another
 *	one in the scroll list. This will have to re-order the
 *	attachments accordingly.
 */

Bool
OPBuildConViewer::ActorCellDragged ()
{
	OP_METHOD (ActorCellDragged, 0);
	
	OPSelGridRef grid;
	
	Count sh = 0;
	Count sv = 0;
	Count dh = 0;
	Count dv = 0;
	Real  ph = 0;
	Real  pv = 0;
	
	this->view_events->GetSelectedName (this->event_name);
	
	grid = this->view_actors->ReturnGrid ();
	grid->GetDragOrigin (sh, sv);
	grid->GetDropPosition (dh, dv, ph, pv);
	
	if (pv > 0) {
		dv++;
	}
	if (sv < dv) {
		dv--;
	}
	if (sv == dv) {
		return TRUE;
	}
	
	//	Find the absolute position of the selected source and destination
	//	cells in the list of attached actors. This must take into account
	//	the event to which the actors have been attached.
	
	Count num = this->object->CountAttached ();
	Count pos = 0;
	Count i   = 0;
	Count real_s = 0;
	Count real_d = 0;
	
	for (i = 0, pos = 0; i < num; i++) {
		
		OPAtomRef event;
		OPAtomRef actor;
		OPObjectRef observer;
		Card32 id;
		
		this->object->GetAttached (i, event, actor, observer, id);
		
		if ( (event == this->event_name)
		  && (observer->IsKindOf (OP_RUNTIME (OPDataReport)) == FALSE) ) {
			if (sv == pos) real_s = i;
			if (dv == pos) real_d = i;
			pos++;
		}
	}
	
	//	Reorder the elements so that the source moves to the destination
	//	position. This means that we will have to remove every observer and
	//	then attach them back again in the new order.
	
	OP_ASSERT (num < MAX_CONNECT);
	
	OPAtomRef   events[MAX_CONNECT];
	OPAtomRef   actors[MAX_CONNECT];
	OPObjectRef observers[MAX_CONNECT];
	Card32      ids[MAX_CONNECT];
	Count       total = 1;
	
	//	Copy the element from its source position to its destination in the array.
	
	this->object->GetAttached (real_s, events[dv], actors[dv], observers[dv], ids[dv]);
	
	//	Copy the other elements to their respective position (skip the element which
	//	has already been copied).
	
	for (i = 0, pos = 0; i < num; i++) {
		
		if (i == real_s) continue;
		
		OPAtomRef event;
		OPAtomRef actor;
		OPObjectRef observer;
		Card32 id;
		
		this->object->GetAttached (i, event, actor, observer, id);
		
		if ( (event == this->event_name)
		  && (observer->IsKindOf (OP_RUNTIME (OPDataReport)) == FALSE) ) {
			
			if (pos == dv) pos++;
			
			events[pos]    = event;
			actors[pos]    = actor;
			observers[pos] = observer;
			ids[pos]       = id;
			
			pos++;
			total++;
		}
	}
	
	//	Remove all the observers and then add them back again in the proper order.
	
	for (i = 0; i < total; i++) {
		this->object->DetachObserver (observers[i], events[i], actors[i], ids[i]);
	}
	for (i = 0; i < total; i++) {
		this->object->AttachObserver (observers[i], events[i], actors[i], ids[i]);
	}
	
	this->Update ();
	
	return TRUE;
}


/*
 *	The user has clicked on the disconnection button. This
 *	will just remove the highlighted action attached to the
 *	event shown in the left scroll list.
 */

Bool
OPBuildConViewer::DisconnectActor ()
{
	OP_METHOD (DisconnectActor, 0);
	
	this->UpdateSelectedInfo ();
	
	if ( (this->view_events->CountSelectedItems ())
	  && (this->view_actors->CountSelectedItems ()) ) {
		this->object->DetachObserver (this->observer, this->event_name,
									  this->observer_name, this->observer_id);
		this->Update ();
	}
	
	return TRUE;
}


/*
 *	Update internal information about the selected items
 *	in both scroll lists.
 */

Bool
OPBuildConViewer::UpdateSelectedInfo ()
{
	OP_METHOD (UpdateSelectedInfo, 0);
	
	this->event_name    = OPAtomNull;
	this->observer      = OPObjectNull;
	this->observer_name = OPAtomNull;
	this->observer_id   = 0;
	
	if (this->view_events->CountSelectedItems ()) {
		this->view_events->GetSelectedName (this->event_name);
	}
	
	if ( (this->view_events->CountSelectedItems ())
	  && (this->view_actors->CountSelectedItems ()) ) {
		
		OPAtomRef   name;
		OPValueRef  value;
		
		this->view_actors->GetSelectedIndex (value);
		this->view_actors->GetSelectedName (name);
		this->view_actors->GetSelectedData (this->observer);
		
		value->GetAttribute ("Actor", this->observer_name);
		value->GetAttribute ("ID", value);
		
		this->observer_id = value->ReturnCard32 ();
	}
	
	return TRUE;
}


/*
 *	Return the selected event in the left scroll list.
 */

Bool
OPBuildConViewer::GetSelectedEvent (OPAtomOut event)
{
	this->UpdateSelectedInfo ();
	
	event = this->event_name;
	
	return TRUE;
}


/*
 *	Return the selected observer/actor in the right
 *	scroll list.
 */

Bool
OPBuildConViewer::GetSelectedActor (OPObjectOut observer, OPAtomOut name, Card32Out id)
{
	this->UpdateSelectedInfo ();
	
	observer = this->observer;
	name     = this->observer_name;
	id       = this->observer_id;
	
	return TRUE;
}

